diff -pruN 1.16.4-4/.github/workflows/build-and-test.yml 1.16.5-1/.github/workflows/build-and-test.yml
--- 1.16.4-4/.github/workflows/build-and-test.yml	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/.github/workflows/build-and-test.yml	2024-06-28 13:17:26.000000000 +0000
@@ -12,12 +12,13 @@ jobs:
         os:
           - ubuntu-20.04
           - ubuntu-22.04
+          - ubuntu-24.04
         compiler:
           - gcc
           - clang
-        tracing:
-          - LIBDQLITE_TRACE=1
-          - NOLIBDQLITE_TRACE=1
+        dqlite-next:
+          - yes
+          - no
     runs-on: ${{ matrix.os }}
 
     steps:
@@ -33,20 +34,26 @@ jobs:
         CC: ${{ matrix.compiler }}
       run: |
           autoreconf -i
-          ./configure --enable-debug --enable-code-coverage --enable-sanitize --enable-build-raft
-          make -j4
+          ./configure --enable-debug --enable-code-coverage --enable-sanitize \
+                      --enable-build-raft --enable-dqlite-next=${{ matrix.dqlite-next }}
+          make -j4 unit-test integration-test \
+                   raft-core-fuzzy-test \
+                   raft-core-integration-test \
+                   raft-core-unit-test \
+                   raft-uv-integration-test \
+                   raft-uv-unit-test
 
     - name: Test
       env:
         CC: ${{ matrix.compiler }}
+        LIBDQLITE_TRACE: 1
       run: |
-           export ${{ matrix.tracing }}
-           make -j4 check || (cat ./test-suite.log && false)
+           make check || (cat ./test-suite.log && false)
 
     - name: Coverage
       env:
         CC: ${{ matrix.compiler }}
-      if: ${{ matrix.compiler == 'gcc' }}
+      if: ${{ matrix.os == 'ubuntu-22.04' && matrix.compiler == 'gcc' }}
       run: |
         make code-coverage-capture
 
diff -pruN 1.16.4-4/.github/workflows/coverity.yml 1.16.5-1/.github/workflows/coverity.yml
--- 1.16.4-4/.github/workflows/coverity.yml	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/.github/workflows/coverity.yml	2024-06-28 13:17:26.000000000 +0000
@@ -21,9 +21,8 @@ jobs:
 
       - name: Install dependencies
         run: |
-          sudo add-apt-repository -y ppa:dqlite/dev
           sudo apt-get update -qq
-          sudo apt-get install -qq gcc libsqlite3-dev liblz4-dev libuv1-dev libraft-dev
+          sudo apt-get install -qq gcc libsqlite3-dev liblz4-dev libuv1-dev
 
       - name: Run coverity
         run: |
@@ -33,7 +32,7 @@ jobs:
           autoreconf -i
           mkdir build
           cd build
-          ../configure
+          ../configure --enable-build-raft
 
           # Build
           cov-build --dir cov-int make -j4
diff -pruN 1.16.4-4/.github/workflows/latest-deps.yml 1.16.5-1/.github/workflows/latest-deps.yml
--- 1.16.4-4/.github/workflows/latest-deps.yml	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/.github/workflows/latest-deps.yml	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,73 @@
+name: CI Tests (latest deps)
+
+on:
+  - push
+  - pull_request
+
+jobs:
+  build-and-test:
+    runs-on: ubuntu-latest
+
+    steps:
+    - uses: actions/checkout@v4
+
+    - name: Raise aio-max-nr
+      run: |
+          sysctl fs.aio-max-nr
+          sudo sysctl -w fs.aio-max-nr=1000000
+
+    - name: Install latest libuv
+      run: |
+          version="$(curl -L 'https://dist.libuv.org/dist' | grep -o 'v[0-9]\.[0-9]\{1,2\}\.[0-9]\{1,2\}' | sort -V -r | head -n1)"
+          echo "Selected libuv $version"
+          curl -LO "https://dist.libuv.org/dist/$version/libuv-$version.tar.gz"
+          tar xzf "libuv-$version.tar.gz"
+          cd "libuv-$version"
+          sh autogen.sh
+          ./configure
+          make -j4
+          sudo make install
+
+    - name: Install latest liblz4
+      run: |
+          mkdir lz4
+          cd lz4
+          git init
+          git remote add github 'https://github.com/lz4/lz4'
+          git fetch github 'refs/tags/*:refs/tags/*'
+          version="$(git tag | sort -V -r | head -n1)"
+          echo "Selected lz4 $version"
+          git checkout "$version"
+          make -j4
+          sudo make install
+
+    - name: ldconfig
+      run: |
+          sudo ldconfig
+
+    - name: Get latest SQLite
+      run: |
+          relative="$(curl -L 'https://sqlite.org/download.html' | grep '^PRODUCT' | grep 'amalgamation' | cut -d',' -f3)"
+          curl -LO "https://sqlite.org/$relative"
+          name="$(basename "$relative" .zip)"
+          echo "Selected $name"
+          unzip "$name.zip"
+          cd "$name"
+          cp sqlite3.{c,h} "$GITHUB_WORKSPACE"
+
+    - name: Build dqlite
+      run: |
+          autoreconf -i
+          ./configure --enable-debug --enable-sanitize --enable-build-raft --enable-build-sqlite
+          make -j4 unit-test integration-test \
+                   raft-core-fuzzy-test \
+                   raft-core-integration-test \
+                   raft-core-unit-test \
+                   raft-uv-integration-test \
+                   raft-uv-unit-test
+          ldd .libs/libdqlite.so
+
+    - name: Test
+      run: |
+           export LIBDQLITE_TRACE=1
+           make check || (cat ./test-suite.log && false)
diff -pruN 1.16.4-4/.gitignore 1.16.5-1/.gitignore
--- 1.16.4-4/.gitignore	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/.gitignore	2024-06-28 13:17:26.000000000 +0000
@@ -26,3 +26,8 @@ dqlite.pc
 libtool
 stamp-h*
 sqlite3.c
+raft-core-fuzzy-test
+raft-core-integration-test
+raft-core-unit-test
+raft-uv-integration-test
+raft-uv-unit-test
diff -pruN 1.16.4-4/CODE_OF_CONDUCT.md 1.16.5-1/CODE_OF_CONDUCT.md
--- 1.16.4-4/CODE_OF_CONDUCT.md	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/CODE_OF_CONDUCT.md	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,3 @@
+dqlite has adopted the [Ubuntu Code of Conduct](coc).
+
+[coc]: https://ubuntu.com/community/ethos/code-of-conduct
diff -pruN 1.16.4-4/CONTRIBUTING.md 1.16.5-1/CONTRIBUTING.md
--- 1.16.4-4/CONTRIBUTING.md	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/CONTRIBUTING.md	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,11 @@
+# Contributing to dqlite
+
+The dqlite team welcomes external contributions via GitHub pull
+requests. To get your PR merged, you need to sign [Canonical's
+contributor license agreement (CLA)][cla]. This is straightforward to
+do once you have an account on [Launchpad][lp]; if you don't, you can
+create one [here][signup].
+
+[cla]: https://ubuntu.com/legal/contributors
+[lp]: https://launchpad.net
+[signup]: https://launchpad.net/+login
diff -pruN 1.16.4-4/Makefile.am 1.16.5-1/Makefile.am
--- 1.16.4-4/Makefile.am	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/Makefile.am	2024-06-28 13:17:26.000000000 +0000
@@ -1,28 +1,31 @@
 ACLOCAL_AMFLAGS = -I m4
+
 AM_CFLAGS += $(CODE_COVERAGE_CFLAGS)
 AM_CFLAGS += $(SQLITE_CFLAGS) $(UV_CFLAGS) $(PTHREAD_CFLAGS)
 AM_LDFLAGS = $(UV_LIBS) $(PTHREAD_LIBS)
 
-if !BUILD_SQLITE_ENABLED
-AM_LDFLAGS += $(SQLITE_LIBS)
+if DQLITE_NEXT_ENABLED
+AM_CFLAGS += -DDQLITE_NEXT
 endif
 
 if !BUILD_RAFT_ENABLED
-AM_CFLAGS += $(RAFT_CFLAGS)
+AM_CFLAGS += $(RAFT_CFLAGS) -DUSE_SYSTEM_RAFT
 AM_LDFLAGS += $(RAFT_LIBS)
 endif
 
 if DEBUG_ENABLED
-  AM_CFLAGS += -g3
+AM_CFLAGS += -O0
 else
-  AM_CFLAGS += -O2
+AM_CFLAGS += -O2
 endif
+
 if SANITIZE_ENABLED
-  AM_CFLAGS += -fsanitize=address
+AM_CFLAGS += -fsanitize=address
 endif
+
 if BACKTRACE_ENABLED
-  AM_CFLAGS += -DDQLITE_ASSERT_WITH_BACKTRACE -DRAFT_ASSERT_WITH_BACKTRACE
-  AM_LDFLAGS += -lbacktrace
+AM_CFLAGS += -DDQLITE_ASSERT_WITH_BACKTRACE -DRAFT_ASSERT_WITH_BACKTRACE
+AM_LDFLAGS += -lbacktrace
 endif
 
 include_HEADERS = include/dqlite.h
@@ -61,11 +64,8 @@ basic_dqlite_sources = \
   src/transport.c \
   src/translate.c \
   src/tuple.c \
-  src/vfs.c
-
-if BUILD_SQLITE_ENABLED
-basic_dqlite_sources += sqlite3.c
-endif
+  src/vfs.c \
+  src/vfs2.c
 
 lib_LTLIBRARIES = libdqlite.la
 libdqlite_la_CFLAGS = $(AM_CFLAGS) -fvisibility=hidden -DRAFT_API=''
@@ -168,6 +168,7 @@ unit_test_SOURCES += \
   test/unit/test_sm.c \
   test/unit/test_tuple.c \
   test/unit/test_vfs.c \
+  test/unit/test_vfs2.c \
   test/unit/main.c
 unit_test_CFLAGS = $(AM_CFLAGS) -Wno-unknown-warning-option -Wno-uninitialized -Wno-maybe-uninitialized -Wno-float-equal -Wno-conversion
 unit_test_LDFLAGS = $(AM_LDFLAGS)
@@ -322,6 +323,7 @@ libraft_la_LDFLAGS += $(LZ4_LIBS)
 raft_uv_integration_test_CFLAGS += -DLZ4_AVAILABLE
 raft_uv_integration_test_LDFLAGS += $(LZ4_LIBS)
 endif
+
 if LZ4_ENABLED
 libdqlite_la_CFLAGS += -DLZ4_ENABLED
 raft_uv_integration_test_CFLAGS += -DLZ4_ENABLED
@@ -331,6 +333,17 @@ endif
 
 endif # BUILD_RAFT_ENABLED
 
+if BUILD_SQLITE_ENABLED
+noinst_LTLIBRARIES = libsqlite3.la
+libsqlite3_la_SOURCES = sqlite3.c
+libsqlite3_la_CFLAGS = -g3
+
+unit_test_LDADD += libsqlite3.la
+libdqlite_la_LIBADD = libsqlite3.la
+else
+AM_LDFLAGS += $(SQLITE_LIBS)
+endif
+
 TESTS = $(check_PROGRAMS)
 
 if CODE_COVERAGE_ENABLED
diff -pruN 1.16.4-4/README.md 1.16.5-1/README.md
--- 1.16.4-4/README.md	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/README.md	2024-06-28 13:17:26.000000000 +0000
@@ -70,6 +70,11 @@ sudo apt update
 sudo apt install libdqlite-dev
 ```
 
+Contributing
+------------
+
+See [CONTRIBUTING.md](./CONTRIBUTING.md).
+
 Build
 -----
 
diff -pruN 1.16.4-4/bt/request 1.16.5-1/bt/request
--- 1.16.4-4/bt/request	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/bt/request	2024-06-28 13:17:26.000000000 +0000
@@ -11,7 +11,7 @@ struct request
     void *data;
     int type;
     raft_index index;
-    void *queue[2];
+    queue queue;
 };
 
 uprobe:$libraft_path:lifecycleRequestStart
diff -pruN 1.16.4-4/configure.ac 1.16.5-1/configure.ac
--- 1.16.4-4/configure.ac	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/configure.ac	2024-06-28 13:17:26.000000000 +0000
@@ -1,11 +1,13 @@
 AC_PREREQ(2.60)
-AC_INIT([libdqlite], [1.16.4], [https://github.com/canonical/dqlite])
+AC_INIT([libdqlite], [1.16.5], [https://github.com/canonical/dqlite])
 AC_CONFIG_MACRO_DIR([m4])
 AC_CONFIG_AUX_DIR([ac])
 
 AM_INIT_AUTOMAKE([subdir-objects -Wall -Werror -Wno-portability foreign])
 AM_SILENT_RULES([yes])
 
+# Without this line, AC_PROG_CC boneheadedly adds `-g -O2` to our CFLAGS.
+AC_SUBST(CFLAGS, "")
 AC_PROG_CC
 AC_USE_SYSTEM_EXTENSIONS
 
@@ -37,6 +39,10 @@ AM_CONDITIONAL(BUILD_SQLITE_ENABLED, tes
 AC_ARG_ENABLE(build-raft, AS_HELP_STRING([--enable-build-raft[=ARG]], [use the bundled raft sources instead of linking to libraft [default=no]]))
 AM_CONDITIONAL(BUILD_RAFT_ENABLED, test "x$enable_build_raft" = "xyes")
 
+AC_ARG_ENABLE(dqlite-next, AS_HELP_STRING([--enable-dqlite-next[=ARG]], [build with the experimental dqlite backend [default=no]]))
+AM_CONDITIONAL(DQLITE_NEXT_ENABLED, test "x$enable_dqlite_next" = "xyes")
+AS_IF([test "x$enable_build_raft" != "xyes" -a "x$enable_dqlite_next" = "xyes"], [AC_MSG_ERROR([dqlite-next requires bundled raft])], [])
+
 # Whether to enable code coverage.
 AX_CODE_COVERAGE
 
@@ -53,7 +59,7 @@ AC_SYS_LARGEFILE
 
 # Checks for libraries
 PKG_CHECK_MODULES(SQLITE, [sqlite3 >= 3.22.0], [], [])
-PKG_CHECK_MODULES(UV, [libuv >= 1.8.0], [], [])
+PKG_CHECK_MODULES(UV, [libuv >= 1.34.0], [], [])
 AS_IF([test "x$enable_build_raft" != "xyes"], [PKG_CHECK_MODULES(RAFT, [raft >= 0.18.1], [], [])], [])
 
 
diff -pruN 1.16.4-4/debian/changelog 1.16.5-1/debian/changelog
--- 1.16.4-4/debian/changelog	2024-05-31 14:12:00.000000000 +0000
+++ 1.16.5-1/debian/changelog	2024-06-28 13:24:34.000000000 +0000
@@ -1,3 +1,11 @@
+dqlite (1.16.5-1) unstable; urgency=medium
+
+  * New upstream release
+    - Drop patches applied upstream
+    - Bump minimum required version of libuv1 in d/control
+
+ -- Mathias Gibbens <gibmat@debian.org>  Fri, 28 Jun 2024 13:24:34 +0000
+
 dqlite (1.16.4-4) unstable; urgency=medium
 
   * Cherry-pick upstream fix for nanosecond calculations
diff -pruN 1.16.4-4/debian/control 1.16.5-1/debian/control
--- 1.16.4-4/debian/control	2024-05-27 16:01:23.000000000 +0000
+++ 1.16.5-1/debian/control	2024-06-28 13:23:20.000000000 +0000
@@ -1,5 +1,5 @@
 Source: dqlite
-Build-Depends: debhelper-compat (= 13), pkgconf, liblz4-dev, libuv1-dev (>= 1.8.0~), libsqlite3-dev (>= 3.22.0~)
+Build-Depends: debhelper-compat (= 13), pkgconf, liblz4-dev, libuv1-dev (>= 1.34.0~), libsqlite3-dev (>= 3.22.0~)
 Section: devel
 Priority: optional
 Maintainer: Free Ekanayaka <freee@debian.org>
@@ -30,7 +30,7 @@ Description: Embeddable, replicated and
 Package: libdqlite-dev
 Section: libdevel
 Architecture: any
-Depends: libdqlite0 (= ${binary:Version}), liblz4-dev, libuv1-dev (>= 1.8.0~), libsqlite3-dev (>= 3.22.0~), ${misc:Depends}
+Depends: libdqlite0 (= ${binary:Version}), liblz4-dev, libuv1-dev (>= 1.34.0~), libsqlite3-dev (>= 3.22.0~), ${misc:Depends}
 Multi-Arch: same
 Description: Embeddable, replicated and fault tolerant SQL engine - development files
  Embeddable SQL database engine with high availability and automatic failover.
diff -pruN 1.16.4-4/debian/patches/0003-fix-alpha-x32-build.patch 1.16.5-1/debian/patches/0003-fix-alpha-x32-build.patch
--- 1.16.4-4/debian/patches/0003-fix-alpha-x32-build.patch	2024-05-31 14:09:14.000000000 +0000
+++ 1.16.5-1/debian/patches/0003-fix-alpha-x32-build.patch	1970-01-01 00:00:00.000000000 +0000
@@ -1,63 +0,0 @@
-From 822519192bbccc04b261434c0952bc646381254d Mon Sep 17 00:00:00 2001
-From: Cole Miller <cole.miller@canonical.com>
-Date: Tue, 23 Apr 2024 12:55:50 +0100
-Subject: [PATCH] Three portability fixes around integer width
-
-This unbreaks the dqlite build on alpha and x32.
-
-Co-authored-by: Mathias Gibbens <gibmat@debian.org>
-Signed-off-by: Cole Miller <cole.miller@canonical.com>
----
- src/raft/uv_fs.c | 10 ++--------
- src/raft/uv_os.c |  2 +-
- src/tracing.c    |  2 +-
- 3 files changed, 4 insertions(+), 10 deletions(-)
-
-diff --git a/src/raft/uv_fs.c b/src/raft/uv_fs.c
-index b4bf42114..e2e1c8dd4 100644
---- a/src/raft/uv_fs.c
-+++ b/src/raft/uv_fs.c
-@@ -704,15 +704,9 @@ static int probeDirectIO(int fd, size_t *size, char *errmsg)
- 			default:
- 				/* UNTESTED: this is an unsupported file system.
- 				 */
--#if defined(__s390x__)
- 				ErrMsgPrintf(errmsg,
--					     "unsupported file system: %ux",
--					     fs_info.f_type);
--#else
--				ErrMsgPrintf(errmsg,
--					     "unsupported file system: %zx",
--					     fs_info.f_type);
--#endif
-+					     "unsupported file system: %llx",
-+					     (unsigned long long)fs_info.f_type);
- 				return RAFT_IOERR;
- 		}
- 	}
-diff --git a/src/raft/uv_os.c b/src/raft/uv_os.c
-index 8a96ab130..a8a972210 100644
---- a/src/raft/uv_os.c
-+++ b/src/raft/uv_os.c
-@@ -52,7 +52,7 @@ int UvOsFallocateEmulation(int fd, off_t offset, off_t len)
- 	if (f.f_bsize == 0) {
- 		increment = 512;
- 	} else if (f.f_bsize < 4096) {
--		increment = f.f_bsize;
-+		increment = (ssize_t)f.f_bsize;
- 	} else {
- 		increment = 4096;
- 	}
-diff --git a/src/tracing.c b/src/tracing.c
-index 96c148e28..0314a54fe 100644
---- a/src/tracing.c
-+++ b/src/tracing.c
-@@ -79,7 +79,7 @@ static inline void tracerEmit(const char *file,
- 		tracerPidCached,
- 
- 		tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour,
--		tm.tm_min, tm.tm_sec, ts.tv_nsec,
-+		tm.tm_min, tm.tm_sec, (unsigned long)ts.tv_nsec,
- 
- 		(unsigned)tid, tracerTraceLevelName(level), func,
- 		tracerShortFileName(file), line, message);
diff -pruN 1.16.4-4/debian/patches/0004-cherry-pick-bea9e6f.patch 1.16.5-1/debian/patches/0004-cherry-pick-bea9e6f.patch
--- 1.16.4-4/debian/patches/0004-cherry-pick-bea9e6f.patch	2024-04-24 21:08:37.000000000 +0000
+++ 1.16.5-1/debian/patches/0004-cherry-pick-bea9e6f.patch	1970-01-01 00:00:00.000000000 +0000
@@ -1,45 +0,0 @@
-From bea9e6f72c6a6aff628a3362cdc8dc6fabbb11c0 Mon Sep 17 00:00:00 2001
-From: Free Ekanayaka <free@ekanayaka.io>
-Date: Fri, 1 Mar 2024 09:36:09 -0500
-Subject: [PATCH] raft/uv: Drop AI_V4MAPPED | AI_ADDRCONFIG from getaddrinfo
-
-Since we're currently only supporting IPv4 and since we're setting
-ai_family to AF_INET, the AI_ADDRCONFIG flag is not really making any
-difference as getaddrinfo returns only IPv4 addresses.
-
-This makes the test suite pass also if there isn't a non-loopback
-network interface on the system configured with IPv4.
-
-Signed-off-by: Free Ekanayaka <free@ekanayaka.io>
-Signed-off-by: Cole Miller <cole.miller@canonical.com>
----
- src/raft/uv_ip.c          | 2 +-
- src/raft/uv_tcp_connect.c | 2 +-
- 2 files changed, 2 insertions(+), 2 deletions(-)
-
-diff --git a/src/raft/uv_ip.c b/src/raft/uv_ip.c
-index 4e4ff9f3f..ffc0ba44f 100644
---- a/src/raft/uv_ip.c
-+++ b/src/raft/uv_ip.c
-@@ -58,7 +58,7 @@ int uvIpAddrSplit(const char *address,
- int uvIpResolveBindAddresses(const char *address, struct addrinfo **ai_result)
- {
- 	static struct addrinfo hints = {
--	    .ai_flags = AI_ADDRCONFIG | AI_PASSIVE | AI_NUMERICSERV,
-+	    .ai_flags = AI_PASSIVE | AI_NUMERICSERV,
- 	    .ai_family = AF_INET,
- 	    .ai_socktype = SOCK_STREAM,
- 	    .ai_protocol = 0};
-diff --git a/src/raft/uv_tcp_connect.c b/src/raft/uv_tcp_connect.c
-index e493d14a8..a13c68531 100644
---- a/src/raft/uv_tcp_connect.c
-+++ b/src/raft/uv_tcp_connect.c
-@@ -261,7 +261,7 @@ static void uvGetAddrInfoCb(uv_getaddrinfo_t *req,
- /* Create a new TCP handle and submit a connection request to the event loop. */
- static int uvTcpConnectStart(struct uvTcpConnect *r, const char *address)
- {
--	static struct addrinfo hints = {.ai_flags = AI_V4MAPPED | AI_ADDRCONFIG,
-+	static struct addrinfo hints = {.ai_flags = 0,
- 					.ai_family = AF_INET,
- 					.ai_socktype = SOCK_STREAM,
- 					.ai_protocol = 0};
diff -pruN 1.16.4-4/debian/patches/0005-cherry-pick-0d74f3b.patch 1.16.5-1/debian/patches/0005-cherry-pick-0d74f3b.patch
--- 1.16.4-4/debian/patches/0005-cherry-pick-0d74f3b.patch	2024-05-31 14:09:36.000000000 +0000
+++ 1.16.5-1/debian/patches/0005-cherry-pick-0d74f3b.patch	1970-01-01 00:00:00.000000000 +0000
@@ -1,32 +0,0 @@
-From 0d74f3be70225740002466edc72118cfc2c4064f Mon Sep 17 00:00:00 2001
-From: Cole Miller <cole.miller@canonical.com>
-Date: Fri, 1 Mar 2024 08:53:24 -0500
-Subject: [PATCH] client/protocol: Ensure 64 bits for nanosecond calculations
-
-Signed-off-by: Cole Miller <cole.miller@canonical.com>
----
- src/client/protocol.c | 4 ++--
- 1 file changed, 2 insertions(+), 2 deletions(-)
-
-diff --git a/src/client/protocol.c b/src/client/protocol.c
-index 5d5e5694e..6ccc6d2ea 100644
---- a/src/client/protocol.c
-+++ b/src/client/protocol.c
-@@ -126,7 +126,7 @@ static ssize_t doRead(int fd,
- 	ssize_t total;
- 	struct pollfd pfd;
- 	struct timespec now;
--	long millis;
-+	long long millis;
- 	ssize_t n;
- 	int rv;
- 
-@@ -206,7 +206,7 @@ static ssize_t doWrite(int fd,
- 	ssize_t total;
- 	struct pollfd pfd;
- 	struct timespec now;
--	long millis;
-+	long long millis;
- 	ssize_t n;
- 	int rv;
- 
diff -pruN 1.16.4-4/debian/patches/series 1.16.5-1/debian/patches/series
--- 1.16.4-4/debian/patches/series	2024-05-31 14:10:09.000000000 +0000
+++ 1.16.5-1/debian/patches/series	2024-06-28 13:21:53.000000000 +0000
@@ -1,5 +1,2 @@
 0001-fix-flaky-time-dependent-tests.patch
 0002-skip-flaky-tests-on-ppc64el.patch
-0003-fix-alpha-x32-build.patch
-0004-cherry-pick-bea9e6f.patch
-0005-cherry-pick-0d74f3b.patch
diff -pruN 1.16.4-4/include/dqlite.h 1.16.5-1/include/dqlite.h
--- 1.16.4-4/include/dqlite.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/include/dqlite.h	2024-06-28 13:17:26.000000000 +0000
@@ -28,7 +28,7 @@
  */
 #define DQLITE_VERSION_MAJOR 1
 #define DQLITE_VERSION_MINOR 16
-#define DQLITE_VERSION_RELEASE 4
+#define DQLITE_VERSION_RELEASE 5
 #define DQLITE_VERSION_NUMBER                                            \
 	(DQLITE_VERSION_MAJOR * 100 * 100 + DQLITE_VERSION_MINOR * 100 + \
 	 DQLITE_VERSION_RELEASE)
@@ -431,6 +431,16 @@ DQLITE_API int dqlite_node_set_target_vo
  */
 DQLITE_API int dqlite_node_set_target_standbys(dqlite_node *n, int standbys);
 
+
+/**
+ * Set the target number of threads in the thread pool processing sqlite3 disk
+ * operations.
+ *
+ * The default pool thread count is 4.
+ */
+DQLITE_API int dqlite_node_set_pool_thread_count(dqlite_node *n,
+						 unsigned thread_count);
+
 /**
  * Enable or disable auto-recovery for corrupted disk files.
  *
diff -pruN 1.16.4-4/src/client/protocol.c 1.16.5-1/src/client/protocol.c
--- 1.16.4-4/src/client/protocol.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/client/protocol.c	2024-06-28 13:17:26.000000000 +0000
@@ -126,7 +126,7 @@ static ssize_t doRead(int fd,
 	ssize_t total;
 	struct pollfd pfd;
 	struct timespec now;
-	long millis;
+	long long millis;
 	ssize_t n;
 	int rv;
 
@@ -206,7 +206,7 @@ static ssize_t doWrite(int fd,
 	ssize_t total;
 	struct pollfd pfd;
 	struct timespec now;
-	long millis;
+	long long millis;
 	ssize_t n;
 	int rv;
 
diff -pruN 1.16.4-4/src/config.c 1.16.5-1/src/config.c
--- 1.16.4-4/src/config.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/config.c	2024-06-28 13:17:26.000000000 +0000
@@ -31,7 +31,8 @@ static unsigned serial = 1;
 int config__init(struct config *c,
 		 dqlite_node_id id,
 		 const char *address,
-		 const char *dir)
+		 const char *raft_dir,
+		 const char *database_dir)
 {
 	int rv;
 	c->id = id;
@@ -49,11 +50,14 @@ int config__init(struct config *c,
 	c->logger.emit = loggerDefaultEmit;
 	c->failure_domain = 0;
 	c->weight = 0;
-	strncpy(c->dir, dir, sizeof(c->dir) - 1);
-	c->dir[sizeof(c->dir) - 1] = '\0';
+
+	snprintf(c->raft_dir, sizeof(c->raft_dir), "%s", (raft_dir != NULL) ? raft_dir : "");
+	snprintf(c->database_dir, sizeof(c->database_dir), "%s", database_dir);
+
 	c->disk = false;
 	c->voters = 3;
 	c->standbys = 0;
+	c->pool_thread_count = 4;
 	serial++;
 	return 0;
 }
diff -pruN 1.16.4-4/src/config.h 1.16.5-1/src/config.h
--- 1.16.4-4/src/config.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/config.h	2024-06-28 13:17:26.000000000 +0000
@@ -6,8 +6,7 @@
 /**
  * Value object holding dqlite configuration.
  */
-struct config
-{
+struct config {
 	dqlite_node_id id;             /* Unique instance ID */
 	char *address;                 /* Instance address */
 	unsigned heartbeat_timeout;    /* In milliseconds */
@@ -17,10 +16,12 @@ struct config
 	char name[256];                /* VFS/replication registriatio name */
 	unsigned long long failure_domain; /* User-provided failure domain */
 	unsigned long long int weight;     /* User-provided node weight */
-	char dir[1024];                    /* Data dir for on-disk database */
+	char raft_dir[1024];               /* Directory used by raft */
+	char database_dir[1024];           /* Data dir for on-disk database */
 	bool disk;                         /* Disk-mode or not */
 	int voters;                        /* Target number of voters */
 	int standbys;                      /* Target number of standbys */
+	unsigned pool_thread_count;    /* Number of threads in thread pool */
 };
 
 /**
@@ -30,7 +31,8 @@ struct config
 int config__init(struct config *c,
 		 dqlite_node_id id,
 		 const char *address,
-		 const char *dir);
+		 const char *raft_dir,
+		 const char *database_dir);
 
 /**
  * Release any memory held by the config object.
diff -pruN 1.16.4-4/src/conn.c 1.16.5-1/src/conn.c
--- 1.16.4-4/src/conn.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/conn.c	2024-06-28 13:17:26.000000000 +0000
@@ -5,6 +5,8 @@
 #include "tracing.h"
 #include "transport.h"
 
+#include <uv.h>
+
 /* Initialize the given buffer for reading, ensure it has the given size. */
 static int init_read(struct conn *c, uv_buf_t *buf, size_t size)
 {
diff -pruN 1.16.4-4/src/db.c 1.16.5-1/src/db.c
--- 1.16.4-4/src/db.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/db.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,3 +1,4 @@
+#include <stdint.h>
 #include <string.h>
 
 #include "../include/dqlite.h"
@@ -16,12 +17,25 @@ static int open_follower_conn(const char
 			      unsigned page_size,
 			      sqlite3 **conn);
 
+static uint32_t str_hash(const char* name)
+{
+	const unsigned char *p;
+	uint32_t h = 5381U;
+
+	for (p = (const unsigned char *) name; *p != '\0'; p++) {
+		h = (h << 5) + h + *p;
+	}
+
+	return h;
+}
+
 int db__init(struct db *db, struct config *config, const char *filename)
 {
-	tracef("db init %s", filename);
+	tracef("db init filename=`%s'", filename);
 	int rv;
 
 	db->config = config;
+	db->cookie = str_hash(filename);
 	db->filename = sqlite3_malloc((int)(strlen(filename) + 1));
 	if (db->filename == NULL) {
 		rv = DQLITE_NOMEM;
@@ -35,7 +49,7 @@ int db__init(struct db *db, struct confi
 	}
 	if (db->config->disk) {
 		rv = snprintf(db->path, MAX_PATHNAME + 1, "%s/%s",
-			      db->config->dir, db->filename);
+			      db->config->database_dir, db->filename);
 	} else {
 		rv = snprintf(db->path, MAX_PATHNAME + 1, "%s", db->filename);
 	}
@@ -46,7 +60,7 @@ int db__init(struct db *db, struct confi
 	db->follower = NULL;
 	db->tx_id = 0;
 	db->read_lock = 0;
-	QUEUE__INIT(&db->leaders);
+	queue_init(&db->leaders);
 	return 0;
 
 err_after_path_alloc:
@@ -59,7 +73,7 @@ err:
 
 void db__close(struct db *db)
 {
-	assert(QUEUE__IS_EMPTY(&db->leaders));
+	assert(queue_empty(&db->leaders));
 	if (db->follower != NULL) {
 		int rc;
 		rc = sqlite3_close(db->follower);
diff -pruN 1.16.4-4/src/db.h 1.16.5-1/src/db.h
--- 1.16.4-4/src/db.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/db.h	2024-06-28 13:17:26.000000000 +0000
@@ -5,6 +5,7 @@
 #ifndef DB_H_
 #define DB_H_
 
+#include <stdint.h>
 #include "lib/queue.h"
 
 #include "config.h"
@@ -14,6 +15,7 @@ struct db
 	struct config *config; /* Dqlite configuration */
 	char *filename;        /* Database filename */
 	char *path;            /* Used for on-disk db */
+	uint32_t cookie;       /* Used to bind to the pool's thread */
 	sqlite3 *follower;     /* Follower connection */
 	queue leaders;         /* Open leader connections */
 	unsigned tx_id;        /* Current ongoing transaction ID, if any */
diff -pruN 1.16.4-4/src/fsm.c 1.16.5-1/src/fsm.c
--- 1.16.4-4/src/fsm.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/fsm.c	2024-06-28 13:17:26.000000000 +0000
@@ -502,10 +502,10 @@ static unsigned snapshotNumBufs(struct f
 	queue *head;
 	unsigned n = 1; /* snapshot header */
 
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		n += 2; /* database header & wal */
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		n += dbNumPages(db); /* 1 buffer per page (zero copy) */
 	}
 
@@ -539,12 +539,12 @@ static void freeSnapshotBufs(struct fsm
 
 	i = 1;
 	/* Free all database headers & WAL buffers */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		if (i == n_bufs) {
 			break;
 		}
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		/* i is the index of the database header */
 		sqlite3_free(bufs[i].base);
 		/* i is now the index of the next database header (if any) */
@@ -567,9 +567,9 @@ static int fsm__snapshot(struct raft_fsm
 
 	/* First count how many databases we have and check that no transaction
 	 * nor checkpoint nor other snapshot is in progress. */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		if (db->tx_id != 0 || db->read_lock) {
 			return RAFT_BUSY;
 		}
@@ -577,9 +577,9 @@ static int fsm__snapshot(struct raft_fsm
 	}
 
 	/* Lock all databases, preventing the checkpoint from running */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		rv = databaseReadLock(db);
 		assert(rv == 0);
 	}
@@ -598,9 +598,9 @@ static int fsm__snapshot(struct raft_fsm
 
 	/* Encode individual databases. */
 	i = 1;
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		/* database_header + num_pages + wal */
 		unsigned n = 1 + dbNumPages(db) + 1;
 		rv = encodeDatabase(db, &(*bufs)[i], n);
@@ -618,9 +618,9 @@ err_after_encode_header:
 err_after_bufs_alloc:
 	sqlite3_free(*bufs);
 err:
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		databaseReadUnlock(db);
 	}
 	assert(rv != 0);
@@ -663,12 +663,12 @@ static int fsm__snapshot_finalize(struct
 	/* Unlock all databases that were locked for the snapshot, this is safe
 	 * because DB's are only ever added at the back of the queue. */
 	n_db = 0;
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		if (n_db == header.n) {
 			break;
 		}
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		rv = databaseReadUnlock(db);
 		assert(rv == 0);
 		n_db++;
@@ -815,7 +815,7 @@ static unsigned snapshotNumBufsDisk(stru
 	queue *head;
 	unsigned n = 1; /* snapshot header */
 
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		n += 3; /* database header, database file and wal */
 	}
@@ -849,7 +849,7 @@ static void freeSnapshotBufsDisk(struct
 
 	i = 1;
 	/* Free all database headers & WAL buffers. Unmap the DB file. */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		if (i == n_bufs) {
 			break;
@@ -878,9 +878,9 @@ static int fsm__snapshot_disk(struct raf
 
 	/* First count how many databases we have and check that no transaction
 	 * nor checkpoint nor other snapshot is in progress. */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		if (db->tx_id != 0 || db->read_lock) {
 			return RAFT_BUSY;
 		}
@@ -890,9 +890,9 @@ static int fsm__snapshot_disk(struct raf
 	/* Lock all databases, preventing the checkpoint from running. This
 	 * ensures the database is not written while it is mmap'ed and copied by
 	 * raft. */
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		rv = databaseReadLock(db);
 		assert(rv == 0);
 	}
@@ -917,9 +917,9 @@ static int fsm__snapshot_disk(struct raf
 
 	/* Copy WAL of all databases. */
 	i = 1;
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		/* database_header + db + WAL */
 		unsigned n = 3;
 		/* pass pointer to buffer that will contain WAL. */
@@ -938,9 +938,9 @@ err_after_encode_sync:
 err_after_bufs_alloc:
 	sqlite3_free(*bufs);
 err:
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		databaseReadUnlock(db);
 	}
 	assert(rv != 0);
@@ -972,13 +972,13 @@ static int fsm__snapshot_async_disk(stru
 
 	/* Encode individual databases. */
 	i = 1;
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		if (i == *n_bufs) {
 			/* In case a db was added in meanwhile */
 			break;
 		}
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		/* database_header + database file + wal */
 		unsigned n = 3;
 		rv = encodeDiskDatabaseAsync(db, &(*bufs)[i], n);
@@ -1031,12 +1031,12 @@ static int fsm__snapshot_finalize_disk(s
 	/* Unlock all databases that were locked for the snapshot, this is safe
 	 * because DB's are only ever added at the back of the queue. */
 	n_db = 0;
-	QUEUE__FOREACH(head, &f->registry->dbs)
+	QUEUE_FOREACH(head, &f->registry->dbs)
 	{
 		if (n_db == header.n) {
 			break;
 		}
-		db = QUEUE__DATA(head, struct db, queue);
+		db = QUEUE_DATA(head, struct db, queue);
 		databaseReadUnlock(db);
 		n_db++;
 	}
diff -pruN 1.16.4-4/src/gateway.c 1.16.5-1/src/gateway.c
--- 1.16.4-4/src/gateway.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/gateway.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,11 +1,14 @@
 #include "gateway.h"
 
 #include "bind.h"
+#include "conn.h"
 #include "id.h"
+#include "lib/threadpool.h"
 #include "protocol.h"
 #include "query.h"
 #include "request.h"
 #include "response.h"
+#include "server.h"
 #include "tracing.h"
 #include "translate.h"
 #include "tuple.h"
@@ -96,8 +99,8 @@ void gateway__close(struct gateway *g)
  * decode the request. This is used in the common case where only one schema
  * version is extant. */
 #define START_V0(REQ, RES, ...)                                       \
-	struct request_##REQ request = {0};                           \
-	struct response_##RES response = {0};                         \
+	struct request_##REQ request = { 0 };                         \
+	struct response_##RES response = { 0 };                       \
 	{                                                             \
 		int rv_;                                              \
 		if (req->schema != 0) {                               \
@@ -300,8 +303,8 @@ static void prepareBarrierCb(struct barr
 	tracef("prepare barrier cb status:%d", status);
 	struct gateway *g = barrier->data;
 	struct handle *req = g->req;
-	struct response_stmt response_v0 = {0};
-	struct response_stmt_with_offset response_v1 = {0};
+	struct response_stmt response_v0 = { 0 };
+	struct response_stmt_with_offset response_v1 = { 0 };
 	const char *sql = req->sql;
 	struct stmt *stmt;
 	const char *tail;
@@ -372,7 +375,7 @@ static int handle_prepare(struct gateway
 	tracef("handle prepare");
 	struct cursor *cursor = &req->cursor;
 	struct stmt *stmt;
-	struct request_prepare request = {0};
+	struct request_prepare request = { 0 };
 	int rc;
 
 	if (req->schema != DQLITE_PREPARE_STMT_SCHEMA_V0 &&
@@ -462,7 +465,7 @@ static int handle_exec(struct gateway *g
 	tracef("handle exec schema:%" PRIu8, req->schema);
 	struct cursor *cursor = &req->cursor;
 	struct stmt *stmt;
-	struct request_exec request = {0};
+	struct request_exec request = { 0 };
 	int tuple_format;
 	uint64_t req_id;
 	int rv;
@@ -514,17 +517,20 @@ static int handle_exec(struct gateway *g
  * given request with a single batch of rows.
  *
  * A single batch of rows is typically about the size of a memory page. */
-static void query_batch(struct gateway *g)
+static void query_batch_async(struct handle *req, enum pool_half half)
 {
-	struct handle *req = g->req;
-	assert(req != NULL);
-	g->req = NULL;
+	struct gateway *g = req->gw;
 	sqlite3_stmt *stmt = req->stmt;
 	assert(stmt != NULL);
 	struct response_rows response;
 	int rc;
 
-	rc = query__batch(stmt, req->buffer);
+	if (half == POOL_TOP_HALF) {
+		req->work.rc = query__batch(stmt, req->buffer);
+		return;
+	}  /* else POOL_BOTTOM_HALF => */
+	rc = req->work.rc;
+
 	if (rc != SQLITE_ROW && rc != SQLITE_DONE) {
 		assert(g->leader != NULL);
 		failure(req, rc, sqlite3_errmsg(g->leader->conn));
@@ -548,6 +554,41 @@ done:
 	}
 }
 
+#ifdef DQLITE_NEXT
+
+static void qb_top(pool_work_t *w)
+{
+	struct handle *req = CONTAINER_OF(w, struct handle, work);
+	query_batch_async(req, POOL_TOP_HALF);
+}
+
+static void qb_bottom(pool_work_t *w)
+{
+	struct handle *req = CONTAINER_OF(w, struct handle, work);
+	query_batch_async(req, POOL_BOTTOM_HALF);
+}
+
+#endif
+
+static void query_batch(struct gateway *g)
+{
+	struct handle *req = g->req;
+	assert(req != NULL);
+	g->req = NULL;
+	req->gw = g;
+
+#ifdef DQLITE_NEXT
+	struct dqlite_node *node = g->raft->data;
+	pool_t *pool = !!(pool_ut_fallback()->flags & POOL_FOR_UT)
+		? pool_ut_fallback() : &node->pool;
+	pool_queue_work(pool, &req->work, g->leader->db->cookie,
+			WT_UNORD, qb_top, qb_bottom);
+#else
+	query_batch_async(req, POOL_TOP_HALF);
+	query_batch_async(req, POOL_BOTTOM_HALF);
+#endif
+}
+
 static void query_barrier_cb(struct barrier *barrier, int status)
 {
 	tracef("query barrier cb status:%d", status);
@@ -591,7 +632,7 @@ static int handle_query(struct gateway *
 	tracef("handle query schema:%" PRIu8, req->schema);
 	struct cursor *cursor = &req->cursor;
 	struct stmt *stmt;
-	struct request_query request = {0};
+	struct request_query request = { 0 };
 	int tuple_format;
 	bool is_readonly;
 	uint64_t req_id;
@@ -692,7 +733,7 @@ static void handle_exec_sql_next(struct
 {
 	tracef("handle exec sql next");
 	struct cursor *cursor = &req->cursor;
-	struct response_result response = {0};
+	struct response_result response = { 0 };
 	sqlite3_stmt *stmt = NULL;
 	const char *tail;
 	int tuple_format;
@@ -781,7 +822,7 @@ static int handle_exec_sql(struct gatewa
 {
 	tracef("handle exec sql schema:%" PRIu8, req->schema);
 	struct cursor *cursor = &req->cursor;
-	struct request_exec_sql request = {0};
+	struct request_exec_sql request = { 0 };
 	int rc;
 
 	/* Fail early if the schema version isn't recognized, even though we
@@ -916,7 +957,7 @@ static int handle_query_sql(struct gatew
 {
 	tracef("handle query sql schema:%" PRIu8, req->schema);
 	struct cursor *cursor = &req->cursor;
-	struct request_query_sql request = {0};
+	struct request_query_sql request = { 0 };
 	int rv;
 
 	/* Fail early if the schema version isn't recognized. */
@@ -960,8 +1001,7 @@ static int handle_interrupt(struct gatew
 	return 0;
 }
 
-struct change
-{
+struct change {
 	struct gateway *gateway;
 	struct raft_change req;
 };
@@ -973,7 +1013,7 @@ static void raftChangeCb(struct raft_cha
 	struct change *r = change->data;
 	struct gateway *g = r->gateway;
 	struct handle *req = g->req;
-	struct response_empty response = {0};
+	struct response_empty response = { 0 };
 	g->req = NULL;
 	sqlite3_free(r);
 	if (status != 0) {
@@ -1146,7 +1186,7 @@ static int handle_dump(struct gateway *g
 	bool err = true;
 	sqlite3_vfs *vfs;
 	char *cur;
-	char filename[1024] = {0};
+	char filename[1024] = { 0 };
 	void *data;
 	size_t n;
 	uint8_t *page;
@@ -1312,7 +1352,7 @@ void raftTransferCb(struct raft_transfer
 {
 	struct gateway *g = r->data;
 	struct handle *req = g->req;
-	struct response_empty response = {0};
+	struct response_empty response = { 0 };
 	g->req = NULL;
 	sqlite3_free(r);
 	if (g->raft->state == RAFT_LEADER) {
@@ -1425,6 +1465,7 @@ handle:
 	req->sql = NULL;
 	req->stmt = stmt;
 	req->exec_count = 0;
+	req->work = (pool_work_t){};
 
 	switch (type) {
 #define DISPATCH(LOWER, UPPER, _)            \
@@ -1451,6 +1492,8 @@ int gateway__resume(struct gateway *g, b
 	}
 	tracef("gateway resume - not finished");
 	*finished = false;
+
+	g->req->work = (pool_work_t){};
 	query_batch(g);
 	return 0;
 }
diff -pruN 1.16.4-4/src/gateway.h 1.16.5-1/src/gateway.h
--- 1.16.4-4/src/gateway.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/gateway.h	2024-06-28 13:17:26.000000000 +0000
@@ -23,8 +23,7 @@ struct handle;
  * Handle requests from a single connected client and forward them to
  * SQLite.
  */
-struct gateway
-{
+struct gateway {
 	struct config *config;       /* Configuration */
 	struct registry *registry;   /* Register of existing databases */
 	struct raft *raft;           /* Raft instance */
@@ -62,8 +61,7 @@ typedef void (*handle_cb)(struct handle
 			  int status,
 			  uint8_t type,
 			  uint8_t schema);
-struct handle
-{
+struct handle {
 	/* User data. */
 	void *data;
 	/* Type code for this request. */
@@ -106,6 +104,10 @@ struct handle
 	/* Callback that will be invoked at the end of request processing to
 	 * write the response. */
 	handle_cb cb;
+	/* A link into thread pool's queues. */
+	pool_work_t work;
+	/* Gateway the handle belongs to. */
+	struct gateway *gw;
 };
 
 /**
diff -pruN 1.16.4-4/src/leader.c 1.16.5-1/src/leader.c
--- 1.16.4-4/src/leader.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/leader.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,3 +1,4 @@
+#include <stdint.h>
 #include <stdio.h>
 
 #include "../include/dqlite.h"
@@ -5,9 +6,14 @@
 #include "./lib/assert.h"
 
 #include "command.h"
+#include "conn.h"
+#include "gateway.h"
 #include "id.h"
 #include "leader.h"
+#include "lib/threadpool.h"
+#include "server.h"
 #include "tracing.h"
+#include "utils.h"
 #include "vfs.h"
 
 /* Called when a leader exec request terminates and the associated callback can
@@ -136,7 +142,7 @@ int leader__init(struct leader *l, struc
 
 	l->exec = NULL;
 	l->inflight = NULL;
-	QUEUE__PUSH(&db->leaders, &l->queue);
+	queue_insert_tail(&db->leaders, &l->queue);
 	return 0;
 }
 
@@ -153,7 +159,7 @@ void leader__close(struct leader *l)
 	rc = sqlite3_close(l->conn);
 	assert(rc == 0);
 
-	QUEUE__REMOVE(&l->queue);
+	queue_remove(&l->queue);
 }
 
 /* A checkpoint command that fails to commit is not a huge issue.
@@ -218,7 +224,11 @@ static void leaderMaybeCheckpointLegacy(
 		tracef("raft_malloc - no mem");
 		goto err_after_buf_alloc;
 	}
+#ifdef USE_SYSTEM_RAFT
 	rv = raft_apply(l->raft, apply, &buf, 1, leaderCheckpointApplyCb);
+#else
+	rv = raft_apply(l->raft, apply, &buf, NULL, 1, leaderCheckpointApplyCb);
+#endif
 	if (rv != 0) {
 		tracef("raft_apply failed %d", rv);
 		raft_free(apply);
@@ -326,7 +336,13 @@ static int leaderApplyFrames(struct exec
 	apply->type = COMMAND_FRAMES;
 	idSet(apply->req.req_id, req->id);
 
+#ifdef USE_SYSTEM_RAFT
 	rv = raft_apply(l->raft, &apply->req, &buf, 1, leaderApplyFramesCb);
+#else
+	/* TODO actual WAL slice goes here */
+	struct raft_entry_local_data local_data = {};
+	rv = raft_apply(l->raft, &apply->req, &buf, &local_data, 1, leaderApplyFramesCb);
+#endif
 	if (rv != 0) {
 		tracef("raft apply failed %d", rv);
 		goto err_after_command_encode;
@@ -346,7 +362,7 @@ err:
 	return rv;
 }
 
-static void leaderExecV2(struct exec *req)
+static void leaderExecV2(struct exec *req, enum pool_half half)
 {
 	tracef("leader exec v2 id:%" PRIu64, req->id);
 	struct leader *l = req->leader;
@@ -358,7 +374,10 @@ static void leaderExecV2(struct exec *re
 	unsigned i;
 	int rv;
 
-	req->status = sqlite3_step(req->stmt);
+	if (half == POOL_TOP_HALF) {
+		req->status = sqlite3_step(req->stmt);
+		return;
+	} /* else POOL_BOTTOM_HALF => */
 
 	rv = VfsPoll(vfs, db->path, &frames, &n);
 	if (rv != 0 || n == 0) {
@@ -398,17 +417,44 @@ finish:
 	leaderExecDone(l->exec);
 }
 
+#ifdef DQLITE_NEXT
+
+static void exec_top(pool_work_t *w)
+{
+	struct exec *req = CONTAINER_OF(w, struct exec, work);
+	leaderExecV2(req, POOL_TOP_HALF);
+}
+
+static void exec_bottom(pool_work_t *w)
+{
+	struct exec *req = CONTAINER_OF(w, struct exec, work);
+	leaderExecV2(req, POOL_BOTTOM_HALF);
+}
+
+#endif
+
 static void execBarrierCb(struct barrier *barrier, int status)
 {
 	tracef("exec barrier cb status %d", status);
 	struct exec *req = barrier->data;
 	struct leader *l = req->leader;
+
 	if (status != 0) {
 		l->exec->status = status;
 		leaderExecDone(l->exec);
 		return;
 	}
-	leaderExecV2(req);
+
+#ifdef DQLITE_NEXT
+	struct dqlite_node *node = l->raft->data;
+	pool_t *pool = !!(pool_ut_fallback()->flags & POOL_FOR_UT)
+		? pool_ut_fallback() : &node->pool;
+	pool_queue_work(pool, &req->work, l->db->cookie,
+			WT_UNORD, exec_top, exec_bottom);
+#else
+	leaderExecV2(req, POOL_TOP_HALF);
+	leaderExecV2(req, POOL_BOTTOM_HALF);
+#endif
 }
 
 int leader__exec(struct leader *l,
@@ -431,6 +477,7 @@ int leader__exec(struct leader *l,
 	req->cb = cb;
 	req->barrier.data = req;
 	req->barrier.cb = NULL;
+	req->work = (pool_work_t){};
 
 	rv = leader__barrier(l, &req->barrier, execBarrierCb);
 	if (rv != 0) {
diff -pruN 1.16.4-4/src/leader.h 1.16.5-1/src/leader.h
--- 1.16.4-4/src/leader.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/leader.h	2024-06-28 13:17:26.000000000 +0000
@@ -10,6 +10,7 @@
 
 #include "./lib/queue.h"
 #include "db.h"
+#include "lib/threadpool.h"
 #include "raft.h"
 
 #define SQLITE_IOERR_NOT_LEADER (SQLITE_IOERR | (40 << 8))
@@ -23,22 +24,19 @@ typedef void (*exec_cb)(struct exec *req
 typedef void (*barrier_cb)(struct barrier *req, int status);
 
 /* Wrapper around raft_apply, saving context information. */
-struct apply
-{
+struct apply {
 	struct raft_apply req; /* Raft apply request */
 	int status;            /* Raft apply result */
 	struct leader *leader; /* Leader connection that triggered the hook */
 	int type;              /* Command type */
 	union {                /* Command-specific data */
-		struct
-		{
+		struct {
 			bool is_commit;
 		} frames;
 	};
 };
 
-struct leader
-{
+struct leader {
 	struct db *db;          /* Database the connection. */
 	sqlite3 *conn;          /* Underlying SQLite connection. */
 	struct raft *raft;      /* Raft instance. */
@@ -47,8 +45,7 @@ struct leader
 	struct apply *inflight; /* TODO: make leader__close async */
 };
 
-struct barrier
-{
+struct barrier {
 	void *data;
 	struct leader *leader;
 	struct raft_barrier req;
@@ -58,8 +55,7 @@ struct barrier
 /**
  * Asynchronous request to execute a statement.
  */
-struct exec
-{
+struct exec {
 	void *data;
 	struct leader *leader;
 	struct barrier barrier;
@@ -68,6 +64,7 @@ struct exec
 	int status;
 	queue queue;
 	exec_cb cb;
+	pool_work_t work;
 };
 
 /**
diff -pruN 1.16.4-4/src/lib/byte.h 1.16.5-1/src/lib/byte.h
--- 1.16.4-4/src/lib/byte.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/byte.h	2024-06-28 13:17:26.000000000 +0000
@@ -115,6 +115,18 @@ DQLITE_INLINE uint32_t ByteGetBe32(const
 	return w | x | y | z;
 }
 
+DQLITE_INLINE uint32_t ByteGetLe32(const uint8_t *buf)
+{
+	uint32_t w = buf[0];
+	uint32_t x = buf[1];
+	uint32_t y = buf[2];
+	uint32_t z = buf[3];
+	z <<= 24;
+	y <<= 16;
+	x <<= 8;
+	return w | x | y | z;
+}
+
 DQLITE_INLINE void BytePutBe32(uint32_t v, uint8_t *buf)
 {
 	buf[0] = (uint8_t)(v >> 24);
diff -pruN 1.16.4-4/src/lib/queue.h 1.16.5-1/src/lib/queue.h
--- 1.16.4-4/src/lib/queue.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/queue.h	2024-06-28 13:17:26.000000000 +0000
@@ -1,106 +1,95 @@
 #ifndef LIB_QUEUE_H_
 #define LIB_QUEUE_H_
 
-#include <stddef.h>
+#include <stddef.h> /* offsetof */
 
-typedef void *queue[2];
-
-/* Private macros. */
-#define QUEUE__NEXT(q) (*(queue **)&((*(q))[0]))
-#define QUEUE__PREV(q) (*(queue **)&((*(q))[1]))
-
-#define QUEUE__PREV_NEXT(q) (QUEUE__NEXT(QUEUE__PREV(q)))
-#define QUEUE__NEXT_PREV(q) (QUEUE__PREV(QUEUE__NEXT(q)))
-
-/**
- * Initialize an empty queue.
- */
-#define QUEUE__INIT(q)                \
-	{                             \
-		QUEUE__NEXT(q) = (q); \
-		QUEUE__PREV(q) = (q); \
-	}
-
-/**
- * Return true if the queue has no element.
- */
-#define QUEUE__IS_EMPTY(q) ((const queue *)(q) == (const queue *)QUEUE__NEXT(q))
-
-/**
- * Insert an element at the back of a queue.
- */
-#define QUEUE__PUSH(q, e)                        \
-	{                                        \
-		QUEUE__NEXT(e) = (q);            \
-		QUEUE__PREV(e) = QUEUE__PREV(q); \
-		QUEUE__PREV_NEXT(e) = (e);       \
-		QUEUE__PREV(q) = (e);            \
-	}
-
-#define QUEUE__INSERT_TAIL(q, e) QUEUE__PUSH(q, e)
-
-/**
- * Insert an element at the front of a queue.
- */
-#define QUEUE__INSERT_HEAD(h, q)                 \
-	{                                        \
-		QUEUE__NEXT(q) = QUEUE__NEXT(h); \
-		QUEUE__PREV(q) = (h);            \
-		QUEUE__NEXT_PREV(q) = (q);       \
-		QUEUE__NEXT(h) = (q);            \
-	}
-
-/**
- * Remove the given element from the queue. Any element can be removed at any
- * time.
- */
-#define QUEUE__REMOVE(e)                              \
-	{                                             \
-		QUEUE__PREV_NEXT(e) = QUEUE__NEXT(e); \
-		QUEUE__NEXT_PREV(e) = QUEUE__PREV(e); \
-	}
-
-/**
- * Moves elements from queue @h to queue @n
- * Note: Removed QUEUE__SPLIT() and merged it into QUEUE__MOVE().
- */
-#define QUEUE__MOVE(h, n)                                  \
-	{                                                  \
-		if (QUEUE__IS_EMPTY(h)) {                  \
-			QUEUE__INIT(n);                    \
-		} else {                                   \
-			queue *__q = QUEUE__HEAD(h);       \
-			QUEUE__PREV(n) = QUEUE__PREV(h);   \
-			QUEUE__PREV_NEXT(n) = (n);         \
-			QUEUE__NEXT(n) = (__q);            \
-			QUEUE__PREV(h) = QUEUE__PREV(__q); \
-			QUEUE__PREV_NEXT(h) = (h);         \
-			QUEUE__PREV(__q) = (n);            \
-		}                                          \
-	}
-
-/**
- * Return the element at the front of the queue.
- */
-#define QUEUE__HEAD(q) (QUEUE__NEXT(q))
-
-/**
- * Return the element at the back of the queue.
- */
-#define QUEUE__TAIL(q) (QUEUE__PREV(q))
-
-/**
- * Iternate over the element of a queue.
- *
- * Mutating the queue while iterating results in undefined behavior.
- */
-#define QUEUE__FOREACH(q, e) \
-	for ((q) = QUEUE__NEXT(e); (q) != (e); (q) = QUEUE__NEXT(q))
-
-/**
- * Return the structure holding the given element.
- */
-#define QUEUE__DATA(e, type, field) \
-	((type *)(uintptr_t)((char *)(e)-offsetof(type, field)))
+struct queue
+{
+	struct queue *next;
+	struct queue *prev;
+};
+
+typedef struct queue queue;
+
+#define QUEUE_DATA(e, type, field) \
+	((type *)((void *)((char *)(e)-offsetof(type, field))))
+
+#define QUEUE_FOREACH(q, h) for ((q) = (h)->next; (q) != (h); (q) = (q)->next)
+
+static inline void queue_init(struct queue *q)
+{
+	q->next = q;
+	q->prev = q;
+}
+
+static inline int queue_empty(const struct queue *q)
+{
+	return q == q->next;
+}
+
+static inline struct queue *queue_head(const struct queue *q)
+{
+	return q->next;
+}
+
+static inline struct queue *queue_next(const struct queue *q)
+{
+	return q->next;
+}
+
+static inline struct queue *queue_tail(const struct queue *q)
+{
+	return q->prev;
+}
+
+static inline void queue_add(struct queue *h, struct queue *n)
+{
+	h->prev->next = n->next;
+	n->next->prev = h->prev;
+	h->prev = n->prev;
+	h->prev->next = h;
+}
+
+static inline void queue_split(struct queue *h,
+			       struct queue *q,
+			       struct queue *n)
+{
+	n->prev = h->prev;
+	n->prev->next = n;
+	n->next = q;
+	h->prev = q->prev;
+	h->prev->next = h;
+	q->prev = n;
+}
+
+static inline void queue_move(struct queue *h, struct queue *n)
+{
+	if (queue_empty(h))
+		queue_init(n);
+	else
+		queue_split(h, h->next, n);
+}
+
+static inline void queue_insert_head(struct queue *h, struct queue *q)
+{
+	q->next = h->next;
+	q->prev = h;
+	q->next->prev = q;
+	h->next = q;
+}
+
+static inline void queue_insert_tail(struct queue *h, struct queue *q)
+{
+	q->next = h;
+	q->prev = h->prev;
+	q->prev->next = q;
+	h->prev = q;
+}
+
+static inline void queue_remove(struct queue *q)
+{
+	q->prev->next = q->next;
+	q->next->prev = q->prev;
+}
 
 #endif /* LIB_QUEUE_H_*/
diff -pruN 1.16.4-4/src/lib/sm.c 1.16.5-1/src/lib/sm.c
--- 1.16.4-4/src/lib/sm.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/sm.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,5 +1,7 @@
 #include "sm.h"
 #include <stddef.h> /* NULL */
+#include <stdio.h> /* fprintf */
+#include "../tracing.h"
 #include "../utils.h"
 
 
@@ -40,6 +42,8 @@ void sm_move(struct sm *m, int next_stat
 {
 	int prev = sm_state(m);
 
+	tracef("SM_MOVE %s => %s", m->conf[prev].name, m->conf[next_state].name);
+
 	PRE(sm_is_locked(m));
 	PRE(m->conf[sm_state(m)].allowed & BITS(next_state));
 	m->state = next_state;
@@ -58,3 +62,17 @@ void sm_fail(struct sm *m, int fail_stat
 	m->state = fail_state;
 	POST(m->invariant != NULL && m->invariant(m, prev));
 }
+
+static __attribute__((noinline)) bool check_failed(const char *f, int n, const char *s)
+{
+	fprintf(stderr, "%s:%d check failed: %s\n", f, n, s);
+	return false;
+}
+
+bool sm_check(bool b, const char *f, int n, const char *s)
+{
+	if (!b) {
+		return check_failed(f, n, s);
+	}
+	return true;
+}
diff -pruN 1.16.4-4/src/lib/sm.h 1.16.5-1/src/lib/sm.h
--- 1.16.4-4/src/lib/sm.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/sm.h	2024-06-28 13:17:26.000000000 +0000
@@ -6,6 +6,8 @@
 
 #define BITS(state) (1ULL << (state))
 
+#define CHECK(cond) sm_check((cond), __FILE__, __LINE__, #cond)
+
 enum {
 	SM_PREV_NONE = -1,
 	/* sizeof(sm_conf::allowed * 8) */
@@ -42,5 +44,6 @@ void sm_fini(struct sm *m);
 void sm_move(struct sm *m, int next_state);
 void sm_fail(struct sm *m, int fail_state, int rc);
 int sm_state(const struct sm *m);
+bool sm_check(bool b, const char *f, int n, const char *s);
 
 #endif /* __LIB_SM__ */
diff -pruN 1.16.4-4/src/lib/threadpool.c 1.16.5-1/src/lib/threadpool.c
--- 1.16.4-4/src/lib/threadpool.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/threadpool.c	2024-06-28 13:17:26.000000000 +0000
@@ -7,6 +7,7 @@
 #include "../../src/lib/queue.h"
 #include "../../src/lib/sm.h"
 #include "../../src/utils.h"
+#include "../tracing.h"
 
 /**
  *  Planner thread state machine.
@@ -53,14 +54,14 @@ static const struct sm_conf planner_stat
 	[PS_DRAINING] = {
 	    .name = "draining",
 	    .allowed = BITS(PS_DRAINING)
-	             | BITS(PS_NOTHING)
-	             | BITS(PS_BARRIER),
+		     | BITS(PS_NOTHING)
+		     | BITS(PS_BARRIER),
 	},
 	[PS_BARRIER] = {
 	    .name = "barrier",
 	    .allowed = BITS(PS_DRAINING_UNORD)
-	             | BITS(PS_DRAINING)
-	             | BITS(PS_BARRIER),
+		     | BITS(PS_DRAINING)
+		     | BITS(PS_BARRIER),
 	},
 	[PS_DRAINING_UNORD] = {
 	    .name = "draining-unord",
@@ -94,33 +95,40 @@ struct pool_thread {
 	struct targs arg;
 };
 
+/* clang-format off */
 struct pool_impl {
-	uv_mutex_t mutex; /* Input queue, planner_sm,
-			     worker and planner threads lock */
-	uint32_t threads_nr;
+	uv_mutex_t     mutex;          /* Input queue, planner_sm,
+					  worker and planner threads lock */
+	uint32_t       threads_nr;
 	pool_thread_t *threads;
 
-	queue outq;            /* Output queue used by libuv part */
-	uv_mutex_t outq_mutex; /* Output queue lock */
-	uv_async_t outq_async; /* Signalled when output queue is not
-				  empty and libuv loop has to process
-				  items from it */
-	uint64_t active_ws;    /* Number of all work items in flight,
-				  accessed from the main thread only */
-
-	queue ordered;        /* Queue of WT_ORD{N} items */
-	queue unordered;      /* Queue of WT_UNORD items */
-	struct sm planner_sm; /* State machine of the scheduler */
-	uv_cond_t planner_cond;
-	uv_thread_t planner_thread; /* Scheduler's thread */
-
-	uint32_t ord_in_flight; /* Number of WT_ORD{N} in flight */
-	bool exiting;           /* True when the pool is being stopped */
-	enum pool_work_type     /* Type of the previous work item, */
-	    ord_prev;           /* used in WT_ORD{N} ivariants */
-	uint32_t qos;           /* QoS token */
-	uint32_t qos_prio;      /* QoS prio */
+	queue          outq;           /* Output queue used by libuv part */
+	uv_mutex_t     outq_mutex;     /* Output queue lock */
+	uv_async_t     outq_async;     /* Signalled when output queue is not
+				          empty and libuv loop has to process
+				          items from it */
+	uint64_t       active_ws;      /* Number of all work items in flight,
+				          accessed from the main thread only */
+
+	queue          ordered;        /* Queue of WT_ORD{N} items */
+	queue          unordered;      /* Queue of WT_UNORD items */
+	struct sm      planner_sm;     /* State machine of the scheduler */
+	uv_cond_t      planner_cond;
+	uv_thread_t    planner_thread; /* Scheduler's thread */
+
+	uint32_t       ord_in_flight;  /* Number of WT_ORD{N} in flight */
+	bool           exiting;        /* True when the pool is being stopped */
+	enum pool_work_type            /* Type of the previous work item, */
+	               ord_prev;       /* used in WT_ORD{N} ivariants */
+	uint32_t       qos;            /* QoS token */
+	uint32_t       qos_prio;       /* QoS prio */
 };
+/* clang-format on */
+
+static inline bool pool_is_inited(const pool_t *pool)
+{
+	return pool->pi != NULL;
+}
 
 static inline bool has_active_ws(pool_t *pool)
 {
@@ -143,25 +151,25 @@ static inline void w_unregister(pool_t *
 
 static bool empty(const queue *q)
 {
-	return QUEUE__IS_EMPTY(q);
+	return queue_empty(q);
 }
 
 static queue *head(const queue *q)
 {
-	return QUEUE__HEAD(q);
+	return queue_head(q);
 }
 
 static void push(queue *to, queue *what)
 {
-	QUEUE__INSERT_TAIL(to, what);
+	queue_insert_tail(to, what);
 }
 
 static queue *pop(queue *from)
 {
-	queue *q = QUEUE__HEAD(from);
+	queue *q = queue_head(from);
 	PRE(q != NULL);
-	QUEUE__REMOVE(q);
-	QUEUE__INIT(q);
+	queue_remove(q);
+	queue_init(q);
 	return q;
 }
 
@@ -180,7 +188,7 @@ static queue *qos_pop(pool_impl_t *pi, q
 
 static pool_work_t *q_to_w(const queue *q)
 {
-	return QUEUE__DATA(q, pool_work_t, link);
+	return QUEUE_DATA(q, pool_work_t, link);
 }
 
 static enum pool_work_type q_type(const queue *q)
@@ -199,19 +207,21 @@ static bool planner_invariant(const stru
 	queue *o = &pi->ordered;
 	queue *u = &pi->unordered;
 
+	/* clang-format off */
 	return ERGO(sm_state(m) == PS_NOTHING, empty(o) && empty(u)) &&
-	       ERGO(sm_state(m) == PS_DRAINING,
-		    ERGO(prev_state == PS_BARRIER,
-			 pi->ord_in_flight == 0 && empty(u)) &&
-			ERGO(prev_state == PS_NOTHING,
-			     !empty(u) || !empty(o))) &&
-	       ERGO(sm_state(m) == PS_EXITED,
-		    pi->exiting && empty(o) && empty(u)) &&
-	       ERGO(
-		   sm_state(m) == PS_BARRIER,
-		   ERGO(prev_state == PS_DRAINING, q_type(head(o)) == WT_BAR) &&
-		       ERGO(prev_state == PS_DRAINING_UNORD, empty(u))) &&
-	       ERGO(sm_state(m) == PS_DRAINING_UNORD, !empty(u));
+		ERGO(sm_state(m) == PS_DRAINING,
+		     ERGO(prev_state == PS_BARRIER,
+			  pi->ord_in_flight == 0 && empty(u)) &&
+		     ERGO(prev_state == PS_NOTHING,
+			  !empty(u) || !empty(o))) &&
+		ERGO(sm_state(m) == PS_EXITED,
+		     pi->exiting && empty(o) && empty(u)) &&
+		ERGO(sm_state(m) == PS_BARRIER,
+		     ERGO(prev_state == PS_DRAINING,
+			  q_type(head(o)) == WT_BAR) &&
+		     ERGO(prev_state == PS_DRAINING_UNORD, empty(u))) &&
+		ERGO(sm_state(m) == PS_DRAINING_UNORD, !empty(u));
+	/* clang-format on */
 }
 
 static void planner(void *arg)
@@ -237,7 +247,9 @@ static void planner(void *arg)
 					uv_cond_wait(&pi->planner_cond, mutex);
 				}
 				sm_move(planner_sm,
-					pi->exiting ? PS_EXITED : PS_DRAINING);
+					pi->exiting && empty(o) && empty(u)
+					    ? PS_EXITED
+					    : PS_DRAINING);
 				break;
 			case PS_DRAINING:
 				while (!(empty(o) && empty(u))) {
@@ -356,7 +368,6 @@ static void pool_cleanup(pool_t *pool)
 		return;
 	}
 
-	pi->exiting = true;
 	uv_cond_signal(&pi->planner_cond);
 
 	if (uv_thread_join(&pi->planner_thread)) {
@@ -413,7 +424,7 @@ static void pool_threads_init(pool_t *po
 			.idx = i,
 		};
 
-		QUEUE__INIT(&ts[i].inq);
+		queue_init(&ts[i].inq);
 		if (uv_cond_init(&ts[i].cond)) {
 			abort();
 		}
@@ -451,6 +462,7 @@ static void pool_work_submit(pool_t *poo
 	}
 
 	uv_mutex_lock(&pi->mutex);
+	POST(!pi->exiting);
 	push(w->type == WT_UNORD ? u : o, &w->link);
 	uv_cond_signal(&pi->planner_cond);
 	uv_mutex_unlock(&pi->mutex);
@@ -462,7 +474,7 @@ void work_done(uv_async_t *handle)
 	pool_impl_t *pi = CONTAINER_OF(handle, pool_impl_t, outq_async);
 
 	uv_mutex_lock(&pi->outq_mutex);
-	QUEUE__MOVE(&pi->outq, &q);
+	queue_move(&pi->outq, &q);
 	uv_mutex_unlock(&pi->outq_mutex);
 
 	while (!empty(&q)) {
@@ -477,8 +489,16 @@ void pool_queue_work(pool_t *pool,
 		     void (*work_cb)(pool_work_t *w),
 		     void (*after_work_cb)(pool_work_t *w))
 {
+	PRE(memcmp(w, &(pool_work_t){}, sizeof *w) == 0);
 	PRE(work_cb != NULL && type < WT_NR);
 
+	if (!!(pool->flags & POOL_FOR_UT_NOT_ASYNC)) {
+		work_cb(w);
+		after_work_cb(w);
+		return;
+	}
+
+	PRE(pool_is_inited(pool));
 	*w = (pool_work_t){
 		.pool = pool,
 		.type = type,
@@ -486,6 +506,7 @@ void pool_queue_work(pool_t *pool,
 		.work_cb = work_cb,
 		.after_work_cb = after_work_cb,
 	};
+
 	w_register(pool, w);
 	pool_work_submit(pool, w);
 }
@@ -500,6 +521,7 @@ int pool_init(pool_t *pool,
 
 	PRE(threads_nr <= THREADPOOL_SIZE_MAX);
 
+	pool->flags = 0x0;
 	pi = pool->pi = calloc(1, sizeof(*pool->pi));
 	if (pi == NULL) {
 		return UV_ENOMEM;
@@ -513,9 +535,9 @@ int pool_init(pool_t *pool,
 		.threads_nr = threads_nr,
 		.ord_in_flight = 0,
 	};
-	QUEUE__INIT(&pi->outq);
-	QUEUE__INIT(&pi->ordered);
-	QUEUE__INIT(&pi->unordered);
+	queue_init(&pi->outq);
+	queue_init(&pi->ordered);
+	queue_init(&pi->unordered);
 
 	rc = uv_mutex_init(&pi->outq_mutex);
 	if (rc != 0) {
@@ -541,7 +563,8 @@ void pool_fini(pool_t *pool)
 	pool_cleanup(pool);
 
 	uv_mutex_lock(&pi->outq_mutex);
-	POST(empty(&pi->outq) && !has_active_ws(pool));
+	POST(!!(pool->flags & POOL_FOR_UT_NON_CLEAN_FINI) ||
+	     (empty(&pi->outq) && !has_active_ws(pool)));
 	uv_mutex_unlock(&pi->outq_mutex);
 
 	uv_mutex_destroy(&pi->outq_mutex);
@@ -550,5 +573,16 @@ void pool_fini(pool_t *pool)
 
 void pool_close(pool_t *pool)
 {
-	uv_close((uv_handle_t *)&pool->pi->outq_async, NULL);
+	pool_impl_t *pi = pool->pi;
+
+	uv_close((uv_handle_t *)&pi->outq_async, NULL);
+	uv_mutex_lock(&pi->mutex);
+	pi->exiting = true;
+	uv_mutex_unlock(&pi->mutex);
+}
+
+pool_t *pool_ut_fallback(void)
+{
+	static pool_t pool;
+	return &pool;
 }
diff -pruN 1.16.4-4/src/lib/threadpool.h 1.16.5-1/src/lib/threadpool.h
--- 1.16.4-4/src/lib/threadpool.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/lib/threadpool.h	2024-06-28 13:17:26.000000000 +0000
@@ -56,26 +56,49 @@ enum pool_work_type {
 	WT_NR,
 };
 
-struct pool_work_s
-{
+struct pool_work_s {
 	queue link;         /* Link into ordered, unordered and outq */
 	uint32_t thread_id; /* Identifier of the thread the item is affined */
 	pool_t *pool;       /* The pool, item is being associated with */
 	enum pool_work_type type;
-
+	int rc; /* Return code used to deliver pool work operation result to the
+		 * uv_loop's thread. */
 	void (*work_cb)(pool_work_t *w);
 	void (*after_work_cb)(pool_work_t *w);
 };
 
-struct pool_s
-{
+struct pool_s {
 	struct pool_impl *pi;
+	int flags;
 };
 
 enum {
 	POOL_QOS_PRIO_FAIR = 2,
 };
 
+enum pool_half {
+	POOL_TOP_HALF = 0x109,
+	POOL_BOTTOM_HALF = 0xb01103,
+};
+
+enum {
+	/**
+	 * Setting POOL_FOR_UT_NON_CLEAN_FINI relaxes pool's invariant during
+	 * the finalization w.r.t. to pass a few tests checking failures with
+	 * non-clean unit-test termination.
+	 */
+	POOL_FOR_UT_NON_CLEAN_FINI = 1u << 0,
+	/**
+	 * Set this flag if there's no event loop in unit test. Top- and
+	 * bottom- halves will be called in the current thread.
+	 */
+	POOL_FOR_UT_NOT_ASYNC = 1u << 1,
+	/**
+	 * Set if the pool runs in the context of unit test.
+	 */
+	POOL_FOR_UT = 1u << 2,
+};
+
 int pool_init(pool_t *pool,
 	      uv_loop_t *loop,
 	      uint32_t threads_nr,
@@ -89,4 +112,6 @@ void pool_queue_work(pool_t *pool,
 		     void (*work_cb)(pool_work_t *w),
 		     void (*after_work_cb)(pool_work_t *w));
 
+pool_t *pool_ut_fallback(void);
+
 #endif /* __THREAD_POOL__ */
diff -pruN 1.16.4-4/src/raft/client.c 1.16.5-1/src/raft/client.c
--- 1.16.4-4/src/raft/client.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/client.c	2024-06-28 13:17:26.000000000 +0000
@@ -7,13 +7,14 @@
 #include "log.h"
 #include "membership.h"
 #include "progress.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "replication.h"
 #include "request.h"
 
 int raft_apply(struct raft *r,
 	       struct raft_apply *req,
 	       const struct raft_buffer bufs[],
+	       const struct raft_entry_local_data local_data[],
 	       const unsigned n,
 	       raft_apply_cb cb)
 {
@@ -41,7 +42,7 @@ int raft_apply(struct raft *r,
 	req->cb = cb;
 
 	/* Append the new entries to the log. */
-	rv = logAppendCommands(r->log, r->current_term, bufs, n);
+	rv = logAppendCommands(r->log, r->current_term, bufs, local_data, n);
 	if (rv != 0) {
 		goto err;
 	}
@@ -57,7 +58,7 @@ int raft_apply(struct raft *r,
 
 err_after_log_append:
 	logDiscard(r->log, index);
-	QUEUE_REMOVE(&req->queue);
+	queue_remove(&req->queue);
 err:
 	assert(rv != 0);
 	return rv;
@@ -90,7 +91,7 @@ int raft_barrier(struct raft *r, struct
 	req->index = index;
 	req->cb = cb;
 
-	rv = logAppend(r->log, r->current_term, RAFT_BARRIER, &buf, NULL);
+	rv = logAppend(r->log, r->current_term, RAFT_BARRIER, buf, (struct raft_entry_local_data){}, true, NULL);
 	if (rv != 0) {
 		goto err_after_buf_alloc;
 	}
@@ -106,7 +107,7 @@ int raft_barrier(struct raft *r, struct
 
 err_after_log_append:
 	logDiscard(r->log, index);
-	QUEUE_REMOVE(&req->queue);
+	queue_remove(&req->queue);
 err_after_buf_alloc:
 	raft_free(buf.base);
 err:
diff -pruN 1.16.4-4/src/raft/convert.c 1.16.5-1/src/raft/convert.c
--- 1.16.4-4/src/raft/convert.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/convert.c	2024-06-28 13:17:26.000000000 +0000
@@ -9,7 +9,7 @@
 #include "log.h"
 #include "membership.h"
 #include "progress.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "replication.h"
 #include "request.h"
 
@@ -93,11 +93,11 @@ static void convertClearLeader(struct ra
 	}
 
 	/* Fail all outstanding requests */
-	while (!QUEUE_IS_EMPTY(&r->leader_state.requests)) {
+	while (!queue_empty(&r->leader_state.requests)) {
 		struct request *req;
 		queue *head;
-		head = QUEUE_HEAD(&r->leader_state.requests);
-		QUEUE_REMOVE(head);
+		head = queue_head(&r->leader_state.requests);
+		queue_remove(head);
 		req = QUEUE_DATA(head, struct request, queue);
 		assert(req->type == RAFT_COMMAND || req->type == RAFT_BARRIER);
 		switch (req->type) {
@@ -209,7 +209,7 @@ int convertToLeader(struct raft *r)
 	r->election_timer_start = r->io->time(r->io);
 
 	/* Reset apply requests queue */
-	QUEUE_INIT(&r->leader_state.requests);
+	queue_init(&r->leader_state.requests);
 
 	/* Allocate and initialize the progress array. */
 	rv = progressBuildArray(r);
diff -pruN 1.16.4-4/src/raft/fixture.c 1.16.5-1/src/raft/fixture.c
--- 1.16.4-4/src/raft/fixture.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/fixture.c	2024-06-28 13:17:26.000000000 +0000
@@ -10,7 +10,7 @@
 #include "convert.h"
 #include "entry.h"
 #include "log.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "snapshot.h"
 
 /* Defaults */
@@ -365,7 +365,7 @@ static void ioFlushSend(struct io *io, s
 	src = &send->message;
 	dst = &transmit->message;
 
-	QUEUE_PUSH(&io->requests, &transmit->queue);
+	queue_insert_tail(&io->requests, &transmit->queue);
 
 	*dst = *src;
 	switch (dst->type) {
@@ -416,12 +416,12 @@ static void ioDestroyTransmit(struct tra
 /* Flush all requests in the queue. */
 static void ioFlushAll(struct io *io)
 {
-	while (!QUEUE_IS_EMPTY(&io->requests)) {
+	while (!queue_empty(&io->requests)) {
 		queue *head;
 		struct ioRequest *r;
 
-		head = QUEUE_HEAD(&io->requests);
-		QUEUE_REMOVE(head);
+		head = queue_head(&io->requests);
+		queue_remove(head);
 
 		r = QUEUE_DATA(head, struct ioRequest, queue);
 		switch (r->type) {
@@ -592,7 +592,7 @@ static int ioMethodAppend(struct raft_io
 
 	req->cb = cb;
 
-	QUEUE_PUSH(&io->requests, &r->queue);
+	queue_insert_tail(&io->requests, &r->queue);
 
 	return 0;
 }
@@ -660,7 +660,7 @@ static int ioMethodSnapshotPut(struct ra
 	r->completion_time = *io->time + io->disk_latency;
 	r->trailing = trailing;
 
-	QUEUE_PUSH(&io->requests, &r->queue);
+	queue_insert_tail(&io->requests, &r->queue);
 
 	return 0;
 }
@@ -680,7 +680,7 @@ static int ioMethodAsyncWork(struct raft
 	r->req->cb = cb;
 	r->completion_time = *io->time + io->work_duration;
 
-	QUEUE_PUSH(&io->requests, &r->queue);
+	queue_insert_tail(&io->requests, &r->queue);
 	return 0;
 }
 
@@ -699,7 +699,7 @@ static int ioMethodSnapshotGet(struct ra
 	r->req->cb = cb;
 	r->completion_time = *io->time + io->disk_latency;
 
-	QUEUE_PUSH(&io->requests, &r->queue);
+	queue_insert_tail(&io->requests, &r->queue);
 
 	return 0;
 }
@@ -749,7 +749,7 @@ static int ioMethodSend(struct raft_io *
 	peer = ioGetPeer(io, message->server_id);
 	r->completion_time = *io->time + peer->send_latency;
 
-	QUEUE_PUSH(&io->requests, &r->queue);
+	queue_insert_tail(&io->requests, &r->queue);
 
 	return 0;
 }
@@ -881,7 +881,7 @@ static int ioInit(struct raft_io *raft_i
 	io->snapshot = NULL;
 	io->entries = NULL;
 	io->n = 0;
-	QUEUE_INIT(&io->requests);
+	queue_init(&io->requests);
 	io->n_peers = 0;
 	io->randomized_election_timeout = ELECTION_TIMEOUT + index * 100;
 	io->network_latency = NETWORK_LATENCY;
@@ -1316,7 +1316,8 @@ static void copyLeaderLog(struct raft_fi
 		buf.base = raft_malloc(buf.len);
 		assert(buf.base != NULL);
 		memcpy(buf.base, entry->buf.base, buf.len);
-		rv = logAppend(f->log, entry->term, entry->type, &buf, NULL);
+		/* FIXME(cole) what to do here for is_local? */
+		rv = logAppend(f->log, entry->term, entry->type, buf, (struct raft_entry_local_data){}, false, NULL);
 		assert(rv == 0);
 	}
 	logRelease(raft->log, 1, entries, n);
@@ -1400,7 +1401,7 @@ static void completeRequest(struct raft_
 		}
 	}
 	assert(found);
-	QUEUE_REMOVE(head);
+	queue_remove(head);
 	switch (r->type) {
 		case APPEND:
 			ioFlushAppend(io, (struct append *)r);
diff -pruN 1.16.4-4/src/raft/lifecycle.c 1.16.5-1/src/raft/lifecycle.c
--- 1.16.4-4/src/raft/lifecycle.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/lifecycle.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,6 +1,6 @@
 #include "lifecycle.h"
 #include "../tracing.h"
-#include "queue.h"
+#include "../lib/queue.h"
 
 #include <inttypes.h>
 #include <stdlib.h>
@@ -23,7 +23,7 @@ void lifecycleRequestStart(struct raft *
 	if (reqIdIsSet(req)) {
 		tracef("request start id:%" PRIu64, extractReqId(req));
 	}
-	QUEUE_PUSH(&r->leader_state.requests, &req->queue);
+	queue_insert_tail(&r->leader_state.requests, &req->queue);
 }
 
 void lifecycleRequestEnd(struct raft *r, struct request *req)
@@ -32,5 +32,5 @@ void lifecycleRequestEnd(struct raft *r,
 	if (reqIdIsSet(req)) {
 		tracef("request end id:%" PRIu64, extractReqId(req));
 	}
-	QUEUE_REMOVE(&req->queue);
+	queue_remove(&req->queue);
 }
diff -pruN 1.16.4-4/src/raft/log.c 1.16.5-1/src/raft/log.c
--- 1.16.4-4/src/raft/log.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/log.c	2024-06-28 13:17:26.000000000 +0000
@@ -545,7 +545,9 @@ int logReinstate(struct raft_log *l,
 int logAppend(struct raft_log *l,
 	      const raft_term term,
 	      const unsigned short type,
-	      const struct raft_buffer *buf,
+	      struct raft_buffer buf,
+	      struct raft_entry_local_data local_data,
+	      bool is_local,
 	      void *batch)
 {
 	int rv;
@@ -556,7 +558,6 @@ int logAppend(struct raft_log *l,
 	assert(term > 0);
 	assert(type == RAFT_CHANGE || type == RAFT_BARRIER ||
 	       type == RAFT_COMMAND);
-	assert(buf != NULL);
 
 	rv = ensureCapacity(l);
 	if (rv != 0) {
@@ -565,7 +566,7 @@ int logAppend(struct raft_log *l,
 
 	index = logLastIndex(l) + 1;
 
-	rv = refsInit(l, term, index, *buf, batch);
+	rv = refsInit(l, term, index, buf, batch);
 	if (rv != 0) {
 		return rv;
 	}
@@ -573,8 +574,10 @@ int logAppend(struct raft_log *l,
 	entry = &l->entries[l->back];
 	entry->term = term;
 	entry->type = type;
-	entry->buf = *buf;
+	entry->buf = buf;
 	entry->batch = batch;
+	entry->local_data = local_data;
+	entry->is_local = is_local;
 
 	l->back += 1;
 	l->back = l->back % l->size;
@@ -585,6 +588,7 @@ int logAppend(struct raft_log *l,
 int logAppendCommands(struct raft_log *l,
 		      const raft_term term,
 		      const struct raft_buffer bufs[],
+		      const struct raft_entry_local_data local_data[],
 		      const unsigned n)
 {
 	unsigned i;
@@ -596,8 +600,8 @@ int logAppendCommands(struct raft_log *l
 	assert(n > 0);
 
 	for (i = 0; i < n; i++) {
-		const struct raft_buffer *buf = &bufs[i];
-		rv = logAppend(l, term, RAFT_COMMAND, buf, NULL);
+		struct raft_entry_local_data loc = (local_data != NULL) ? local_data[i] : (struct raft_entry_local_data){};
+		rv = logAppend(l, term, RAFT_COMMAND, bufs[i], loc, true, NULL);
 		if (rv != 0) {
 			return rv;
 		}
@@ -624,7 +628,7 @@ int logAppendConfiguration(struct raft_l
 	}
 
 	/* Append the new entry to the log. */
-	rv = logAppend(l, term, RAFT_CHANGE, &buf, NULL);
+	rv = logAppend(l, term, RAFT_CHANGE, buf, (struct raft_entry_local_data){}, true, NULL);
 	if (rv != 0) {
 		goto err_after_encode;
 	}
diff -pruN 1.16.4-4/src/raft/log.h 1.16.5-1/src/raft/log.h
--- 1.16.4-4/src/raft/log.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/log.h	2024-06-28 13:17:26.000000000 +0000
@@ -113,13 +113,16 @@ int logReinstate(struct raft_log *l,
 int logAppend(struct raft_log *l,
 	      raft_term term,
 	      unsigned short type,
-	      const struct raft_buffer *buf,
+	      struct raft_buffer buf,
+	      struct raft_entry_local_data local_data,
+	      bool is_local,
 	      void *batch);
 
 /* Convenience to append a series of #RAFT_COMMAND entries. */
 int logAppendCommands(struct raft_log *l,
 		      const raft_term term,
 		      const struct raft_buffer bufs[],
+		      const struct raft_entry_local_data local_data[],
 		      const unsigned n);
 
 /* Convenience to encode and append a single #RAFT_CHANGE entry. */
diff -pruN 1.16.4-4/src/raft/queue.h 1.16.5-1/src/raft/queue.h
--- 1.16.4-4/src/raft/queue.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/queue.h	1970-01-01 00:00:00.000000000 +0000
@@ -1,57 +0,0 @@
-#ifndef QUEUE_H_
-#define QUEUE_H_
-
-#include <stddef.h>
-
-typedef void *queue[2];
-
-/* Private macros. */
-#define QUEUE_NEXT(q) (*(queue **)&((*(q))[0]))
-#define QUEUE_PREV(q) (*(queue **)&((*(q))[1]))
-
-#define QUEUE_PREV_NEXT(q) (QUEUE_NEXT(QUEUE_PREV(q)))
-#define QUEUE_NEXT_PREV(q) (QUEUE_PREV(QUEUE_NEXT(q)))
-
-/* Initialize an empty queue. */
-#define QUEUE_INIT(q)                \
-	{                            \
-		QUEUE_NEXT(q) = (q); \
-		QUEUE_PREV(q) = (q); \
-	}
-
-/* Return true if the queue has no element. */
-#define QUEUE_IS_EMPTY(q) ((const queue *)(q) == (const queue *)QUEUE_NEXT(q))
-
-/* Insert an element at the back of a queue. */
-#define QUEUE_PUSH(q, e)                       \
-	{                                      \
-		QUEUE_NEXT(e) = (q);           \
-		QUEUE_PREV(e) = QUEUE_PREV(q); \
-		QUEUE_PREV_NEXT(e) = (e);      \
-		QUEUE_PREV(q) = (e);           \
-	}
-
-/* Remove the given element from the queue. Any element can be removed at any *
- * time. */
-#define QUEUE_REMOVE(e)                             \
-	{                                           \
-		QUEUE_PREV_NEXT(e) = QUEUE_NEXT(e); \
-		QUEUE_NEXT_PREV(e) = QUEUE_PREV(e); \
-	}
-
-/* Return the element at the front of the queue. */
-#define QUEUE_HEAD(q) (QUEUE_NEXT(q))
-
-/* Return the element at the back of the queue. */
-#define QUEUE_TAIL(q) (QUEUE_PREV(q))
-
-/* Iterate over the element of a queue. * Mutating the queue while iterating
- * results in undefined behavior. */
-#define QUEUE_FOREACH(q, e) \
-	for ((q) = QUEUE_NEXT(e); (q) != (e); (q) = QUEUE_NEXT(q))
-
-/* Return the structure holding the given element. */
-#define QUEUE_DATA(e, type, field) \
-	((type *)((void *)((char *)(e)-offsetof(type, field))))
-
-#endif /* QUEUE_H_*/
diff -pruN 1.16.4-4/src/raft/recv_install_snapshot.c 1.16.5-1/src/raft/recv_install_snapshot.c
--- 1.16.4-4/src/raft/recv_install_snapshot.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/recv_install_snapshot.c	2024-06-28 13:17:26.000000000 +0000
@@ -8,6 +8,354 @@
 #include "recv.h"
 #include "replication.h"
 
+#include "../lib/sm.h"
+
+/**
+ * =Overview
+ *
+ * This detailed level design is based on PL018 and describes
+ * significant implementation details of data structures, RPCs
+ * introduced in it; provides model of operation and failure handling
+ * based on Leader's and Follower's states.
+ *
+ * =Data structures
+ *
+ * Among other structures it's needed to introduce a (persistent) container
+ * `HT` to efficiently store and map checksums to their page numbers on both
+ * the leader's and  follower's side. HT is implemented on top of sqlite3
+ * database with unix VFS. Every database corresponds to a raft-related
+ * database and maintains the following schema:
+ *
+ * CREATE TABLE "map" ("checksum" INTEGER NOT NULL, "pageno" INTEGER NOT NULL UNIQUE)
+ * CREATE INDEX map_idx on map(checksum);
+ *
+ * Each database stores a mapping from checksum to page number. This
+ * provides an efficient way to insert and lookup records
+ * corresponding to the checksums and page numbers.
+ */
+
+typedef unsigned int checksum_t;
+typedef unsigned long int pageno_t;
+
+struct page_checksum_t {
+	pageno_t   page_no;
+	checksum_t checksum;
+};
+
+struct page_from_to {
+	pageno_t from;
+	pageno_t to;
+};
+
+enum result {
+	OK = 0,
+	UNEXPECTED = 1,
+	DONE = 2,
+};
+
+struct raft_signature {
+	int version;
+
+	const char *db;
+	struct page_from_to page_from_to;
+	unsigned int cs_page_no;
+};
+
+struct raft_signature_result {
+	int version;
+
+	const char *db;
+	struct page_checksum_t *cs;
+	unsigned int cs_nr;
+	unsigned int cs_page_no;
+	enum result result;
+};
+
+struct raft_install_snapshot_mv {
+	int version;
+
+	const char *db;
+	struct page_from_to *mv;
+	unsigned int mv_nr;
+};
+
+struct raft_install_snapshot_mv_result {
+	int version;
+
+	const char *db;
+	pageno_t last_known_page_no; /* used for retries and message losses */
+	enum result result;
+};
+
+struct raft_install_snapshot_cp {
+	int version;
+
+	const char *db;
+	pageno_t page_no;
+	struct raft_buffer page_data;
+	enum result result;
+};
+
+struct raft_install_snapshot_cp_result {
+	int version;
+
+	pageno_t last_known_page_no; /* used for retries and message losses */
+	enum result result;
+};
+
+/**
+ * =Operation
+ *
+ * 0. Leader creates one state machine per Follower to track of their states
+ * and moves it to FOLLOWER_ONLINE state. Follower creates a state machine to
+ * keep track of its states and moves it to NORMAL state.
+ *
+ * 1. The Leader learns the Follower’s follower.lastLogIndex during receiving
+ * replies on AppendEntries() RPC, fails to find follower.lastLogIndex in its
+ * RAFT log or tries and fails to construct an AppendEntries() message because
+ * of the WAL that contained some necessary frames has been rotated out, and
+ * understands that the snapshot installation procedure is required.
+ *
+ * Leader calls leader_tick() putting struct raft_message as a parameter which
+ * logic moves it from FOLLOWER_ONLINE to FOLLOWER_NEEDS_SNAPSHOT state.
+ *
+ * 2. The Leader initiates the snapshot installation by sending
+ * InstallSnapshot() message.
+ *
+ * 3. Upon receiving this message on the Follower's side, Follower calls
+ * follower_tick() putting struct raft_message as a parameter which logic moves
+ * it from NORMAL to SIGNATURES_CALC_STARTED state. The Follower then creates
+ * its HT and starts calculating checksums and recording them. Once finished it
+ * sends the leader the InstallSnapshotResult() message and the Leader moves to
+ * SIGNATURES_CALC_STARTED and creates its HT.
+ *
+ * 3. The Leader sends Signature() messages to the Follower containing the page
+ * range for which we want to get the checksums.
+ *
+ * The Follower sends the requested checksums in a SignatureResult() message
+ * back to the Leader and the leader puts incomming payloads of Signature()
+ * message into the HT.
+ *
+ * 4. When the follower sends the checksum of its highest numbered page to the
+ * Leader, it sends the SignatureResult() message using the done=true flag,
+ * upon receiving it the Leader moves into SNAPSHOT_INSTALLATION_STARTED state.
+ *
+ * 5. In SNAPSHOT_INSTALLATION_STARTED state, the Leader starts iterating over
+ * the local persistent state, and calculates the checksum for each page the
+ * state has. Then, it tries to find the checksum it calculated in HT. Based on
+ * the result of this calculation, the Leader sends InstallShapshot(CP..) or
+ * InstallShapshot(MV..) to the Follower.
+ *
+ * Upon receving these messages, the Follower moves into
+ * SNAPSHOT_CHUNCK_RECEIVED state. The Leader moves into SNAPSHOT_CHUNCK_SENT
+ * state after receiving first reply from the Follower.
+ *
+ * 6. When the iteration has finished the Leader sends
+ * InstallShapshot(..., done=true) message to the Follower. It moves the
+ * Follower back to NORMAL state and the state machine corresponding to the
+ * Follower on the Leader is moved to SNAPSHOT_DONE_SENT state.
+ *
+ * 7. The Leader sends AppendEntries() RPC to the Follower and restarts the
+ * algorithm from (1). The Leader's state machine is being moved to
+ * FOLLOWER_ONLINE state.
+ *
+ * =Failure model
+ *
+ * ==Unavailability of the Leader and Follower.
+ *
+ * To handle use-cases when any party of the communication becomes
+ * unavailable for a while without crash the following assumtions are
+ * made:
+ *
+ * - Signature() or InstallSnapshot(MV/CP) messages are idempotent and
+ *   can be applied to the persistent state many times resulting the
+ *   same transition.
+ *
+ * - Each message with data chuncks has an information about the
+ *   "chunk index". Chunk indexes come in monotonically increasing
+ *   order.
+ *
+ * - Each reply message acknowledges that the data received (or
+ *   ignored) by sending `result` field back to the counter part along
+ *   with last known chunk index as a confirmation that the receiver
+ *   "knows everything up to the given chunck index".
+ *
+ * - If a party notices that last known chunk index sent back to it
+ *   doesn't match it's own, the communication get's restarted from
+ *   the lowest known index.
+ *
+ * If a reply is not received the Leader will eventually timeout and retry
+ * sending the same message.
+ *
+ * ==Crashes of the Leader and Follower.
+ *
+ * Crashes of the Leader are handled by Raft when a new leader is elected
+ * and the snapshot process is restarted.
+ *
+ * If the Follower receives an message which is not expected in the Follower's
+ * current state, the Follower will reply using the message's result RPC
+ * setting the unexpected=true flag. This response suggests the Leader to
+ * restart the snapshot installation procedure.
+ *
+ * In particular, if the follower crashes it will restart its state machine to
+ * the NORMAL state and reply using the unexpected=true flag for any messages
+ * not expected in the NORMAL state, suggesting the Leader to restart the
+ * procedure.
+ *
+ * =State model
+ *
+ * Definitions:
+ *
+ * Rf -- raft index sent in AppendEntriesResult() from Follower to Leader
+ * Tf -- Follower's term sent in AppendEntriesResult() from Follower to Leader
+ *
+ * Tl -- Leader's term
+ * Rl -- raft index of the Leader
+ *
+  * Leader's state machine:
+ *
+ * +---------------------------------+
+ * |                                 |     AppendEntriesResult() received
+ * |        *Result(unexpected=true) |     raft_log.find(Rf) == "FOUND"
+ * |        received                 |      +------------+
+ * |        +---------------> FOLLOWER_ONLINE <----------+
+ * |        |                        |
+ * |        |                        | AppendEntriesResult() received
+ * |        |                        V Rf << Rl && raft_log.find(Rf) == "ENOENTRY"
+ * |        +------------- FOLLOWER_NEEDS_SNAPSHOT
+ * |        |                        |
+ * |        |                        | InstallSnapshotResult() received
+ * |        |                        V
+ * |        +------------ SIGNATURES_CALC_STARTED  -------+ SignatureResult() received
+ * |        |                        |             <------+
+ * |        |                        | SignatureResult(done=true) received
+ * |        |                        V
+ * |        +----------- SNAPSHOT_INSTALLATION_STARTED
+ * |        |                        |
+ * |        |                        | CP_Result()/MV_Result() received
+ * |        |                        V
+ * |        +--------------- SNAPSHOT_CHUNCK_SENT -------+ CP_Result()/MV_Result()
+ * |        |                        |            <------+  received
+ * |        |                        | Raw snapshot iteration done
+ * |        |                        V
+ * |        +--------------- SNAPSHOT_DONE_SENT
+ * |                                 | InstallSnapshotResult() received
+ * +---------------------------------+
+ *
+ * Follower's state machine:
+ *
+ *                            +------+ (%)
+ * +-------------------> NORMAL <----+
+ * |       +----------->   |
+ * |       |               | InstallSnapshot() received
+ * |       |               V
+ * |       +--- SIGNATURES_CALC_STARTED
+ * |       |               |
+ * |       |               | Signatures for all db pages have been calculated.
+ * |       |               | InstallSnapshotResult() sent.
+ * |       |               V
+ * |       +---- SIGNATURES_CALC_DONE
+ * |       |               |
+ * |       |               | First SignatureResult() sent
+ * |       |               V
+ * |       +---- SIGNATURES_PART_SENT ---------+ Signature() received
+ * |       |               |          <--------+ SignatureResult() sent
+ * |       |               |
+ * |       |               | CP()/MV() received
+ * |       |               V
+ * |       +--- SNAPSHOT_CHUNCK_RECEIVED -------+ CP()/MV() received
+ * |    (@ || %)           |             <------+
+ * |                       | InstallSnapshot(done=true) received
+ * +-----------------------+
+ *
+ * (@) -- AppendEntries() received && Tf < Tl
+ * (%) -- Signature()/CP()/MV() received and in the current state receving a
+ *        message of such type is unexpected. *Result(unexpected=true) sent.
+ */
+
+enum follower_states {
+	FS_NORMAL,
+	FS_SIGNATURES_CALC_STARTED,
+	FS_SIGNATURES_CALC_DONE,
+	FS_SIGNATURES_PART_SENT,
+	FS_SNAPSHOT_CHUNCK_RECEIVED,
+	FS_NR,
+};
+
+static const struct sm_conf follower_states[FS_NR] = {
+	[FS_NORMAL] = {
+		.flags = SM_INITIAL | SM_FINAL,
+		.name = "normal",
+		.allowed = BITS(FS_SIGNATURES_CALC_STARTED),
+	},
+/*	[PS_DRAINING] = {
+	    .name = "draining",
+	    .allowed = BITS(PS_DRAINING)
+		     | BITS(PS_NOTHING)
+		     | BITS(PS_BARRIER),
+	},
+*/
+};
+
+ __attribute__((unused)) static void follower_tick(struct sm *follower, const struct raft_message *msg)
+{
+	(void) follower;
+	(void) msg;
+	(void) follower_states;
+	//switch (sm_state(follower)) {
+	//}
+}
+
+__attribute__((unused)) static bool follower_invariant(const struct sm *m, int prev_state)
+{
+	(void) m;
+	(void) prev_state;
+	//pool_impl_t *pi = CONTAINER_OF(m, pool_impl_t, planner_sm);
+	return true;
+}
+
+enum leader_states {
+	LS_FOLLOWER_ONLINE,
+	LS_FOLLOWER_NEEDS_SNAPSHOT,
+	LS_SIGNATURES_CALC_STARTED,
+	LS_SNAPSHOT_INSTALLATION_STARTED,
+	LS_SNAPSHOT_CHUNCK_SENT,
+	LS_SNAPSHOT_DONE_SENT,
+	LS_NR,
+};
+
+static const struct sm_conf leader_states[LS_NR] = {
+	[LS_FOLLOWER_ONLINE] = {
+		.flags = SM_INITIAL | SM_FINAL,
+		.name = "online",
+		.allowed = BITS(LS_FOLLOWER_NEEDS_SNAPSHOT),
+	},
+/*	[PS_DRAINING] = {
+	    .name = "draining",
+	    .allowed = BITS(PS_DRAINING)
+		     | BITS(PS_NOTHING)
+		     | BITS(PS_BARRIER),
+	},
+*/
+};
+
+__attribute__((unused)) static void leader_tick(struct sm *leader, const struct raft_message *msg)
+{
+	(void) leader;
+	(void) msg;
+	(void) leader_states;
+	//switch (sm_state(leader)) {
+	//}
+}
+
+__attribute__((unused)) static bool leader_invariant(const struct sm *m, int prev_state)
+{
+	(void) m;
+	(void) prev_state;
+	//pool_impl_t *pi = CONTAINER_OF(m, pool_impl_t, planner_sm);
+	return true;
+}
+
 static void installSnapshotSendCb(struct raft_io_send *req, int status)
 {
 	(void)status;
diff -pruN 1.16.4-4/src/raft/replication.c 1.16.5-1/src/raft/replication.c
--- 1.16.4-4/src/raft/replication.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/replication.c	2024-06-28 13:17:26.000000000 +0000
@@ -15,7 +15,7 @@
 #include "log.h"
 #include "membership.h"
 #include "progress.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "replication.h"
 #include "request.h"
 #include "snapshot.h"
@@ -1235,7 +1235,7 @@ int replicationAppend(struct raft *r,
 			goto err_after_request_alloc;
 		}
 
-		rv = logAppend(r->log, copy.term, copy.type, &copy.buf, NULL);
+		rv = logAppend(r->log, copy.term, copy.type, copy.buf, (struct raft_entry_local_data){}, false, NULL);
 		if (rv != 0) {
 			goto err_after_request_alloc;
 		}
diff -pruN 1.16.4-4/src/raft/request.h 1.16.5-1/src/raft/request.h
--- 1.16.4-4/src/raft/request.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/request.h	2024-06-28 13:17:26.000000000 +0000
@@ -10,7 +10,7 @@ struct request
 	void *data;
 	int type;
 	raft_index index;
-	void *queue[2];
+	queue queue;
 	uint8_t req_id[16];
 	uint8_t client_id[16];
 	uint8_t unique_id[16];
diff -pruN 1.16.4-4/src/raft/start.c 1.16.5-1/src/raft/start.c
--- 1.16.4-4/src/raft/start.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/start.c	2024-06-28 13:17:26.000000000 +0000
@@ -71,8 +71,8 @@ static int restoreEntries(struct raft *r
 	r->last_stored = start_index - 1;
 	for (i = 0; i < n; i++) {
 		struct raft_entry *entry = &entries[i];
-		rv = logAppend(r->log, entry->term, entry->type, &entry->buf,
-			       entry->batch);
+		rv = logAppend(r->log, entry->term, entry->type, entry->buf,
+			       entry->local_data, entry->is_local, entry->batch);
 		if (rv != 0) {
 			goto err;
 		}
diff -pruN 1.16.4-4/src/raft/state.c 1.16.5-1/src/raft/state.c
--- 1.16.4-4/src/raft/state.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/state.c	2024-06-28 13:17:26.000000000 +0000
@@ -2,7 +2,7 @@
 #include "configuration.h"
 #include "election.h"
 #include "log.h"
-#include "queue.h"
+#include "../lib/queue.h"
 
 int raft_state(struct raft *r)
 {
diff -pruN 1.16.4-4/src/raft/uv.c 1.16.5-1/src/raft/uv.c
--- 1.16.4-4/src/raft/uv.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv.c	2024-06-28 13:17:26.000000000 +0000
@@ -172,10 +172,10 @@ void uvMaybeFireCloseCb(struct uv *uv)
 	if (uv->timer.data != NULL) {
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->append_segments)) {
+	if (!queue_empty(&uv->append_segments)) {
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->finalize_reqs)) {
+	if (!queue_empty(&uv->finalize_reqs)) {
 		return;
 	}
 	if (uv->finalize_work.data != NULL) {
@@ -190,13 +190,13 @@ void uvMaybeFireCloseCb(struct uv *uv)
 	if (uv->snapshot_put_work.data != NULL) {
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->snapshot_get_reqs)) {
+	if (!queue_empty(&uv->snapshot_get_reqs)) {
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->async_work_reqs)) {
+	if (!queue_empty(&uv->async_work_reqs)) {
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->aborting)) {
+	if (!queue_empty(&uv->aborting)) {
 		return;
 	}
 
@@ -698,28 +698,28 @@ int raft_uv_init(struct raft_io *io,
 #endif
 	uv->segment_size = UV__MAX_SEGMENT_SIZE;
 	uv->block_size = 0;
-	QUEUE_INIT(&uv->clients);
-	QUEUE_INIT(&uv->servers);
+	queue_init(&uv->clients);
+	queue_init(&uv->servers);
 	uv->connect_retry_delay = CONNECT_RETRY_DELAY;
 	uv->prepare_inflight = NULL;
-	QUEUE_INIT(&uv->prepare_reqs);
-	QUEUE_INIT(&uv->prepare_pool);
+	queue_init(&uv->prepare_reqs);
+	queue_init(&uv->prepare_pool);
 	uv->prepare_next_counter = 1;
 	uv->append_next_index = 1;
-	QUEUE_INIT(&uv->append_segments);
-	QUEUE_INIT(&uv->append_pending_reqs);
-	QUEUE_INIT(&uv->append_writing_reqs);
+	queue_init(&uv->append_segments);
+	queue_init(&uv->append_pending_reqs);
+	queue_init(&uv->append_writing_reqs);
 	uv->barrier = NULL;
-	QUEUE_INIT(&uv->finalize_reqs);
+	queue_init(&uv->finalize_reqs);
 	uv->finalize_work.data = NULL;
 	uv->truncate_work.data = NULL;
-	QUEUE_INIT(&uv->snapshot_get_reqs);
-	QUEUE_INIT(&uv->async_work_reqs);
+	queue_init(&uv->snapshot_get_reqs);
+	queue_init(&uv->async_work_reqs);
 	uv->snapshot_put_work.data = NULL;
 	uv->timer.data = NULL;
 	uv->tick_cb = NULL; /* Set by raft_io->start() */
 	uv->recv_cb = NULL; /* Set by raft_io->start() */
-	QUEUE_INIT(&uv->aborting);
+	queue_init(&uv->aborting);
 	uv->closing = false;
 	uv->close_cb = NULL;
 	uv->auto_recovery = true;
diff -pruN 1.16.4-4/src/raft/uv.h 1.16.5-1/src/raft/uv.h
--- 1.16.4-4/src/raft/uv.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv.h	2024-06-28 13:17:26.000000000 +0000
@@ -6,7 +6,7 @@
 #include "../raft.h"
 #include "../tracing.h"
 #include "err.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "uv_fs.h"
 #include "uv_os.h"
 
diff -pruN 1.16.4-4/src/raft/uv_append.c 1.16.5-1/src/raft/uv_append.c
--- 1.16.4-4/src/raft/uv_append.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_append.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,7 +1,7 @@
 #include "assert.h"
 #include "byte.h"
 #include "heap.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "uv.h"
 #include "uv_encoding.h"
 #include "uv_writer.h"
@@ -92,7 +92,7 @@ static void uvAliveSegmentFinalize(struc
 		 * close the file handle and release the segment memory. */
 	}
 
-	QUEUE_REMOVE(&s->queue);
+	queue_remove(&s->queue);
 	UvWriterClose(&s->writer, uvAliveSegmentWriterCloseCb);
 }
 
@@ -102,10 +102,10 @@ static void uvAppendFinishRequestsInQueu
 {
 	queue queue_copy;
 	struct uvAppend *append;
-	QUEUE_INIT(&queue_copy);
-	while (!QUEUE_IS_EMPTY(q)) {
+	queue_init(&queue_copy);
+	while (!queue_empty(q)) {
 		queue *head;
-		head = QUEUE_HEAD(q);
+		head = queue_head(q);
 		append = QUEUE_DATA(head, struct uvAppend, queue);
 		/* Rollback the append next index if the result was
 		 * unsuccessful. */
@@ -116,15 +116,15 @@ static void uvAppendFinishRequestsInQueu
 			tracef("rollback uv->append_next_index now:%llu",
 			       uv->append_next_index);
 		}
-		QUEUE_REMOVE(head);
-		QUEUE_PUSH(&queue_copy, head);
+		queue_remove(head);
+		queue_insert_tail(&queue_copy, head);
 	}
-	while (!QUEUE_IS_EMPTY(&queue_copy)) {
+	while (!queue_empty(&queue_copy)) {
 		queue *head;
 		struct raft_io_append *req;
-		head = QUEUE_HEAD(&queue_copy);
+		head = queue_head(&queue_copy);
 		append = QUEUE_DATA(head, struct uvAppend, queue);
-		QUEUE_REMOVE(head);
+		queue_remove(head);
 		req = append->req;
 		RaftHeapFree(append);
 		req->cb(req, status);
@@ -150,10 +150,10 @@ static void uvAppendFinishPendingRequest
 static struct uvAliveSegment *uvGetCurrentAliveSegment(struct uv *uv)
 {
 	queue *head;
-	if (QUEUE_IS_EMPTY(&uv->append_segments)) {
+	if (queue_empty(&uv->append_segments)) {
 		return NULL;
 	}
-	head = QUEUE_HEAD(&uv->append_segments);
+	head = queue_head(&uv->append_segments);
 	return QUEUE_DATA(head, struct uvAliveSegment, queue);
 }
 
@@ -282,14 +282,14 @@ out:
 	/* During the closing sequence we should have already canceled all
 	 * pending request. */
 	if (uv->closing) {
-		assert(QUEUE_IS_EMPTY(&uv->append_pending_reqs));
+		assert(queue_empty(&uv->append_pending_reqs));
 		assert(s->finalize);
 		uvAliveSegmentFinalize(s);
 		return;
 	}
 
 	/* Possibly process waiting requests. */
-	if (!QUEUE_IS_EMPTY(&uv->append_pending_reqs)) {
+	if (!queue_empty(&uv->append_pending_reqs)) {
 		rv = uvAppendMaybeStart(uv);
 		if (rv != 0) {
 			uv->errored = true;
@@ -338,10 +338,10 @@ static int uvAppendMaybeStart(struct uv
 	int rv;
 
 	assert(!uv->closing);
-	assert(!QUEUE_IS_EMPTY(&uv->append_pending_reqs));
+	assert(!queue_empty(&uv->append_pending_reqs));
 
 	/* If we are already writing, let's wait. */
-	if (!QUEUE_IS_EMPTY(&uv->append_writing_reqs)) {
+	if (!queue_empty(&uv->append_writing_reqs)) {
 		return 0;
 	}
 
@@ -372,18 +372,18 @@ start:
 
 	/* Let's add to the segment's write buffer all pending requests targeted
 	 * to this segment. */
-	QUEUE_INIT(&q);
+	queue_init(&q);
 
 	n_reqs = 0;
-	while (!QUEUE_IS_EMPTY(&uv->append_pending_reqs)) {
-		head = QUEUE_HEAD(&uv->append_pending_reqs);
+	while (!queue_empty(&uv->append_pending_reqs)) {
+		head = queue_head(&uv->append_pending_reqs);
 		append = QUEUE_DATA(head, struct uvAppend, queue);
 		assert(append->segment != NULL);
 		if (append->segment != segment) {
 			break; /* Not targeted to this segment */
 		}
-		QUEUE_REMOVE(head);
-		QUEUE_PUSH(&q, head);
+		queue_remove(head);
+		queue_insert_tail(&q, head);
 		n_reqs++;
 		rv = uvAliveSegmentEncodeEntriesToWriteBuf(segment, append);
 		if (rv != 0) {
@@ -397,21 +397,21 @@ start:
 	 * request, in that case we need to wait for it). Otherwise it must mean
 	 * we have exhausted the queue of pending append requests. */
 	if (n_reqs == 0) {
-		assert(QUEUE_IS_EMPTY(&uv->append_writing_reqs));
+		assert(queue_empty(&uv->append_writing_reqs));
 		if (segment->finalize) {
 			uvAliveSegmentFinalize(segment);
-			if (!QUEUE_IS_EMPTY(&uv->append_pending_reqs)) {
+			if (!queue_empty(&uv->append_pending_reqs)) {
 				goto start;
 			}
 		}
-		assert(QUEUE_IS_EMPTY(&uv->append_pending_reqs));
+		assert(queue_empty(&uv->append_pending_reqs));
 		return 0;
 	}
 
-	while (!QUEUE_IS_EMPTY(&q)) {
-		head = QUEUE_HEAD(&q);
-		QUEUE_REMOVE(head);
-		QUEUE_PUSH(&uv->append_writing_reqs, head);
+	while (!queue_empty(&q)) {
+		head = queue_head(&q);
+		queue_remove(head);
+		queue_insert_tail(&uv->append_writing_reqs, head);
 	}
 
 	rv = uvAliveSegmentWrite(segment);
@@ -457,7 +457,7 @@ static void uvAliveSegmentPrepareCb(stru
 
 	/* If we have been closed, let's discard the segment. */
 	if (uv->closing) {
-		QUEUE_REMOVE(&segment->queue);
+		queue_remove(&segment->queue);
 		assert(status ==
 		       RAFT_CANCELED); /* UvPrepare cancels pending reqs */
 		uvSegmentBufferClose(&segment->pending);
@@ -476,7 +476,7 @@ static void uvAliveSegmentPrepareCb(stru
 
 	/* There must be pending appends that were waiting for this prepare
 	 * requests. */
-	assert(!QUEUE_IS_EMPTY(&uv->append_pending_reqs));
+	assert(!queue_empty(&uv->append_pending_reqs));
 
 	rv = uvAliveSegmentReady(uv, req->fd, req->counter, segment);
 	if (rv != 0) {
@@ -493,7 +493,7 @@ static void uvAliveSegmentPrepareCb(stru
 	return;
 
 err:
-	QUEUE_REMOVE(&segment->queue);
+	queue_remove(&segment->queue);
 	RaftHeapFree(segment);
 	uv->errored = true;
 	uvAppendFinishPendingRequests(uv, rv);
@@ -535,7 +535,7 @@ static int uvAppendPushAliveSegment(stru
 	}
 	uvAliveSegmentInit(segment, uv);
 
-	QUEUE_PUSH(&uv->append_segments, &segment->queue);
+	queue_insert_tail(&uv->append_segments, &segment->queue);
 
 	rv = UvPrepare(uv, &fd, &counter, &segment->prepare,
 		       uvAliveSegmentPrepareCb);
@@ -557,7 +557,7 @@ err_after_prepare:
 	UvOsClose(fd);
 	UvFinalize(uv, counter, 0, 0, 0);
 err_after_alloc:
-	QUEUE_REMOVE(&segment->queue);
+	queue_remove(&segment->queue);
 	RaftHeapFree(segment);
 err:
 	assert(rv != 0);
@@ -568,10 +568,10 @@ err:
 static struct uvAliveSegment *uvGetLastAliveSegment(struct uv *uv)
 {
 	queue *tail;
-	if (QUEUE_IS_EMPTY(&uv->append_segments)) {
+	if (queue_empty(&uv->append_segments)) {
 		return NULL;
 	}
-	tail = QUEUE_TAIL(&uv->append_segments);
+	tail = queue_tail(&uv->append_segments);
 	return QUEUE_DATA(tail, struct uvAliveSegment, queue);
 }
 
@@ -597,7 +597,7 @@ static size_t uvAppendSize(struct uvAppe
 {
 	size_t size = sizeof(uint32_t) * 2; /* CRC checksums */
 	unsigned i;
-	size += uvSizeofBatchHeader(a->n); /* Batch header */
+	size += uvSizeofBatchHeader(a->n, true); /* Batch header */
 	for (i = 0; i < a->n; i++) {       /* Entries data */
 		size += bytePad64(a->entries[i].buf.len);
 	}
@@ -647,7 +647,7 @@ static int uvAppendEnqueueRequest(struct
 	uvAliveSegmentReserveSegmentCapacity(segment, size);
 
 	append->segment = segment;
-	QUEUE_PUSH(&uv->append_pending_reqs, &append->queue);
+	queue_insert_tail(&uv->append_pending_reqs, &append->queue);
 	uv->append_next_index += append->n;
 	tracef("set uv->append_next_index %llu", uv->append_next_index);
 
@@ -711,7 +711,7 @@ int UvAppend(struct raft_io *io,
 	}
 
 	assert(append->segment != NULL);
-	assert(!QUEUE_IS_EMPTY(&uv->append_pending_reqs));
+	assert(!queue_empty(&uv->append_pending_reqs));
 
 	/* Try to write immediately. */
 	rv = uvAppendMaybeStart(uv);
@@ -753,7 +753,7 @@ static void uvFinalizeCurrentAliveSegmen
 			break;
 		}
 	}
-	has_writing_reqs = !QUEUE_IS_EMPTY(&uv->append_writing_reqs);
+	has_writing_reqs = !queue_empty(&uv->append_writing_reqs);
 
 	/* If there is no pending append request or inflight write against the
 	 * current segment, we can submit a request for it to be closed
@@ -792,11 +792,11 @@ bool UvBarrierMaybeTrigger(struct UvBarr
 		return false;
 	}
 
-	if (!QUEUE_IS_EMPTY(&barrier->reqs)) {
+	if (!queue_empty(&barrier->reqs)) {
 		queue *head;
 		struct UvBarrierReq *r;
-		head = QUEUE_HEAD(&barrier->reqs);
-		QUEUE_REMOVE(head);
+		head = queue_head(&barrier->reqs);
+		queue_remove(head);
 		r = QUEUE_DATA(head, struct UvBarrierReq, queue);
 		r->cb(r);
 		return true;
@@ -821,7 +821,7 @@ static struct UvBarrier *uvBarrierCreate
 		return NULL;
 	}
 	barrier->blocking = false;
-	QUEUE_INIT(&barrier->reqs);
+	queue_init(&barrier->reqs);
 	return barrier;
 }
 
@@ -902,8 +902,8 @@ int UvBarrier(struct uv *uv, raft_index
 		 * the callback immediately.
 		 *
 		 * TODO: find a way to avoid invoking this synchronously. */
-		if (QUEUE_IS_EMPTY(&uv->append_segments) &&
-		    QUEUE_IS_EMPTY(&uv->finalize_reqs) &&
+		if (queue_empty(&uv->append_segments) &&
+		    queue_empty(&uv->finalize_reqs) &&
 		    uv->finalize_work.data == NULL) {
 			/* Not interested in return value. */
 			UvBarrierMaybeTrigger(barrier);
@@ -930,7 +930,7 @@ void UvUnblock(struct uv *uv)
 		uvMaybeFireCloseCb(uv);
 		return;
 	}
-	if (!QUEUE_IS_EMPTY(&uv->append_pending_reqs)) {
+	if (!queue_empty(&uv->append_pending_reqs)) {
 		int rv;
 		rv = uvAppendMaybeStart(uv);
 		if (rv != 0) {
@@ -945,7 +945,7 @@ void UvBarrierAddReq(struct UvBarrier *b
 	assert(req != NULL);
 	/* Once there's a blocking req, this barrier becomes blocking. */
 	barrier->blocking |= req->blocking;
-	QUEUE_PUSH(&barrier->reqs, &req->queue);
+	queue_insert_tail(&barrier->reqs, &req->queue);
 }
 
 /* Fire all pending barrier requests, the barrier callback will notice that
@@ -1022,7 +1022,7 @@ void uvAppendClose(struct uv *uv)
 	/* Also finalize the segments that we didn't write at all and are just
 	 * sitting in the append_segments queue waiting for writes against the
 	 * current segment to complete. */
-	while (!QUEUE_IS_EMPTY(&uv->append_segments)) {
+	while (!queue_empty(&uv->append_segments)) {
 		segment = uvGetLastAliveSegment(uv);
 		assert(segment != NULL);
 		if (segment == uvGetCurrentAliveSegment(uv)) {
diff -pruN 1.16.4-4/src/raft/uv_encoding.c 1.16.5-1/src/raft/uv_encoding.c
--- 1.16.4-4/src/raft/uv_encoding.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_encoding.c	2024-06-28 13:17:26.000000000 +0000
@@ -86,10 +86,16 @@ static size_t sizeofTimeoutNow(void)
 	       sizeof(uint64_t) /* Last log term. */;
 }
 
-size_t uvSizeofBatchHeader(size_t n)
+size_t uvSizeofBatchHeader(size_t n, bool with_local_data)
 {
-	return 8 + /* Number of entries in the batch, little endian */
-	       16 * n /* One header per entry */;
+	size_t res = 8 + /* Number of entries in the batch, little endian */
+		16 * n; /* One header per entry */;
+	if (with_local_data) {
+#ifdef DQLITE_NEXT
+		res += 8; /* Local data length, applies to all entries */
+#endif
+	}
+	return res;
 }
 
 static void encodeRequestVote(const struct raft_request_vote *p, void *buf)
@@ -137,7 +143,7 @@ static void encodeAppendEntries(const st
 	bytePut64(&cursor, p->prev_log_term);  /* Previous term. */
 	bytePut64(&cursor, p->leader_commit);  /* Commit index. */
 
-	uvEncodeBatchHeader(p->entries, p->n_entries, cursor);
+	uvEncodeBatchHeader(p->entries, p->n_entries, cursor, false /* no local data */);
 }
 
 static void encodeAppendEntriesResult(
@@ -295,7 +301,8 @@ oom:
 
 void uvEncodeBatchHeader(const struct raft_entry *entries,
 			 unsigned n,
-			 void *buf)
+			 void *buf,
+			 bool with_local_data)
 {
 	unsigned i;
 	void *cursor = buf;
@@ -303,6 +310,13 @@ void uvEncodeBatchHeader(const struct ra
 	/* Number of entries in the batch, little endian */
 	bytePut64(&cursor, n);
 
+	if (with_local_data) {
+#ifdef DQLITE_NEXT
+		/* Local data size per entry, little endian */
+		bytePut64(&cursor, (uint64_t)sizeof(struct raft_entry_local_data));
+#endif
+	}
+
 	for (i = 0; i < n; i++) {
 		const struct raft_entry *entry = &entries[i];
 
@@ -363,7 +377,8 @@ static void decodeRequestVoteResult(cons
 
 int uvDecodeBatchHeader(const void *batch,
 			struct raft_entry **entries,
-			unsigned *n)
+			unsigned *n,
+			uint64_t *local_data_size)
 {
 	const void *cursor = batch;
 	size_t i;
@@ -376,6 +391,17 @@ int uvDecodeBatchHeader(const void *batc
 		return 0;
 	}
 
+	if (local_data_size != NULL) {
+#ifdef DQLITE_NEXT
+		uint64_t z = byteGet64(&cursor);
+		if (z == 0 || z > sizeof(struct raft_entry_local_data) || z % sizeof(uint64_t) != 0) {
+			rv = RAFT_MALFORMED;
+			goto err;
+		}
+		*local_data_size = z;
+#endif
+	}
+
 	*entries = raft_malloc(*n * sizeof **entries);
 
 	if (*entries == NULL) {
@@ -430,7 +456,7 @@ static int decodeAppendEntries(const uv_
 	args->prev_log_term = byteGet64(&cursor);
 	args->leader_commit = byteGet64(&cursor);
 
-	rv = uvDecodeBatchHeader(cursor, &args->entries, &args->n_entries);
+	rv = uvDecodeBatchHeader(cursor, &args->entries, &args->n_entries, false);
 	if (rv != 0) {
 		return rv;
 	}
@@ -549,33 +575,38 @@ int uvDecodeMessage(uint16_t type,
 	return rv;
 }
 
-void uvDecodeEntriesBatch(uint8_t *batch,
-			  size_t offset,
-			  struct raft_entry *entries,
-			  unsigned n)
+int uvDecodeEntriesBatch(uint8_t *batch,
+			 size_t offset,
+			 struct raft_entry *entries,
+			 unsigned n,
+			 uint64_t local_data_size)
 {
 	uint8_t *cursor;
-	size_t i;
 
 	assert(batch != NULL);
 
 	cursor = batch + offset;
 
-	for (i = 0; i < n; i++) {
+	for (size_t i = 0; i < n; i++) {
 		struct raft_entry *entry = &entries[i];
 		entry->batch = batch;
+		entry->buf.base = (entry->buf.len > 0) ? cursor : NULL;
 
-		if (entry->buf.len == 0) {
-			entry->buf.base = NULL;
-			continue;
-		}
-
-		entry->buf.base = cursor;
-
-		cursor = cursor + entry->buf.len;
+		cursor += entry->buf.len;
 		if (entry->buf.len % 8 != 0) {
 			/* Add padding */
 			cursor = cursor + 8 - (entry->buf.len % 8);
 		}
+
+		entry->is_local = false;
+
+		entry->local_data = (struct raft_entry_local_data){};
+		assert(local_data_size <= sizeof(entry->local_data.buf));
+		assert(local_data_size % 8 == 0);
+#ifdef DQLITE_NEXT
+		memcpy(entry->local_data.buf, cursor, local_data_size);
+		cursor += local_data_size;
+#endif
 	}
+	return 0;
 }
diff -pruN 1.16.4-4/src/raft/uv_encoding.h 1.16.5-1/src/raft/uv_encoding.h
--- 1.16.4-4/src/raft/uv_encoding.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_encoding.h	2024-06-28 13:17:26.000000000 +0000
@@ -8,7 +8,11 @@
 #include "../raft.h"
 
 /* Current disk format version. */
+#ifdef DQLITE_NEXT
+#define UV__DISK_FORMAT 2
+#else
 #define UV__DISK_FORMAT 1
+#endif
 
 int uvEncodeMessage(const struct raft_message *message,
 		    uv_buf_t **bufs,
@@ -21,12 +25,14 @@ int uvDecodeMessage(uint16_t type,
 
 int uvDecodeBatchHeader(const void *batch,
 			struct raft_entry **entries,
-			unsigned *n);
+			unsigned *n,
+			uint64_t *local_data_size);
 
-void uvDecodeEntriesBatch(uint8_t *batch,
-			  size_t offset,
-			  struct raft_entry *entries,
-			  unsigned n);
+int uvDecodeEntriesBatch(uint8_t *batch,
+			 size_t offset,
+			 struct raft_entry *entries,
+			 unsigned n,
+			 uint64_t local_data_size);
 
 /**
  * The layout of the memory pointed at by a @batch pointer is the following:
@@ -45,15 +51,17 @@ void uvDecodeEntriesBatch(uint8_t *batch
  * [1 byte ] Message type (Either RAFT_COMMAND or RAFT_CHANGE)
  * [3 bytes] Currently unused.
  * [4 bytes] Size of the log entry data, little endian.
+ * [8 bytes] Size of the local buffer, little endian.
  *
  * A payload data section for an entry is simply a sequence of bytes of
  * arbitrary lengths, possibly padded with extra bytes to reach 8-byte boundary
  * (which means that all entry data pointers are 8-byte aligned).
  */
-size_t uvSizeofBatchHeader(size_t n);
+size_t uvSizeofBatchHeader(size_t n, bool with_local_data);
 
 void uvEncodeBatchHeader(const struct raft_entry *entries,
 			 unsigned n,
-			 void *buf);
+			 void *buf,
+			 bool with_local_data);
 
 #endif /* UV_ENCODING_H_ */
diff -pruN 1.16.4-4/src/raft/uv_finalize.c 1.16.5-1/src/raft/uv_finalize.c
--- 1.16.4-4/src/raft/uv_finalize.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_finalize.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,6 +1,6 @@
 #include "assert.h"
 #include "heap.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "uv.h"
 #include "uv_os.h"
 
@@ -88,7 +88,7 @@ static void uvFinalizeAfterWorkCb(uv_wor
 
 	/* If we have no more dismissed segments to close, check if there's a
 	 * barrier to unblock or if we are done closing. */
-	if (QUEUE_IS_EMPTY(&uv->finalize_reqs)) {
+	if (queue_empty(&uv->finalize_reqs)) {
 		tracef("unblock barrier or close");
 		if (uv->barrier != NULL && UvBarrierReady(uv)) {
 			UvBarrierMaybeTrigger(uv->barrier);
@@ -98,9 +98,9 @@ static void uvFinalizeAfterWorkCb(uv_wor
 	}
 
 	/* Grab a new dismissed segment to close. */
-	head = QUEUE_HEAD(&uv->finalize_reqs);
+	head = queue_head(&uv->finalize_reqs);
 	segment = QUEUE_DATA(head, struct uvDyingSegment, queue);
-	QUEUE_REMOVE(&segment->queue);
+	queue_remove(&segment->queue);
 
 	rv = uvFinalizeStart(segment);
 	if (rv != 0) {
@@ -160,7 +160,7 @@ int UvFinalize(struct uv *uv,
 	/* If we're already processing a segment, let's put the request in the
 	 * queue and wait. */
 	if (uv->finalize_work.data != NULL) {
-		QUEUE_PUSH(&uv->finalize_reqs, &segment->queue);
+		queue_insert_tail(&uv->finalize_reqs, &segment->queue);
 		return 0;
 	}
 
diff -pruN 1.16.4-4/src/raft/uv_fs.c 1.16.5-1/src/raft/uv_fs.c
--- 1.16.4-4/src/raft/uv_fs.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_fs.c	2024-06-28 13:17:26.000000000 +0000
@@ -704,15 +704,9 @@ static int probeDirectIO(int fd, size_t
 			default:
 				/* UNTESTED: this is an unsupported file system.
 				 */
-#if defined(__s390x__)
 				ErrMsgPrintf(errmsg,
-					     "unsupported file system: %ux",
-					     fs_info.f_type);
-#else
-				ErrMsgPrintf(errmsg,
-					     "unsupported file system: %zx",
-					     fs_info.f_type);
-#endif
+					     "unsupported file system: %llx",
+					     (unsigned long long)fs_info.f_type);
 				return RAFT_IOERR;
 		}
 	}
@@ -840,14 +834,6 @@ static int probeAsyncIO(int fd, size_t s
 		 * errors, since we allocated the file with posix_fallocate and
 		 * the block size is supposed to be correct. */
 		*ok = false;
-		if (event.res == -EAGAIN) {
-			/* If EAGAIN is encountered we assume the functionality
-			 * is supported but this write would have blocked for
-			 * some reason. UvWriter has a fallback mechanism to
-			 * schedule writes on the thread pool in case the async
-			 * write fails with EAGAIN, so this is safe. */
-			*ok = true;
-		}
 	}
 
 	return 0;
@@ -874,6 +860,7 @@ static void probeFallocate(const char *d
 	if (rv == 0) {
 		*fallocate = true;
 	}
+	close(fd);
 
 out:
 	UvFsRemoveFile(dir, UV__FS_PROBE_FALLOCATE_FILE, ignored);
diff -pruN 1.16.4-4/src/raft/uv_ip.c 1.16.5-1/src/raft/uv_ip.c
--- 1.16.4-4/src/raft/uv_ip.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_ip.c	2024-06-28 13:17:26.000000000 +0000
@@ -58,7 +58,7 @@ int uvIpAddrSplit(const char *address,
 int uvIpResolveBindAddresses(const char *address, struct addrinfo **ai_result)
 {
 	static struct addrinfo hints = {
-	    .ai_flags = AI_ADDRCONFIG | AI_PASSIVE | AI_NUMERICSERV,
+	    .ai_flags = AI_PASSIVE | AI_NUMERICSERV,
 	    .ai_family = AF_INET,
 	    .ai_socktype = SOCK_STREAM,
 	    .ai_protocol = 0};
diff -pruN 1.16.4-4/src/raft/uv_os.c 1.16.5-1/src/raft/uv_os.c
--- 1.16.4-4/src/raft/uv_os.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_os.c	2024-06-28 13:17:26.000000000 +0000
@@ -52,7 +52,7 @@ int UvOsFallocateEmulation(int fd, off_t
 	if (f.f_bsize == 0) {
 		increment = 512;
 	} else if (f.f_bsize < 4096) {
-		increment = f.f_bsize;
+		increment = (ssize_t)f.f_bsize;
 	} else {
 		increment = 4096;
 	}
diff -pruN 1.16.4-4/src/raft/uv_prepare.c 1.16.5-1/src/raft/uv_prepare.c
--- 1.16.4-4/src/raft/uv_prepare.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_prepare.c	2024-06-28 13:17:26.000000000 +0000
@@ -77,12 +77,12 @@ err:
  * status. */
 static void uvPrepareFinishAllRequests(struct uv *uv, int status)
 {
-	while (!QUEUE_IS_EMPTY(&uv->prepare_reqs)) {
+	while (!queue_empty(&uv->prepare_reqs)) {
 		queue *head;
 		struct uvPrepare *req;
-		head = QUEUE_HEAD(&uv->prepare_reqs);
+		head = queue_head(&uv->prepare_reqs);
 		req = QUEUE_DATA(head, struct uvPrepare, queue);
-		QUEUE_REMOVE(&req->queue);
+		queue_remove(&req->queue);
 		req->cb(req, status);
 	}
 }
@@ -94,10 +94,10 @@ static void uvPrepareConsume(struct uv *
 	queue *head;
 	struct uvIdleSegment *segment;
 	/* Pop a segment from the pool. */
-	head = QUEUE_HEAD(&uv->prepare_pool);
+	head = queue_head(&uv->prepare_pool);
 	segment = QUEUE_DATA(head, struct uvIdleSegment, queue);
 	assert(segment->fd >= 0);
-	QUEUE_REMOVE(&segment->queue);
+	queue_remove(&segment->queue);
 	*fd = segment->fd;
 	*counter = segment->counter;
 	RaftHeapFree(segment);
@@ -111,13 +111,13 @@ static void uvPrepareFinishOldestRequest
 	struct uvPrepare *req;
 
 	assert(!uv->closing);
-	assert(!QUEUE_IS_EMPTY(&uv->prepare_reqs));
-	assert(!QUEUE_IS_EMPTY(&uv->prepare_pool));
+	assert(!queue_empty(&uv->prepare_reqs));
+	assert(!queue_empty(&uv->prepare_pool));
 
 	/* Pop the head of the prepare requests queue. */
-	head = QUEUE_HEAD(&uv->prepare_reqs);
+	head = queue_head(&uv->prepare_reqs);
 	req = QUEUE_DATA(head, struct uvPrepare, queue);
-	QUEUE_REMOVE(&req->queue);
+	queue_remove(&req->queue);
 
 	/* Finish the request */
 	uvPrepareConsume(uv, &req->fd, &req->counter);
@@ -199,8 +199,8 @@ static void uvPrepareAfterWorkCb(uv_work
 	/* If we are closing, let's discard the segment. All pending requests
 	 * have already being fired with RAFT_CANCELED. */
 	if (uv->closing) {
-		assert(QUEUE_IS_EMPTY(&uv->prepare_pool));
-		assert(QUEUE_IS_EMPTY(&uv->prepare_reqs));
+		assert(queue_empty(&uv->prepare_pool));
+		assert(queue_empty(&uv->prepare_reqs));
 		if (segment->status == 0) {
 			char errmsg[RAFT_ERRMSG_BUF_SIZE];
 			UvOsClose(segment->fd);
@@ -218,7 +218,7 @@ static void uvPrepareAfterWorkCb(uv_work
 	 * Note that if there's no pending request, we don't set the error
 	 * message, to avoid overwriting previous errors. */
 	if (segment->status != 0) {
-		if (!QUEUE_IS_EMPTY(&uv->prepare_reqs)) {
+		if (!queue_empty(&uv->prepare_reqs)) {
 			ErrMsgTransferf(segment->errmsg, uv->io->errmsg,
 					"create segment %s", segment->filename);
 			uvPrepareFinishAllRequests(uv, segment->status);
@@ -231,10 +231,10 @@ static void uvPrepareAfterWorkCb(uv_work
 	assert(segment->fd >= 0);
 
 	tracef("completed creation of %s", segment->filename);
-	QUEUE_PUSH(&uv->prepare_pool, &segment->queue);
+	queue_insert_tail(&uv->prepare_pool, &segment->queue);
 
 	/* Let's process any pending request. */
-	if (!QUEUE_IS_EMPTY(&uv->prepare_reqs)) {
+	if (!queue_empty(&uv->prepare_reqs)) {
 		uvPrepareFinishOldestRequest(uv);
 	}
 
@@ -249,7 +249,7 @@ static void uvPrepareAfterWorkCb(uv_work
 	 * above, thus reducing the pool size and making it smaller than the
 	 * target size. */
 	if (uvPrepareCount(uv) >= UV__TARGET_POOL_SIZE) {
-		assert(QUEUE_IS_EMPTY(&uv->prepare_reqs));
+		assert(queue_empty(&uv->prepare_reqs));
 		return;
 	}
 
@@ -284,7 +284,7 @@ int UvPrepare(struct uv *uv,
 
 	assert(!uv->closing);
 
-	if (!QUEUE_IS_EMPTY(&uv->prepare_pool)) {
+	if (!queue_empty(&uv->prepare_pool)) {
 		uvPrepareConsume(uv, fd, counter);
 		goto maybe_start;
 	}
@@ -292,7 +292,7 @@ int UvPrepare(struct uv *uv,
 	*fd = -1;
 	*counter = 0;
 	req->cb = cb;
-	QUEUE_PUSH(&uv->prepare_reqs, &req->queue);
+	queue_insert_tail(&uv->prepare_reqs, &req->queue);
 
 maybe_start:
 	/* If we are already creating a segment, let's just wait. */
@@ -311,7 +311,7 @@ err:
 	if (*fd != -1) {
 		uvPrepareDiscard(uv, *fd, *counter);
 	} else {
-		QUEUE_REMOVE(&req->queue);
+		queue_remove(&req->queue);
 	}
 	assert(rv != 0);
 	return rv;
@@ -325,12 +325,12 @@ void UvPrepareClose(struct uv *uv)
 	uvPrepareFinishAllRequests(uv, RAFT_CANCELED);
 
 	/* Remove any unused prepared segment. */
-	while (!QUEUE_IS_EMPTY(&uv->prepare_pool)) {
+	while (!queue_empty(&uv->prepare_pool)) {
 		queue *head;
 		struct uvIdleSegment *segment;
-		head = QUEUE_HEAD(&uv->prepare_pool);
+		head = queue_head(&uv->prepare_pool);
 		segment = QUEUE_DATA(head, struct uvIdleSegment, queue);
-		QUEUE_REMOVE(&segment->queue);
+		queue_remove(&segment->queue);
 		uvPrepareDiscard(uv, segment->fd, segment->counter);
 		RaftHeapFree(segment);
 	}
diff -pruN 1.16.4-4/src/raft/uv_recv.c 1.16.5-1/src/raft/uv_recv.c
--- 1.16.4-4/src/raft/uv_recv.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_recv.c	2024-06-28 13:17:26.000000000 +0000
@@ -78,13 +78,13 @@ static int uvServerInit(struct uvServer
 	s->message.type = 0;
 	s->payload.base = NULL;
 	s->payload.len = 0;
-	QUEUE_PUSH(&uv->servers, &s->queue);
+	queue_insert_tail(&uv->servers, &s->queue);
 	return 0;
 }
 
 static void uvServerDestroy(struct uvServer *s)
 {
-	QUEUE_REMOVE(&s->queue);
+	queue_remove(&s->queue);
 
 	if (s->header.base != NULL) {
 		/* This means we were interrupted while reading the header. */
@@ -179,8 +179,8 @@ static void uvServerStreamCloseCb(uv_han
 static void uvServerAbort(struct uvServer *s)
 {
 	struct uv *uv = s->uv;
-	QUEUE_REMOVE(&s->queue);
-	QUEUE_PUSH(&uv->aborting, &s->queue);
+	queue_remove(&s->queue);
+	queue_insert_tail(&uv->aborting, &s->queue);
 	uv_close((struct uv_handle_s *)s->stream, uvServerStreamCloseCb);
 }
 
@@ -290,11 +290,12 @@ static void uvServerReadCb(uv_stream_t *
 				case RAFT_IO_APPEND_ENTRIES:
 					payload.base = s->payload.base;
 					payload.len = s->payload.len;
-					uvDecodeEntriesBatch(
+					(void)uvDecodeEntriesBatch(
 					    payload.base, 0,
 					    s->message.append_entries.entries,
 					    s->message.append_entries
-						.n_entries);
+						.n_entries,
+					    false);
 					break;
 				case RAFT_IO_INSTALL_SNAPSHOT:
 					s->message.install_snapshot.data.base =
@@ -411,10 +412,10 @@ int UvRecvStart(struct uv *uv)
 
 void UvRecvClose(struct uv *uv)
 {
-	while (!QUEUE_IS_EMPTY(&uv->servers)) {
+	while (!queue_empty(&uv->servers)) {
 		queue *head;
 		struct uvServer *server;
-		head = QUEUE_HEAD(&uv->servers);
+		head = queue_head(&uv->servers);
 		server = QUEUE_DATA(head, struct uvServer, queue);
 		uvServerAbort(server);
 	}
diff -pruN 1.16.4-4/src/raft/uv_segment.c 1.16.5-1/src/raft/uv_segment.c
--- 1.16.4-4/src/raft/uv_segment.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_segment.c	2024-06-28 13:17:26.000000000 +0000
@@ -266,11 +266,11 @@ static int uvLoadEntriesBatch(struct uv
 
 	/* Consume the batch header, excluding the first 8 bytes containing the
 	 * number of entries, which we have already read. */
-	header.len = uvSizeofBatchHeader(n);
+	header.len = uvSizeofBatchHeader(n, true);
 	header.base = batch;
 
 	rv = uvConsumeContent(content, offset,
-			      uvSizeofBatchHeader(n) - sizeof(uint64_t), NULL,
+			      uvSizeofBatchHeader(n, true) - sizeof(uint64_t), NULL,
 			      errmsg);
 	if (rv != 0) {
 		ErrMsgTransfer(errmsg, uv->io->errmsg, "read header");
@@ -288,15 +288,20 @@ static int uvLoadEntriesBatch(struct uv
 	}
 
 	/* Decode the batch header, allocating the entries array. */
-	rv = uvDecodeBatchHeader(header.base, entries, n_entries);
+	uint64_t local_data_size = 0;
+	rv = uvDecodeBatchHeader(header.base, entries, n_entries, &local_data_size);
 	if (rv != 0) {
 		goto err;
 	}
 
-	/* Calculate the total size of the batch data */
+	/* Calculate the total size of the batch data. TODO this computation
+	 * should be rolled into the actual parsing part somehow. */
 	data.len = 0;
 	for (i = 0; i < n; i++) {
 		data.len += (*entries)[i].buf.len;
+#ifdef DQLITE_NEXT
+		data.len += sizeof((*entries)[i].local_data);
+#endif
 	}
 	data.base = (uint8_t *)content->base + *offset;
 
@@ -312,13 +317,17 @@ static int uvLoadEntriesBatch(struct uv
 	crc1 = byteFlip32(((uint32_t *)checksums)[1]);
 	crc2 = byteCrc32(data.base, data.len, 0);
 	if (crc1 != crc2) {
+		tracef("batch is bad");
 		ErrMsgPrintf(uv->io->errmsg, "data checksum mismatch");
 		rv = RAFT_CORRUPT;
 		goto err_after_header_decode;
 	}
 
-	uvDecodeEntriesBatch(content->base, *offset - data.len, *entries,
-			     *n_entries);
+	rv = uvDecodeEntriesBatch(content->base, *offset - data.len, *entries,
+			     *n_entries, local_data_size);
+	if (rv != 0) {
+		goto err_after_header_decode;
+	}
 
 	*last = *offset == content->len;
 
@@ -739,9 +748,12 @@ int uvSegmentBufferAppend(struct uvSegme
 	int rv;
 
 	size = sizeof(uint32_t) * 2;            /* CRC checksums */
-	size += uvSizeofBatchHeader(n_entries); /* Batch header */
+	size += uvSizeofBatchHeader(n_entries, true); /* Batch header */
 	for (i = 0; i < n_entries; i++) {       /* Entries data */
 		size += bytePad64(entries[i].buf.len);
+#ifdef DQLITE_NEXT
+		size += sizeof(struct raft_entry_local_data);
+#endif
 	}
 
 	rv = uvEnsureSegmentBufferIsLargeEnough(b, b->n + size);
@@ -758,9 +770,9 @@ int uvSegmentBufferAppend(struct uvSegme
 
 	/* Batch header */
 	header = cursor;
-	uvEncodeBatchHeader(entries, n_entries, cursor);
-	crc1 = byteCrc32(header, uvSizeofBatchHeader(n_entries), 0);
-	cursor = (uint8_t *)cursor + uvSizeofBatchHeader(n_entries);
+	uvEncodeBatchHeader(entries, n_entries, cursor, true /* encode local data */);
+	crc1 = byteCrc32(header, uvSizeofBatchHeader(n_entries, true), 0);
+	cursor = (uint8_t *)cursor + uvSizeofBatchHeader(n_entries, true);
 
 	/* Batch data */
 	crc2 = 0;
@@ -770,6 +782,14 @@ int uvSegmentBufferAppend(struct uvSegme
 		memcpy(cursor, entry->buf.base, entry->buf.len);
 		crc2 = byteCrc32(cursor, entry->buf.len, crc2);
 		cursor = (uint8_t *)cursor + entry->buf.len;
+		static_assert(sizeof(entry->local_data.buf) % sizeof(uint64_t) == 0,
+			      "bad size for entry local data");
+#ifdef DQLITE_NEXT
+		size_t local_data_size = sizeof(entry->local_data.buf);
+		memcpy(cursor, entry->local_data.buf, local_data_size);
+		crc2 = byteCrc32(cursor, local_data_size, crc2);
+		cursor = (uint8_t *)cursor + local_data_size;
+#endif
 	}
 
 	bytePut32(&crc1_p, crc1);
@@ -1010,7 +1030,7 @@ static int uvWriteClosedSegment(struct u
 	 * block */
 	cap = uv->block_size -
 	      (sizeof(uint64_t) /* Format version */ +
-	       sizeof(uint64_t) /* Checksums */ + uvSizeofBatchHeader(1));
+	       sizeof(uint64_t) /* Checksums */ + uvSizeofBatchHeader(1, true /* include local bufs */));
 	if (conf->len > cap) {
 		return RAFT_TOOBIG;
 	}
diff -pruN 1.16.4-4/src/raft/uv_send.c 1.16.5-1/src/raft/uv_send.c
--- 1.16.4-4/src/raft/uv_send.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_send.c	2024-06-28 13:17:26.000000000 +0000
@@ -95,9 +95,9 @@ static int uvClientInit(struct uvClient
 	rv = uv_timer_init(c->uv->loop, &c->timer);
 	assert(rv == 0);
 	strcpy(c->address, address);
-	QUEUE_INIT(&c->pending);
+	queue_init(&c->pending);
 	c->closing = false;
-	QUEUE_PUSH(&uv->clients, &c->queue);
+	queue_insert_tail(&uv->clients, &c->queue);
 	return 0;
 }
 
@@ -119,13 +119,13 @@ static void uvClientMaybeDestroy(struct
 		return;
 	}
 
-	while (!QUEUE_IS_EMPTY(&c->pending)) {
+	while (!queue_empty(&c->pending)) {
 		queue *head;
 		struct uvSend *send;
 		struct raft_io_send *req;
-		head = QUEUE_HEAD(&c->pending);
+		head = queue_head(&c->pending);
 		send = QUEUE_DATA(head, struct uvSend, queue);
-		QUEUE_REMOVE(head);
+		queue_remove(head);
 		req = send->req;
 		uvSendDestroy(send);
 		if (req->cb != NULL) {
@@ -133,7 +133,7 @@ static void uvClientMaybeDestroy(struct
 		}
 	}
 
-	QUEUE_REMOVE(&c->queue);
+	queue_remove(&c->queue);
 
 	assert(c->address != NULL);
 	RaftHeapFree(c->address);
@@ -210,7 +210,7 @@ static int uvClientSend(struct uvClient
 	/* If there's no connection available, let's queue the request. */
 	if (c->stream == NULL) {
 		tracef("no connection available -> enqueue message");
-		QUEUE_PUSH(&c->pending, &send->queue);
+		queue_insert_tail(&c->pending, &send->queue);
 		return 0;
 	}
 
@@ -234,12 +234,12 @@ static void uvClientSendPending(struct u
 	int rv;
 	assert(c->stream != NULL);
 	tracef("send pending messages");
-	while (!QUEUE_IS_EMPTY(&c->pending)) {
+	while (!queue_empty(&c->pending)) {
 		queue *head;
 		struct uvSend *send;
-		head = QUEUE_HEAD(&c->pending);
+		head = queue_head(&c->pending);
 		send = QUEUE_DATA(head, struct uvSend, queue);
-		QUEUE_REMOVE(head);
+		queue_remove(head);
 		rv = uvClientSend(c, send);
 		if (rv != 0) {
 			if (send->req->cb != NULL) {
@@ -322,9 +322,9 @@ static void uvClientConnectCb(struct raf
 			queue *head;
 			struct uvSend *old_send;
 			struct raft_io_send *old_req;
-			head = QUEUE_HEAD(&c->pending);
+			head = queue_head(&c->pending);
 			old_send = QUEUE_DATA(head, struct uvSend, queue);
-			QUEUE_REMOVE(head);
+			queue_remove(head);
 			old_req = old_send->req;
 			uvSendDestroy(old_send);
 			if (old_req->cb != NULL) {
@@ -384,8 +384,8 @@ static void uvClientAbort(struct uvClien
 	       uv_is_active((struct uv_handle_s *)&c->timer) ||
 	       c->connect.data != NULL);
 
-	QUEUE_REMOVE(&c->queue);
-	QUEUE_PUSH(&uv->aborting, &c->queue);
+	queue_remove(&c->queue);
+	queue_insert_tail(&uv->aborting, &c->queue);
 
 	rv = uv_timer_stop(&c->timer);
 	assert(rv == 0);
@@ -507,10 +507,10 @@ err:
 void UvSendClose(struct uv *uv)
 {
 	assert(uv->closing);
-	while (!QUEUE_IS_EMPTY(&uv->clients)) {
+	while (!queue_empty(&uv->clients)) {
 		queue *head;
 		struct uvClient *client;
-		head = QUEUE_HEAD(&uv->clients);
+		head = queue_head(&uv->clients);
 		client = QUEUE_DATA(head, struct uvClient, queue);
 		uvClientAbort(client);
 	}
diff -pruN 1.16.4-4/src/raft/uv_snapshot.c 1.16.5-1/src/raft/uv_snapshot.c
--- 1.16.4-4/src/raft/uv_snapshot.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_snapshot.c	2024-06-28 13:17:26.000000000 +0000
@@ -751,7 +751,7 @@ static void uvSnapshotGetAfterWorkCb(uv_
 	int req_status = get->status;
 	struct uv *uv = get->uv;
 	assert(status == 0);
-	QUEUE_REMOVE(&get->queue);
+	queue_remove(&get->queue);
 	RaftHeapFree(get);
 	req->cb(req, snapshot, req_status);
 	uvMaybeFireCloseCb(uv);
@@ -784,11 +784,11 @@ int UvSnapshotGet(struct raft_io *io,
 	}
 	get->work.data = get;
 
-	QUEUE_PUSH(&uv->snapshot_get_reqs, &get->queue);
+	queue_insert_tail(&uv->snapshot_get_reqs, &get->queue);
 	rv = uv_queue_work(uv->loop, &get->work, uvSnapshotGetWorkCb,
 			   uvSnapshotGetAfterWorkCb);
 	if (rv != 0) {
-		QUEUE_REMOVE(&get->queue);
+		queue_remove(&get->queue);
 		tracef("get last snapshot: %s", uv_strerror(rv));
 		rv = RAFT_IOERR;
 		goto err_after_snapshot_alloc;
diff -pruN 1.16.4-4/src/raft/uv_tcp.c 1.16.5-1/src/raft/uv_tcp.c
--- 1.16.4-4/src/raft/uv_tcp.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_tcp.c	2024-06-28 13:17:26.000000000 +0000
@@ -40,9 +40,9 @@ void UvTcpMaybeFireCloseCb(struct UvTcp
 		return;
 	}
 
-	assert(QUEUE_IS_EMPTY(&t->accepting));
-	assert(QUEUE_IS_EMPTY(&t->connecting));
-	if (!QUEUE_IS_EMPTY(&t->aborting)) {
+	assert(queue_empty(&t->accepting));
+	assert(queue_empty(&t->connecting));
+	if (!queue_empty(&t->aborting)) {
 		return;
 	}
 
@@ -82,9 +82,9 @@ int raft_uv_tcp_init(struct raft_uv_tran
 	t->listeners = NULL;
 	t->n_listeners = 0;
 	t->accept_cb = NULL;
-	QUEUE_INIT(&t->accepting);
-	QUEUE_INIT(&t->connecting);
-	QUEUE_INIT(&t->aborting);
+	queue_init(&t->accepting);
+	queue_init(&t->connecting);
+	queue_init(&t->aborting);
 	t->closing = false;
 	t->close_cb = NULL;
 
diff -pruN 1.16.4-4/src/raft/uv_tcp.h 1.16.5-1/src/raft/uv_tcp.h
--- 1.16.4-4/src/raft/uv_tcp.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_tcp.h	2024-06-28 13:17:26.000000000 +0000
@@ -2,7 +2,7 @@
 #define UV_TCP_H_
 
 #include "../raft.h"
-#include "queue.h"
+#include "../lib/queue.h"
 
 /* Protocol version. */
 #define UV__TCP_HANDSHAKE_PROTOCOL 1
diff -pruN 1.16.4-4/src/raft/uv_tcp_connect.c 1.16.5-1/src/raft/uv_tcp_connect.c
--- 1.16.4-4/src/raft/uv_tcp_connect.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_tcp_connect.c	2024-06-28 13:17:26.000000000 +0000
@@ -77,7 +77,7 @@ static void uvTcpConnectFinish(struct uv
 	struct uv_stream_s *stream = (struct uv_stream_s *)connect->tcp;
 	struct raft_uv_connect *req = connect->req;
 	int status = connect->status;
-	QUEUE_REMOVE(&connect->queue);
+	queue_remove(&connect->queue);
 	RaftHeapFree(connect->handshake.base);
 	uv_freeaddrinfo(connect->getaddrinfo.addrinfo);
 	raft_free(connect);
@@ -101,8 +101,8 @@ static void uvTcpConnectUvCloseCb(struct
 /* Abort a connection request. */
 static void uvTcpConnectAbort(struct uvTcpConnect *connect)
 {
-	QUEUE_REMOVE(&connect->queue);
-	QUEUE_PUSH(&connect->t->aborting, &connect->queue);
+	queue_remove(&connect->queue);
+	queue_insert_tail(&connect->t->aborting, &connect->queue);
 	uv_cancel((struct uv_req_s *)&connect->getaddrinfo);
 	/* Call uv_close on the tcp handle, if there is no getaddrinfo request
 	 * in flight and the handle is not currently closed due to next IP
@@ -261,7 +261,7 @@ static void uvGetAddrInfoCb(uv_getaddrin
 /* Create a new TCP handle and submit a connection request to the event loop. */
 static int uvTcpConnectStart(struct uvTcpConnect *r, const char *address)
 {
-	static struct addrinfo hints = {.ai_flags = AI_V4MAPPED | AI_ADDRCONFIG,
+	static struct addrinfo hints = {.ai_flags = 0,
 					.ai_family = AF_INET,
 					.ai_socktype = SOCK_STREAM,
 					.ai_protocol = 0};
@@ -353,7 +353,7 @@ int UvTcpConnect(struct raft_uv_transpor
 	req->cb = cb;
 
 	/* Keep track of the pending request */
-	QUEUE_PUSH(&t->connecting, &r->queue);
+	queue_insert_tail(&t->connecting, &r->queue);
 
 	/* Start connecting */
 	rv = uvTcpConnectStart(r, address);
@@ -364,7 +364,7 @@ int UvTcpConnect(struct raft_uv_transpor
 	return 0;
 
 err_after_alloc:
-	QUEUE_REMOVE(&r->queue);
+	queue_remove(&r->queue);
 	RaftHeapFree(r);
 err:
 	return rv;
@@ -372,10 +372,10 @@ err:
 
 void UvTcpConnectClose(struct UvTcp *t)
 {
-	while (!QUEUE_IS_EMPTY(&t->connecting)) {
+	while (!queue_empty(&t->connecting)) {
 		struct uvTcpConnect *connect;
 		queue *head;
-		head = QUEUE_HEAD(&t->connecting);
+		head = queue_head(&t->connecting);
 		connect = QUEUE_DATA(head, struct uvTcpConnect, queue);
 		uvTcpConnectAbort(connect);
 	}
diff -pruN 1.16.4-4/src/raft/uv_tcp_listen.c 1.16.5-1/src/raft/uv_tcp_listen.c
--- 1.16.4-4/src/raft/uv_tcp_listen.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_tcp_listen.c	2024-06-28 13:17:26.000000000 +0000
@@ -69,7 +69,7 @@ static void uvTcpIncomingCloseCb(struct
 {
 	struct uvTcpIncoming *incoming = handle->data;
 	struct UvTcp *t = incoming->t;
-	QUEUE_REMOVE(&incoming->queue);
+	queue_remove(&incoming->queue);
 	if (incoming->handshake.address.base != NULL) {
 		RaftHeapFree(incoming->handshake.address.base);
 	}
@@ -84,8 +84,8 @@ static void uvTcpIncomingAbort(struct uv
 	struct UvTcp *t = incoming->t;
 	/* After uv_close() returns we are guaranteed that no more alloc_cb or
 	 * read_cb will be called. */
-	QUEUE_REMOVE(&incoming->queue);
-	QUEUE_PUSH(&t->aborting, &incoming->queue);
+	queue_remove(&incoming->queue);
+	queue_insert_tail(&t->aborting, &incoming->queue);
 	uv_close((struct uv_handle_s *)incoming->tcp, uvTcpIncomingCloseCb);
 }
 
@@ -143,7 +143,7 @@ static void uvTcpIncomingReadCbAddress(u
 	assert(rv == 0);
 	id = byteFlip64(incoming->handshake.preamble[1]);
 	address = incoming->handshake.address.base;
-	QUEUE_REMOVE(&incoming->queue);
+	queue_remove(&incoming->queue);
 	incoming->t->accept_cb(incoming->t->transport, id, address,
 			       (struct uv_stream_s *)incoming->tcp);
 	RaftHeapFree(incoming->handshake.address.base);
@@ -274,7 +274,7 @@ static void uvTcpListenCb(struct uv_stre
 	incoming->listener = (struct uv_tcp_s *)stream;
 	incoming->tcp = NULL;
 
-	QUEUE_PUSH(&t->accepting, &incoming->queue);
+	queue_insert_tail(&t->accepting, &incoming->queue);
 
 	rv = uvTcpIncomingStart(incoming);
 	if (rv != 0) {
@@ -284,7 +284,7 @@ static void uvTcpListenCb(struct uv_stre
 	return;
 
 err_after_accept_alloc:
-	QUEUE_REMOVE(&incoming->queue);
+	queue_remove(&incoming->queue);
 	RaftHeapFree(incoming);
 err:
 	assert(rv != 0);
@@ -411,9 +411,9 @@ void UvTcpListenClose(struct UvTcp *t)
 	queue *head;
 	assert(t->closing);
 
-	while (!QUEUE_IS_EMPTY(&t->accepting)) {
+	while (!queue_empty(&t->accepting)) {
 		struct uvTcpIncoming *incoming;
-		head = QUEUE_HEAD(&t->accepting);
+		head = queue_head(&t->accepting);
 		incoming = QUEUE_DATA(head, struct uvTcpIncoming, queue);
 		uvTcpIncomingAbort(incoming);
 	}
diff -pruN 1.16.4-4/src/raft/uv_truncate.c 1.16.5-1/src/raft/uv_truncate.c
--- 1.16.4-4/src/raft/uv_truncate.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_truncate.c	2024-06-28 13:17:26.000000000 +0000
@@ -137,8 +137,8 @@ static void uvTruncateBarrierCb(struct U
 		return;
 	}
 
-	assert(QUEUE_IS_EMPTY(&uv->append_writing_reqs));
-	assert(QUEUE_IS_EMPTY(&uv->finalize_reqs));
+	assert(queue_empty(&uv->append_writing_reqs));
+	assert(queue_empty(&uv->finalize_reqs));
 	assert(uv->finalize_work.data == NULL);
 	assert(uv->truncate_work.data == NULL);
 
diff -pruN 1.16.4-4/src/raft/uv_work.c 1.16.5-1/src/raft/uv_work.c
--- 1.16.4-4/src/raft/uv_work.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_work.c	2024-06-28 13:17:26.000000000 +0000
@@ -28,7 +28,7 @@ static void uvAsyncAfterWorkCb(uv_work_t
 	struct uv *uv = w->uv;
 	assert(status == 0);
 
-	QUEUE_REMOVE(&w->queue);
+	queue_remove(&w->queue);
 	RaftHeapFree(w);
 	req->cb(req, req_status);
 	uvMaybeFireCloseCb(uv);
@@ -56,11 +56,11 @@ int UvAsyncWork(struct raft_io *io,
 	async_work->work.data = async_work;
 	req->cb = cb;
 
-	QUEUE_PUSH(&uv->async_work_reqs, &async_work->queue);
+	queue_insert_tail(&uv->async_work_reqs, &async_work->queue);
 	rv = uv_queue_work(uv->loop, &async_work->work, uvAsyncWorkCb,
 			   uvAsyncAfterWorkCb);
 	if (rv != 0) {
-		QUEUE_REMOVE(&async_work->queue);
+		queue_remove(&async_work->queue);
 		tracef("async work: %s", uv_strerror(rv));
 		rv = RAFT_IOERR;
 		goto err_after_req_alloc;
diff -pruN 1.16.4-4/src/raft/uv_writer.c 1.16.5-1/src/raft/uv_writer.c
--- 1.16.4-4/src/raft/uv_writer.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_writer.c	2024-06-28 13:17:26.000000000 +0000
@@ -33,7 +33,7 @@ static void uvWriterReqSetStatus(struct
  * callback if set. */
 static void uvWriterReqFinish(struct UvWriterReq *req)
 {
-	QUEUE_REMOVE(&req->queue);
+	queue_remove(&req->queue);
 	if (req->status != 0) {
 		uvWriterReqTransferErrMsg(req);
 	}
@@ -207,7 +207,7 @@ static void uvWriterPollCb(uv_poll_t *po
 				req->status = RAFT_IOERR;
 				goto finish;
 			}
-			return;
+			continue;
 		}
 
 		uvWriterReqSetStatus(req, (int)event->res);
@@ -219,10 +219,10 @@ static void uvWriterPollCb(uv_poll_t *po
 	return;
 
 fail_requests:
-	while (!QUEUE_IS_EMPTY(&w->poll_queue)) {
+	while (!queue_empty(&w->poll_queue)) {
 		queue *head;
 		struct UvWriterReq *req;
-		head = QUEUE_HEAD(&w->poll_queue);
+		head = queue_head(&w->poll_queue);
 		req = QUEUE_DATA(head, struct UvWriterReq, queue);
 		uvWriterReqSetStatus(req, status);
 		uvWriterReqFinish(req);
@@ -251,8 +251,8 @@ int UvWriterInit(struct UvWriter *w,
 	w->event_poller.data = NULL;
 	w->check.data = NULL;
 	w->close_cb = NULL;
-	QUEUE_INIT(&w->poll_queue);
-	QUEUE_INIT(&w->work_queue);
+	queue_init(&w->poll_queue);
+	queue_init(&w->work_queue);
 	w->closing = false;
 	w->errmsg = errmsg;
 
@@ -352,10 +352,10 @@ static void uvWriterPollerCloseCb(struct
 	w->event_poller.data = NULL;
 
 	/* Cancel all pending requests. */
-	while (!QUEUE_IS_EMPTY(&w->poll_queue)) {
+	while (!queue_empty(&w->poll_queue)) {
 		queue *head;
 		struct UvWriterReq *req;
-		head = QUEUE_HEAD(&w->poll_queue);
+		head = queue_head(&w->poll_queue);
 		req = QUEUE_DATA(head, struct UvWriterReq, queue);
 		assert(req->work.data == NULL);
 		req->status = RAFT_CANCELED;
@@ -382,7 +382,7 @@ static void uvWriterCheckCloseCb(struct
 static void uvWriterCheckCb(struct uv_check_s *check)
 {
 	struct UvWriter *w = check->data;
-	if (!QUEUE_IS_EMPTY(&w->work_queue)) {
+	if (!queue_empty(&w->work_queue)) {
 		return;
 	}
 	uv_close((struct uv_handle_s *)&w->check, uvWriterCheckCloseCb);
@@ -407,7 +407,7 @@ void UvWriterClose(struct UvWriter *w, U
 
 	/* If we have requests executing in the threadpool, we need to wait for
 	 * them. That's done in the check callback. */
-	if (!QUEUE_IS_EMPTY(&w->work_queue)) {
+	if (!queue_empty(&w->work_queue)) {
 		uv_check_start(&w->check, uvWriterCheckCb);
 	} else {
 		uv_close((struct uv_handle_s *)&w->check, uvWriterCheckCloseCb);
@@ -440,8 +440,8 @@ int UvWriterSubmit(struct UvWriter *w,
 	 *       writes, so ensure that we're getting write requests
 	 *       sequentially. */
 	if (w->n_events == 1) {
-		assert(QUEUE_IS_EMPTY(&w->poll_queue));
-		assert(QUEUE_IS_EMPTY(&w->work_queue));
+		assert(queue_empty(&w->poll_queue));
+		assert(queue_empty(&w->work_queue));
 	}
 
 	assert(w->fd >= 0);
@@ -490,7 +490,7 @@ int UvWriterSubmit(struct UvWriter *w,
 
 	/* Try to submit the write request asynchronously */
 	if (w->async) {
-		QUEUE_PUSH(&w->poll_queue, &req->queue);
+		queue_insert_tail(&w->poll_queue, &req->queue);
 		rv = UvOsIoSubmit(w->ctx, 1, &iocbs);
 
 		/* If no error occurred, we're done, the write request was
@@ -499,7 +499,7 @@ int UvWriterSubmit(struct UvWriter *w,
 			goto done;
 		}
 
-		QUEUE_REMOVE(&req->queue);
+		queue_remove(&req->queue);
 
 		/* Check the reason of the error. */
 		switch (rv) {
@@ -521,7 +521,7 @@ int UvWriterSubmit(struct UvWriter *w,
 
 	/* If we got here it means we need to run io_submit in the threadpool.
 	 */
-	QUEUE_PUSH(&w->work_queue, &req->queue);
+	queue_insert_tail(&w->work_queue, &req->queue);
 	req->work.data = req;
 	rv = uv_queue_work(w->loop, &req->work, uvWriterWorkCb,
 			   uvWriterAfterWorkCb);
@@ -529,7 +529,7 @@ int UvWriterSubmit(struct UvWriter *w,
 		/* UNTESTED: with the current libuv implementation this can't
 		 * fail. */
 		req->work.data = NULL;
-		QUEUE_REMOVE(&req->queue);
+		queue_remove(&req->queue);
 		UvOsErrMsg(w->errmsg, "uv_queue_work", rv);
 		rv = RAFT_IOERR;
 		goto err;
diff -pruN 1.16.4-4/src/raft/uv_writer.h 1.16.5-1/src/raft/uv_writer.h
--- 1.16.4-4/src/raft/uv_writer.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft/uv_writer.h	2024-06-28 13:17:26.000000000 +0000
@@ -6,7 +6,7 @@
 #include <stdbool.h>
 
 #include "err.h"
-#include "queue.h"
+#include "../lib/queue.h"
 #include "uv_os.h"
 
 /* Perform asynchronous writes to a single file. */
diff -pruN 1.16.4-4/src/raft.h 1.16.5-1/src/raft.h
--- 1.16.4-4/src/raft.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/raft.h	2024-06-28 13:17:26.000000000 +0000
@@ -1,4 +1,11 @@
-#ifndef RAFT_H
+#if defined(USE_SYSTEM_RAFT)
+
+#include <raft.h>
+#include <raft/uv.h>
+#include <raft/fixture.h>
+
+#elif !defined(RAFT_H)
+
 #define RAFT_H
 
 #include <stdarg.h>
@@ -9,6 +16,8 @@
 
 #include <uv.h>
 
+#include "lib/queue.h"
+
 #ifndef RAFT_API
 #define RAFT_API __attribute__((visibility("default")))
 #endif
@@ -189,6 +198,28 @@ enum {
 };
 
 /**
+ * A small fixed-size inline buffer that stores extra data for a raft_entry
+ * that is different for each node in the cluster.
+ *
+ * A leader initializes the local data for an entry before passing it into
+ * raft_apply. This local data is stored in the volatile raft log and also
+ * in the persistent raft log on the leader. AppendEntries messages sent by
+ * the leader never contain the local data for entries.
+ *
+ * When a follower accepts an AppendEntries request, it invokes a callback
+ * provided by the FSM to fill out the local data for each new entry before
+ * appending the entries to its log (volatile and persistent). This local
+ * data doesn't have to be the same as the local data that the leader computed.
+ *
+ * When starting up, a raft node reads the local data for each entry for its
+ * persistent log as part of populating the volatile log.
+ */
+struct raft_entry_local_data {
+	/* Must be the only member of this struct. */
+	uint8_t buf[16];
+};
+
+/**
  * A single entry in the raft log.
  *
  * An entry that originated from this raft instance while it was the leader
@@ -211,12 +242,27 @@ enum {
  *
  * This arrangement makes it possible to minimize the amount of memory-copying
  * when performing I/O.
+ *
+ * The @is_local field is set to `true` by a leader that appends an entry to its
+ * volatile log. It is set to `false` by a follower that copies an entry received
+ * via AppendEntries to its volatile log. It is not represented in the AppendEntries
+ * message or in the persistent log. This field can be used by the FSM's `apply`
+ * callback to handle a COMMAND entry differently depending on whether it
+ * originated locally.
+ *
+ * Note: The @local_data and @is_local fields do not exist when we use an external
+ * libraft, because the last separate release of libraft predates their addition.
+ * The ifdef at the very top of this file ensures that we use the system raft headers
+ * when we build against an external libraft, so there will be no ABI mismatch as
+ * a result of incompatible struct layouts.
  */
 struct raft_entry
 {
-	raft_term term;      /* Term in which the entry was created. */
-	unsigned short type; /* Type (FSM command, barrier, config change). */
+	raft_term term;         /* Term in which the entry was created. */
+	unsigned short type;    /* Type (FSM command, barrier, config change). */
+	bool is_local;          /* Placed here so it goes in the padding after @type. */
 	struct raft_buffer buf; /* Entry data. */
+	struct raft_entry_local_data local_data;
 	void *batch;            /* Batch that buf's memory points to, if any. */
 };
 
@@ -802,7 +848,7 @@ struct raft
 			raft_index
 			    round_index; /* Target of the current round. */
 			raft_time round_start; /* Start of current round. */
-			void *requests[2]; /* Outstanding client requests. */
+			queue requests; /* Outstanding client requests. */
 			uint32_t
 			    voter_contacts; /* Current number of voting nodes we
 					       are in contact with */
@@ -1029,7 +1075,7 @@ RAFT_API int raft_voter_contacts(struct
 	void *data;            \
 	int type;              \
 	raft_index index;      \
-	void *queue[2];        \
+	queue queue;           \
 	uint8_t req_id[16];    \
 	uint8_t client_id[16]; \
 	uint8_t unique_id[16]; \
@@ -1073,6 +1119,7 @@ struct raft_apply
 RAFT_API int raft_apply(struct raft *r,
 			struct raft_apply *req,
 			const struct raft_buffer bufs[],
+			const struct raft_entry_local_data local_data[],
 			const unsigned n,
 			raft_apply_cb cb);
 
diff -pruN 1.16.4-4/src/registry.c 1.16.5-1/src/registry.c
--- 1.16.4-4/src/registry.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/registry.c	2024-06-28 13:17:26.000000000 +0000
@@ -9,17 +9,17 @@
 void registry__init(struct registry *r, struct config *config)
 {
 	r->config = config;
-	QUEUE__INIT(&r->dbs);
+	queue_init(&r->dbs);
 }
 
 void registry__close(struct registry *r)
 {
-	while (!QUEUE__IS_EMPTY(&r->dbs)) {
+	while (!queue_empty(&r->dbs)) {
 		struct db *db;
 		queue *head;
-		head = QUEUE__HEAD(&r->dbs);
-		QUEUE__REMOVE(head);
-		db = QUEUE__DATA(head, struct db, queue);
+		head = queue_head(&r->dbs);
+		queue_remove(head);
+		db = QUEUE_DATA(head, struct db, queue);
 		db__close(db);
 		sqlite3_free(db);
 	}
@@ -28,9 +28,9 @@ void registry__close(struct registry *r)
 int registry__db_get(struct registry *r, const char *filename, struct db **db)
 {
 	queue *head;
-	QUEUE__FOREACH(head, &r->dbs)
+	QUEUE_FOREACH(head, &r->dbs)
 	{
-		*db = QUEUE__DATA(head, struct db, queue);
+		*db = QUEUE_DATA(head, struct db, queue);
 		if (strcmp((*db)->filename, filename) == 0) {
 			return 0;
 		}
@@ -40,6 +40,6 @@ int registry__db_get(struct registry *r,
 		return DQLITE_NOMEM;
 	}
 	db__init(*db, r->config, filename);
-	QUEUE__PUSH(&r->dbs, &(*db)->queue);
+	queue_insert_tail(&r->dbs, &(*db)->queue);
 	return 0;
 }
diff -pruN 1.16.4-4/src/roles.c 1.16.5-1/src/roles.c
--- 1.16.4-4/src/roles.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/roles.c	2024-06-28 13:17:26.000000000 +0000
@@ -183,13 +183,13 @@ static void startChange(struct dqlite_no
 	int role;
 	int rv;
 
-	if (QUEUE__IS_EMPTY(&d->roles_changes)) {
+	if (queue_empty(&d->roles_changes)) {
 		return;
 	}
 
-	head = QUEUE__HEAD(&d->roles_changes);
-	QUEUE__REMOVE(head);
-	rec = QUEUE__DATA(head, struct change_record, queue);
+	head = queue_head(&d->roles_changes);
+	queue_remove(head);
+	rec = QUEUE_DATA(head, struct change_record, queue);
 	id = rec->id;
 	role = rec->role;
 	raft_free(rec);
@@ -228,9 +228,9 @@ static void queueChange(uint64_t id, int
 
 	/* If we already queued a role change for this node, just update
 	 * that record instead of queueing a new one. */
-	QUEUE__FOREACH(head, &d->roles_changes)
+	QUEUE_FOREACH(head, &d->roles_changes)
 	{
-		rec = QUEUE__DATA(head, struct change_record, queue);
+		rec = QUEUE_DATA(head, struct change_record, queue);
 		if (rec->id == id) {
 			rec->role = role;
 			return;
@@ -243,7 +243,7 @@ static void queueChange(uint64_t id, int
 	}
 	rec->id = id;
 	rec->role = role;
-	QUEUE__PUSH(&d->roles_changes, &rec->queue);
+	queue_insert_tail(&d->roles_changes, &rec->queue);
 }
 
 void RolesComputeChanges(int voters,
@@ -666,7 +666,7 @@ void RolesAdjust(struct dqlite_node *d)
 	}
 	/* If a series of role adjustments is already in progress, don't kick
 	 * off another one. */
-	if (!QUEUE__IS_EMPTY(&d->roles_changes)) {
+	if (!queue_empty(&d->roles_changes)) {
 		return;
 	}
 	assert(d->running);
@@ -707,10 +707,10 @@ void RolesCancelPendingChanges(struct dq
 	queue *head;
 	struct change_record *rec;
 
-	while (!QUEUE__IS_EMPTY(&d->roles_changes)) {
-		head = QUEUE__HEAD(&d->roles_changes);
-		rec = QUEUE__DATA(head, struct change_record, queue);
-		QUEUE__REMOVE(head);
+	while (!queue_empty(&d->roles_changes)) {
+		head = queue_head(&d->roles_changes);
+		rec = QUEUE_DATA(head, struct change_record, queue);
+		queue_remove(head);
 		raft_free(rec);
 	}
 }
diff -pruN 1.16.4-4/src/server.c 1.16.5-1/src/server.c
--- 1.16.4-4/src/server.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/server.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,9 +1,12 @@
 #include "server.h"
 
 #include <errno.h>
+#include <sched.h>
 #include <stdlib.h>
+#include <sys/file.h>
 #include <sys/un.h>
 #include <time.h>
+#include <uv.h>
 
 #include "../include/dqlite.h"
 #include "client/protocol.h"
@@ -13,6 +16,7 @@
 #include "lib/addr.h"
 #include "lib/assert.h"
 #include "lib/fs.h"
+#include "lib/threadpool.h"
 #include "logger.h"
 #include "protocol.h"
 #include "roles.h"
@@ -40,9 +44,9 @@ static void state_cb(struct raft *r,
 
 	if (old_state == RAFT_LEADER && new_state != RAFT_LEADER) {
 		tracef("node %llu@%s: leadership lost", r->id, r->address);
-		QUEUE__FOREACH(head, &d->conns)
+		QUEUE_FOREACH(head, &d->conns)
 		{
-			conn = QUEUE__DATA(head, struct conn, queue);
+			conn = QUEUE_DATA(head, struct conn, queue);
 			gateway__leader_close(&conn->gateway,
 					      RAFT_LEADERSHIPLOST);
 		}
@@ -60,6 +64,7 @@ int dqlite__init(struct dqlite_node *d,
 	ssize_t count;
 
 	d->initialized = false;
+	d->lock_fd = -1;
 	memset(d->errmsg, 0, sizeof d->errmsg);
 
 	rv = snprintf(db_dir_path, sizeof db_dir_path, DATABASE_DIR_FMT, dir);
@@ -69,7 +74,7 @@ int dqlite__init(struct dqlite_node *d,
 		goto err;
 	}
 
-	rv = config__init(&d->config, id, address, db_dir_path);
+	rv = config__init(&d->config, id, address, dir, db_dir_path);
 	if (rv != 0) {
 		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE,
 			 "config__init(rv:%d)", rv);
@@ -81,6 +86,7 @@ int dqlite__init(struct dqlite_node *d,
 		goto err_after_config_init;
 	}
 	registry__init(&d->registry, &d->config);
+
 	rv = uv_loop_init(&d->loop);
 	if (rv != 0) {
 		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE,
@@ -88,10 +94,20 @@ int dqlite__init(struct dqlite_node *d,
 		rv = DQLITE_ERROR;
 		goto err_after_vfs_init;
 	}
-	rv = raftProxyInit(&d->raft_transport, &d->loop);
+#ifdef DQLITE_NEXT
+	rv = pool_init(&d->pool, &d->loop, d->config.pool_thread_count,
+		       POOL_QOS_PRIO_FAIR);
 	if (rv != 0) {
+		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE, "pool_init(): %s",
+			 uv_strerror(rv));
+		rv = DQLITE_ERROR;
 		goto err_after_loop_init;
 	}
+#endif
+	rv = raftProxyInit(&d->raft_transport, &d->loop);
+	if (rv != 0) {
+		goto err_after_pool_init;
+	}
 	rv = raft_uv_init(&d->raft_io, &d->loop, dir, &d->raft_transport);
 	if (rv != 0) {
 		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE,
@@ -110,7 +126,8 @@ int dqlite__init(struct dqlite_node *d,
 	if (rv != 0) {
 		snprintf(d->errmsg, DQLITE_ERRMSG_BUF_SIZE, "raft_init(): %s",
 			 raft_errmsg(&d->raft));
-		return DQLITE_ERROR;
+		rv = DQLITE_ERROR;
+		goto err;
 	}
 	/* TODO: expose these values through some API */
 	raft_set_election_timeout(&d->raft, 3000);
@@ -143,9 +160,9 @@ int dqlite__init(struct dqlite_node *d,
 		goto err_after_stopped_init;
 	}
 
-	QUEUE__INIT(&d->queue);
-	QUEUE__INIT(&d->conns);
-	QUEUE__INIT(&d->roles_changes);
+	queue_init(&d->queue);
+	queue_init(&d->conns);
+	queue_init(&d->roles_changes);
 	d->raft_state = RAFT_UNAVAILABLE;
 	d->running = false;
 	d->listener = NULL;
@@ -172,7 +189,12 @@ err_after_raft_io_init:
 	raft_uv_close(&d->raft_io);
 err_after_raft_transport_init:
 	raftProxyClose(&d->raft_transport);
+err_after_pool_init:
+#ifdef DQLITE_NEXT
+	pool_close(&d->pool);
+	pool_fini(&d->pool);
 err_after_loop_init:
+#endif
 	uv_loop_close(&d->loop);
 err_after_vfs_init:
 	VfsClose(&d->vfs);
@@ -199,6 +221,10 @@ void dqlite__close(struct dqlite_node *d
 	// TODO assert rv of uv_loop_close after fixing cleanup logic related to
 	// the TODO above referencing the cleanup logic without running the
 	// node. See https://github.com/canonical/dqlite/issues/504.
+
+#ifdef DQLITE_NEXT
+	pool_fini(&d->pool);
+#endif
 	uv_loop_close(&d->loop);
 	raftProxyClose(&d->raft_transport);
 	registry__close(&d->registry);
@@ -492,7 +518,7 @@ static void raftCloseCb(struct raft *raf
 
 static void destroy_conn(struct conn *conn)
 {
-	QUEUE__REMOVE(&conn->queue);
+	queue_remove(&conn->queue);
 	sqlite3_free(conn);
 }
 
@@ -533,6 +559,9 @@ static void stopCb(uv_async_t *stop)
 		tracef("not running or already stopped");
 		return;
 	}
+#ifdef DQLITE_NEXT
+	pool_close(&d->pool);
+#endif
 	if (d->role_management) {
 		rv = uv_timer_stop(&d->timer);
 		assert(rv == 0);
@@ -540,9 +569,9 @@ static void stopCb(uv_async_t *stop)
 	}
 	d->running = false;
 
-	QUEUE__FOREACH(head, &d->conns)
+	QUEUE_FOREACH(head, &d->conns)
 	{
-		conn = QUEUE__DATA(head, struct conn, queue);
+		conn = QUEUE_DATA(head, struct conn, queue);
 		conn__stop(conn);
 	}
 	raft_close(&d->raft, raftCloseCb);
@@ -649,7 +678,7 @@ static void listenCb(uv_stream_t *listen
 		goto err_after_conn_alloc;
 	}
 
-	QUEUE__PUSH(&t->conns, &conn->queue);
+	queue_insert_tail(&t->conns, &conn->queue);
 
 	return;
 
@@ -756,6 +785,12 @@ int dqlite_node_set_auto_recovery(dqlite
 	return 0;
 }
 
+int dqlite_node_set_pool_thread_count(dqlite_node *n, unsigned thread_count)
+{
+	n->config.pool_thread_count = thread_count;
+	return 0;
+}
+
 const char *dqlite_node_errmsg(dqlite_node *n)
 {
 	if (n != NULL) {
@@ -796,6 +831,32 @@ static bool taskReady(struct dqlite_node
 	return d->running;
 }
 
+#define LOCK_FILENAME "dqlite-lock"
+
+static int acquire_dir(const char *dir, int *fd_out)
+{
+	char path[PATH_MAX];
+	int fd;
+	int rv;
+
+	snprintf(path, sizeof(path), "%s/%s", dir, LOCK_FILENAME);
+	fd = open(path, O_RDWR|O_CREAT|O_CLOEXEC, S_IRUSR|S_IWUSR);
+	if (fd < 0) {
+		return DQLITE_ERROR;
+	}
+	rv = flock(fd, LOCK_EX|LOCK_NB);
+	if (rv != 0) {
+		return DQLITE_ERROR;
+	}
+	*fd_out = fd;
+	return 0;
+}
+
+static void release_dir(int fd)
+{
+	close(fd);
+}
+
 static int dqliteDatabaseDirSetup(dqlite_node *t)
 {
 	int rv;
@@ -804,14 +865,14 @@ static int dqliteDatabaseDirSetup(dqlite
 		return 0;
 	}
 
-	rv = FsEnsureDir(t->config.dir);
+	rv = FsEnsureDir(t->config.database_dir);
 	if (rv != 0) {
 		snprintf(t->errmsg, DQLITE_ERRMSG_BUF_SIZE,
 			 "Error creating database dir: %d", rv);
 		return rv;
 	}
 
-	rv = FsRemoveDirFiles(t->config.dir);
+	rv = FsRemoveDirFiles(t->config.database_dir);
 	if (rv != 0) {
 		snprintf(t->errmsg, DQLITE_ERRMSG_BUF_SIZE,
 			 "Error removing files in database dir: %d", rv);
@@ -834,27 +895,39 @@ int dqlite_node_start(dqlite_node *t)
 		goto err;
 	}
 
+	int lock_fd;
+	rv = acquire_dir(t->config.raft_dir, &lock_fd);
+	if (rv != 0) {
+		snprintf(t->errmsg, DQLITE_ERRMSG_BUF_SIZE,
+			 "couldn't lock the raft directory");
+		return rv;
+	}
+	t->lock_fd = lock_fd;
+
 	rv = maybeBootstrap(t, t->config.id, t->config.address);
 	if (rv != 0) {
 		tracef("bootstrap failed %d", rv);
-		goto err;
+		goto err_after_acquire_dir;
 	}
 
 	rv = pthread_create(&t->thread, 0, &taskStart, t);
 	if (rv != 0) {
 		tracef("pthread create failed %d", rv);
 		rv = DQLITE_ERROR;
-		goto err;
+		goto err_after_acquire_dir;
 	}
 
 	if (!taskReady(t)) {
 		tracef("!taskReady");
 		rv = DQLITE_ERROR;
-		goto err;
+		goto err_after_acquire_dir;
 	}
 
 	return 0;
 
+
+err_after_acquire_dir:
+	release_dir(t->lock_fd);
 err:
 	return rv;
 }
@@ -883,6 +956,8 @@ int dqlite_node_stop(dqlite_node *d)
 	rv = pthread_join(d->thread, &result);
 	assert(rv == 0);
 
+	release_dir(d->lock_fd);
+
 	return (int)((uintptr_t)result);
 }
 
@@ -971,12 +1046,20 @@ int dqlite_node_recover_ext(dqlite_node
 		};
 	}
 
+	int lock_fd;
+	rv = acquire_dir(n->config.raft_dir, &lock_fd);
+	if (rv != 0) {
+		goto out;
+	}
+
 	rv = raft_recover(&n->raft, &configuration);
 	if (rv != 0) {
 		rv = DQLITE_ERROR;
 		goto out;
 	}
 
+	release_dir(lock_fd);
+
 out:
 	raft_configuration_close(&configuration);
 	return rv;
@@ -1317,7 +1400,7 @@ int dqlite_server_set_auto_join(dqlite_s
 	 * connect to. Once we've done that, we immediately fetch a fresh list
 	 * of cluster members that includes ID and role information, and clear
 	 * away the temporary node store cache. */
-	struct client_node_info info = {0};
+	struct client_node_info info = { 0 };
 	unsigned i;
 
 	for (i = 0; i < n; i += 1) {
@@ -1585,6 +1668,11 @@ int dqlite_server_start(dqlite_server *s
 	struct client_context context;
 	int rv;
 
+	rv = sqlite3_threadsafe();
+	if (!(rv == 1 || rv == 2)) {
+		goto err;
+	}
+
 	if (server->started) {
 		goto err;
 	}
diff -pruN 1.16.4-4/src/server.h 1.16.5-1/src/server.h
--- 1.16.4-4/src/server.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/server.h	2024-06-28 13:17:26.000000000 +0000
@@ -9,6 +9,7 @@
 #include "config.h"
 #include "id.h"
 #include "lib/assert.h"
+#include "lib/threadpool.h"
 #include "logger.h"
 #include "raft.h"
 #include "registry.h"
@@ -18,15 +19,16 @@
 /**
  * A single dqlite server instance.
  */
-struct dqlite_node
-{
+struct dqlite_node {
 	bool initialized; /* dqlite__init succeeded */
 
+	int lock_fd;                             /* Locked file in the data directory */
 	pthread_t thread;                        /* Main run loop thread. */
 	struct config config;                    /* Config values */
 	struct sqlite3_vfs vfs;                  /* In-memory VFS */
 	struct registry registry;                /* Databases */
 	struct uv_loop_s loop;                   /* UV loop */
+	struct pool_s pool;                      /* Thread pool */
 	struct raft_uv_transport raft_transport; /* Raft libuv transport */
 	struct raft_io raft_io;                  /* libuv I/O */
 	struct raft_fsm raft_fsm;                /* dqlite FSM */
@@ -59,15 +61,13 @@ struct dqlite_node
 
 /* Dynamic array of node info objects. This is the in-memory representation of
  * the node store. */
-struct node_store_cache
-{
+struct node_store_cache {
 	struct client_node_info *nodes; /* owned */
 	unsigned len;
 	unsigned cap;
 };
 
-struct dqlite_server
-{
+struct dqlite_server {
 	/* Threading stuff: */
 	pthread_cond_t cond;
 	pthread_mutex_t mutex;
diff -pruN 1.16.4-4/src/tracing.c 1.16.5-1/src/tracing.c
--- 1.16.4-4/src/tracing.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/tracing.c	2024-06-28 13:17:26.000000000 +0000
@@ -79,7 +79,7 @@ static inline void tracerEmit(const char
 		tracerPidCached,
 
 		tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour,
-		tm.tm_min, tm.tm_sec, ts.tv_nsec,
+		tm.tm_min, tm.tm_sec, (unsigned long)ts.tv_nsec,
 
 		(unsigned)tid, tracerTraceLevelName(level), func,
 		tracerShortFileName(file), line, message);
diff -pruN 1.16.4-4/src/utils.h 1.16.5-1/src/utils.h
--- 1.16.4-4/src/utils.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/utils.h	2024-06-28 13:17:26.000000000 +0000
@@ -2,6 +2,7 @@
 #define DQLITE_UTILS_H_
 
 #include <assert.h>
+#include <stdbool.h>
 #include <stdint.h>
 
 /* Various utility functions and macros */
@@ -21,4 +22,8 @@
 #define POST(cond) assert((cond))
 #define ERGO(a, b) (!(a) || (b))
 
+static inline bool is_po2(unsigned long n) {
+	return n > 0 && (n & (n - 1)) == 0;
+}
+
 #endif /* DQLITE_UTILS_H_ */
diff -pruN 1.16.4-4/src/vfs.c 1.16.5-1/src/vfs.c
--- 1.16.4-4/src/vfs.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/src/vfs.c	2024-06-28 13:17:26.000000000 +0000
@@ -1681,6 +1681,7 @@ static int vfsFileShmLock(sqlite3_file *
 		/* When releasing the write lock, if we find a pending
 		 * uncommitted transaction then a rollback must have occurred.
 		 * In that case we delete the pending transaction. */
+		tracef("ROLLBACK TIME");
 		if (flags == (SQLITE_SHM_UNLOCK | SQLITE_SHM_EXCLUSIVE)) {
 			vfsWalRollbackIfUncommitted(wal);
 		}
diff -pruN 1.16.4-4/src/vfs2.c 1.16.5-1/src/vfs2.c
--- 1.16.4-4/src/vfs2.c	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/src/vfs2.c	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,1965 @@
+#include "vfs2.h"
+
+#include "lib/byte.h"
+#include "lib/queue.h"
+#include "lib/sm.h"
+#include "tracing.h"
+#include "utils.h"
+
+#include <pthread.h>
+#include <sqlite3.h>
+
+#include <assert.h>
+#include <errno.h>
+#include <limits.h>
+#include <stdbool.h>
+#include <stddef.h>
+#include <stdint.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
+#include <unistd.h>
+
+#define VFS2_WAL_FIXED_SUFFIX1 "-xwal1"
+#define VFS2_WAL_FIXED_SUFFIX2 "-xwal2"
+
+#define VFS2_WAL_INDEX_REGION_SIZE (1 << 15)
+#define VFS2_WAL_FRAME_HDR_SIZE 24
+
+#define VFS2_EXCLUSIVE UINT_MAX
+
+#define WAL_WRITE_LOCK 0
+#define WAL_CKPT_LOCK 1
+#define WAL_RECOVER_LOCK 2
+
+#define LE_MAGIC 0x377f0682
+#define BE_MAGIC 0x377f0683
+
+#define WAL_NREADER (SQLITE_SHM_NLOCK - 3)
+
+#define READ_MARK_UNUSED 0xffffffff
+
+static const uint32_t invalid_magic = 0x17171717;
+
+enum {
+	/* Entry is not yet open. */
+	WTX_CLOSED,
+	/* Next WAL write will be a header write, causing a WAL swap (WAL-cur is empty or fully checkpointed). */
+	WTX_EMPTY,
+	/* Non-leader, at least one transaction in WAL-cur is not committed. */
+	WTX_FOLLOWING,
+	/* Non-leader, all transactions in WAL-cur are committed (but at least one is not checkpointed). */
+	WTX_FLUSH,
+	/* Leader, all transactions in WAL-cur are committed (but at least one is not checkpointed). */
+	WTX_BASE,
+	/* Leader, transaction in progress. */
+	WTX_ACTIVE,
+	/* Leader, transaction committed by SQLite and hidden. */
+	WTX_HIDDEN,
+	/* Leader, transation committed by SQLite, hidden, and polled. */
+	WTX_POLLED
+};
+
+/*
+
+Diagram of the state machine (some transitions omitted when they would crowd the diagram even more):
+
++----------------------+  sqlite3_open             +------------------------------------------------------------------------------+
+|        CLOSED        | ------------------------> |                                  FOLLOWING                                   | <+
++----------------------+                           +------------------------------------------------------------------------------+  |
+  |                                                  ^                          ^                          |                         |
+  | sqlite3_open                                     | vfs2_apply_uncommitted   | vfs2_apply_uncommitted   | vfs2_{commit,unapply}   | vfs2_apply_uncommitted
+  v                                                  |                          |                          v                         |
++----------------------------------------------------------------------------+  |                        +------------------------+  |
+|                                                                            |  +----------------------- |                        |  |
+|                                   EMPTY                                    |                           |         FLUSH          |  |
+|                                                                            | <------------------------ |                        |  |
++----------------------------------------------------------------------------+   sqlite3_wal_checkpoint  +------------------------+  |
+  |                       ^                                                                                |                         |
+  | vfs2_commit_barrier   | sqlite3_wal_checkpoint                                                         |                         |
+  v                       |                                                                                |                         |
++----------------------------------------------------------------------------+  vfs2_commit_barrier        |                         |
+|                                    BASE                                    | <---------------------------+                         |
++----------------------------------------------------------------------------+                                                       |
+  |                       ^                          |                                                                               |
+  | sqlite3_step          | vfs2_unhide              +-------------------------------------------------------------------------------+
+  v                       |
++----------------------+  |
+|        ACTIVE        |  |
++----------------------+  |
+  |                       |
+  | COMMIT_PHASETWO       |
+  v                       |
++----------------------+  |
+|        HIDDEN        |  |
++----------------------+  |
+  |                       |
+  | vfs2_poll             |
+  v                       |
++----------------------+  |
+|        POLLED        | -+
++----------------------+
+
+*/
+
+static const struct sm_conf wtx_states[SM_STATES_MAX] = {
+	[WTX_CLOSED] = {
+		.flags = SM_INITIAL|SM_FINAL,
+		.name = "closed",
+		.allowed = BITS(WTX_EMPTY)|BITS(WTX_FOLLOWING)|BITS(WTX_FLUSH),
+	},
+	[WTX_EMPTY] = {
+		.flags = 0,
+		.name = "empty",
+		.allowed = BITS(WTX_FOLLOWING)|BITS(WTX_FLUSH)|BITS(WTX_ACTIVE)|BITS(WTX_CLOSED),
+	},
+	[WTX_FOLLOWING] = {
+		.flags = 0,
+		.name = "following",
+		.allowed = BITS(WTX_FOLLOWING)|BITS(WTX_FLUSH)|BITS(WTX_CLOSED),
+	},
+	[WTX_FLUSH] = {
+		.flags = 0,
+		.name = "flush",
+		.allowed = BITS(WTX_FOLLOWING)|BITS(WTX_FLUSH)|BITS(WTX_ACTIVE)|BITS(WTX_CLOSED),
+	},
+	[WTX_BASE] = {
+		.flags = 0,
+		.name = "base",
+		.allowed = BITS(WTX_FOLLOWING)|BITS(WTX_BASE)|BITS(WTX_ACTIVE)|BITS(WTX_EMPTY)|BITS(WTX_CLOSED),
+	},
+	[WTX_ACTIVE] = {
+		.flags = 0,
+		.name = "active",
+		.allowed = BITS(WTX_BASE)|BITS(WTX_ACTIVE)|BITS(WTX_HIDDEN)|BITS(WTX_CLOSED),
+	},
+	[WTX_HIDDEN] = {
+		.flags = 0,
+		.name = "hidden",
+		.allowed = BITS(WTX_BASE)|BITS(WTX_POLLED)|BITS(WTX_CLOSED),
+	},
+	[WTX_POLLED] = {
+		.flags = 0,
+		.name = "polled",
+		.allowed = BITS(WTX_BASE)|BITS(WTX_CLOSED),
+	},
+};
+
+/**
+ * Userdata owned by the VFS.
+ */
+struct common {
+	sqlite3_vfs *orig;       /* underlying VFS */
+	pthread_rwlock_t rwlock; /* protects the queue */
+	queue queue; /* queue of entry */
+};
+
+struct cksums {
+	uint32_t cksum1;
+	uint32_t cksum2;
+};
+
+static bool is_bigendian(void)
+{
+	int x = 1;
+	return *(char *)(&x) == 0;
+}
+
+static uint32_t native_magic(void)
+{
+	return is_bigendian() ? BE_MAGIC : LE_MAGIC;
+}
+
+static void update_cksums(uint32_t magic, const uint8_t *p, size_t len, struct cksums *sums)
+{
+	PRE(magic == BE_MAGIC || magic == LE_MAGIC);
+	PRE(len % 8 == 0);
+	const uint8_t *end = p + len;
+	for (; p != end; p += 8) {
+		if (magic == BE_MAGIC) {
+			sums->cksum1 += ByteGetBe32(p) + sums->cksum2;
+			sums->cksum2 += ByteGetBe32(p + 4) + sums->cksum1;
+		} else {
+			sums->cksum1 += ByteGetLe32(p) + sums->cksum2;
+			sums->cksum2 += ByteGetLe32(p + 4) + sums->cksum1;
+		}
+	}
+}
+
+static bool cksums_equal(struct cksums a, struct cksums b)
+{
+	return a.cksum1 == b.cksum1 && a.cksum2 == b.cksum2;
+}
+
+/**
+ * Layout-compatible with the first part of the WAL index header.
+ *
+ * Note: everything is native-endian except the salts, hence the use of
+ * native integer types here.
+ */
+struct wal_index_basic_hdr {
+	uint32_t iVersion;
+	uint8_t unused[4];
+	uint32_t iChange;
+	uint8_t isInit;
+	uint8_t bigEndCksum;
+	uint16_t szPage;
+	uint32_t mxFrame;
+	uint32_t nPage;
+	struct cksums frame_cksums;
+	struct vfs2_salts salts;
+	struct cksums cksums;
+};
+
+struct wal_hdr {
+	uint8_t magic[4];
+	uint8_t version[4];
+	uint8_t page_size[4];
+	uint8_t ckpoint_seqno[4];
+	struct vfs2_salts salts;
+	uint8_t cksum1[4];
+	uint8_t cksum2[4];
+};
+
+struct wal_frame_hdr {
+	uint8_t page_number[4];
+	uint8_t commit[4];
+	struct vfs2_salts salts;
+	uint8_t cksum1[4];
+	uint8_t cksum2[4];
+};
+
+struct wal_index_full_hdr {
+	struct wal_index_basic_hdr basic[2];
+	uint32_t nBackfill;
+	uint32_t marks[WAL_NREADER];
+	uint8_t locks[SQLITE_SHM_NLOCK];
+	uint32_t nBackfillAttempted;
+	uint8_t unused[4];
+};
+
+/**
+ * View of the zeroth shm region, which contains the WAL index header.
+ */
+union vfs2_shm_region0 {
+	struct wal_index_full_hdr hdr;
+	char bytes[VFS2_WAL_INDEX_REGION_SIZE];
+};
+
+struct entry {
+	/* Next/prev entries for this VFS. */
+	queue link;
+
+	/* e.g. /path/to/some.db */
+	char *main_db_name;
+
+	/* The WALs are represented by two physical files (inodes)
+	 * and three filenames. For each of the two physical files
+	 * there is a "fixed name" that always points to that file.
+	 * The "moving name" always points to one of the two physical
+	 * files, but switches between them on every WAL swap. */
+
+	/* e.g. /path/to/some.db-wal */
+	char *wal_moving_name;
+	/* e.g. /path/to/some.db-xwal1 */
+	char *wal_cur_fixed_name;
+	/* Base VFS file object for WAL-cur */
+	sqlite3_file *wal_cur;
+	/* e.g. /path/to/some.db-xwal2 */
+	char *wal_prev_fixed_name;
+	/* Base VFS file object for WAL-prev */
+	sqlite3_file *wal_prev;
+
+	/* Number of `struct file` with SQLITE_OPEN_MAIN_DB that point to this entry */
+	unsigned refcount_main_db;
+	/* Number of `struct file` with SQLITE_OPEN_WAL that point to this entry */
+	unsigned refcount_wal;
+
+	/* if WAL-cur is nonempty at startup, we read its header, verify the checkum,
+	 * and use it to initialize the page size. otherwise, we wait until the first
+	 * write to the WAL, which should be the header */
+	uint32_t page_size;
+
+	/* For ACTIVE, HIDDEN, POLLED: the header that hides the pending txn */
+	struct wal_index_basic_hdr prev_txn_hdr;
+	/* For ACTIVE, HIDDEN, POLLED: the header that shows the pending txn */
+	struct wal_index_basic_hdr pending_txn_hdr;
+
+	/* shm implementation; holds the WAL index */
+	void **shm_regions;
+	int shm_regions_len;
+	unsigned shm_refcount;
+	/* Zero for unlocked, positive for read-locked, UINT_MAX for write-locked */
+	unsigned shm_locks[SQLITE_SHM_NLOCK];
+
+	/* For ACTIVE, HIDDEN: the pending txn. start and len
+	 * are in units of frames. */
+	struct vfs2_wal_frame *pending_txn_frames;
+	uint32_t pending_txn_start;
+	uint32_t pending_txn_len;
+	uint32_t pending_txn_last_frame_commit;
+
+	/* Frame index, points to the physical end of WAL-cur */
+	uint32_t wal_cursor;
+	/* Cached header of WAL-cur */
+	struct wal_hdr wal_cur_hdr;
+	/* Cached header of WAL-prev */
+	struct wal_hdr wal_prev_hdr;
+
+	struct sm wtx_sm;
+	/* VFS-wide data (immutable) */
+	struct common *common;
+};
+
+/**
+ * VFS-specific file object, upcastable to sqlite3_file.
+ */
+struct file {
+	/* Our custom sqlite3_io_methods vtable; must come first. Always
+	 * present. */
+	struct sqlite3_file base;
+	/* Flags passed to the xOpen that created this file. */
+	int flags;
+	/* File object created by the base (unix) VFS. Not used for WAL. */
+	sqlite3_file *orig;
+	/* Common data between main DB and WAL. Not used for other kinds of
+	 * file. */
+	struct entry *entry;
+};
+
+static void free_pending_txn(struct entry *e)
+{
+	if (e->pending_txn_frames != NULL) {
+		for (uint32_t i = 0; i < e->pending_txn_len; i++) {
+			sqlite3_free(e->pending_txn_frames[i].page);
+		}
+		sqlite3_free(e->pending_txn_frames);
+	}
+	e->pending_txn_frames = 0;
+	e->pending_txn_len = 0;
+	e->pending_txn_last_frame_commit = 0;
+}
+
+static uint32_t get_salt1(struct vfs2_salts s)
+{
+	return ByteGetBe32(s.salt1);
+}
+
+static uint32_t get_salt2(struct vfs2_salts s)
+{
+	return ByteGetBe32(s.salt2);
+}
+
+static bool salts_equal(struct vfs2_salts a, struct vfs2_salts b)
+{
+	return get_salt1(a) == get_salt1(b) && get_salt2(a) == get_salt2(b);
+}
+
+static struct wal_index_full_hdr *get_full_hdr(struct entry *e)
+{
+	PRE(e->shm_regions_len > 0);
+	PRE(e->shm_regions != NULL);
+	return e->shm_regions[0];
+}
+
+static bool no_pending_txn(const struct entry *e)
+{
+	return e->pending_txn_len == 0 && e->pending_txn_frames == NULL && e->pending_txn_last_frame_commit == 0;
+}
+
+static bool write_lock_held(const struct entry *e)
+{
+	return e->shm_locks[WAL_WRITE_LOCK] == VFS2_EXCLUSIVE;
+}
+
+static bool wal_index_basic_hdr_equal(struct wal_index_basic_hdr a,
+				      struct wal_index_basic_hdr b)
+{
+	return memcmp(&a, &b, sizeof(struct wal_index_basic_hdr)) == 0;
+}
+
+static bool wal_index_basic_hdr_zeroed(struct wal_index_basic_hdr h)
+{
+	return wal_index_basic_hdr_equal(h, (struct wal_index_basic_hdr){});
+}
+
+static bool wal_index_basic_hdr_advanced(struct wal_index_basic_hdr new,
+					 struct wal_index_basic_hdr old)
+{
+	return new.iChange == old.iChange + 1 &&
+	       new.nPage >= old.nPage /* no vacuums here */
+	       && ((get_salt1(new.salts) == get_salt1(old.salts) &&
+		    get_salt2(new.salts) == get_salt2(old.salts)) ||
+		/* note the weirdness with zero salts */
+		   (get_salt1(old.salts) == 0 &&
+		    get_salt2(old.salts) == 0))
+	       && new.mxFrame > old.mxFrame;
+}
+
+/* Check that the hash tables in the WAL index have been initialized
+ * by looking for nonzero bytes after the WAL index header. (TODO:
+ * actually parse the hash tables?) */
+static bool wal_index_recovered(const struct entry *e)
+{
+	PRE(e->shm_regions_len > 0);
+	char *p = e->shm_regions[0];
+	for (size_t i = sizeof(struct wal_index_full_hdr); i < VFS2_WAL_INDEX_REGION_SIZE; i++) {
+		if (p[i] != 0) {
+			return true;
+		}
+	}
+	return false;
+}
+
+static bool is_valid_page_size(unsigned long n)
+{
+	return n >= 1 << 9 && n <= 1 << 16 && is_po2(n);
+}
+
+static bool is_open(const struct entry *e)
+{
+	return e->main_db_name != NULL
+		&& e->wal_moving_name != NULL
+		&& e->wal_cur_fixed_name != NULL
+		&& e->wal_cur != NULL
+		&& e->wal_prev_fixed_name != NULL
+		&& e->wal_prev != NULL
+		&& (e->refcount_main_db > 0 || e->refcount_wal > 0)
+		&& e->shm_regions != NULL
+		&& e->shm_regions_len > 0
+		&& e->shm_regions[0] != NULL
+		&& e->common != NULL;
+}
+
+static bool basic_hdr_valid(struct wal_index_basic_hdr bhdr)
+{
+	struct cksums sums = {};
+	update_cksums(bhdr.bigEndCksum ? BE_MAGIC : LE_MAGIC, (uint8_t *)&bhdr, offsetof(struct wal_index_basic_hdr, cksums), &sums);
+	return bhdr.iVersion == 3007000
+		&& bhdr.isInit == 1
+		&& cksums_equal(sums, bhdr.cksums);
+}
+
+static bool full_hdr_valid(const struct wal_index_full_hdr *ihdr)
+{
+	return basic_hdr_valid(ihdr->basic[0]) && wal_index_basic_hdr_equal(ihdr->basic[0], ihdr->basic[1]);
+}
+
+static bool wtx_invariant(const struct sm *sm, int prev)
+{
+	/* TODO make use of this */
+	(void)prev;
+
+	struct entry *e = CONTAINER_OF(sm, struct entry, wtx_sm);
+
+	if (sm_state(sm) == WTX_CLOSED) {
+		char *region = (char *)e;
+		char zeroed[offsetof(struct entry, wtx_sm)] = {};
+
+		return CHECK(memcmp(region, zeroed, sizeof(zeroed)) == 0) &&
+			CHECK(e->common != NULL);
+	}
+
+	if (!CHECK(is_open(e))) {
+		return false;
+	}
+	struct wal_index_full_hdr *ihdr = get_full_hdr(e);
+	if (!CHECK(full_hdr_valid(ihdr))) {
+		return false;
+	}
+	uint32_t mx = ihdr->basic[0].mxFrame;
+	uint32_t backfill = ihdr->nBackfill;
+	uint32_t cursor = e->wal_cursor;
+	if (!CHECK(backfill <= mx) || !CHECK(mx <= cursor)) {
+		return false;
+	}
+
+	/* TODO any checks applicable to the read marks and read locks? */
+
+	if (sm_state(sm) == WTX_EMPTY) {
+		return CHECK(mx == backfill) &&
+			CHECK(mx == cursor) &&
+			CHECK(no_pending_txn(e)) &&
+			CHECK(!write_lock_held(e));
+	}
+
+	if (!CHECK(is_valid_page_size(e->page_size))) {
+		return false;
+	}
+
+	if (sm_state(sm) == WTX_FOLLOWING) {
+		return CHECK(no_pending_txn(e)) &&
+			CHECK(write_lock_held(e)) &&
+			CHECK(mx < cursor);
+	}
+
+	if (sm_state(sm) == WTX_FLUSH) {
+		return CHECK(no_pending_txn(e)) &&
+			CHECK(!write_lock_held(e)) &&
+			CHECK(ERGO(mx > 0, backfill < mx)) &&
+			CHECK(mx == cursor);
+	}
+
+	if (sm_state(sm) == WTX_BASE) {
+		return CHECK(no_pending_txn(e)) &&
+			CHECK(!write_lock_held(e)) &&
+			CHECK(ERGO(mx > 0, backfill < mx)) &&
+			CHECK(mx == cursor) &&
+			CHECK(ERGO(mx > 0, wal_index_recovered(e)));
+	}
+
+	if (sm_state(sm) == WTX_ACTIVE) {
+		return CHECK(wal_index_basic_hdr_equal(get_full_hdr(e)->basic[0], e->prev_txn_hdr)) &&
+			CHECK(wal_index_basic_hdr_zeroed(e->pending_txn_hdr)) &&
+			CHECK(write_lock_held(e));
+	}
+
+	if (!CHECK(mx < cursor) ||
+			!CHECK(e->pending_txn_len > 0) ||
+			!CHECK(e->pending_txn_start + e->pending_txn_len == e->wal_cursor)) {
+		return false;
+	}
+
+	if (sm_state(sm) == WTX_HIDDEN) {
+		bool res = CHECK(wal_index_basic_hdr_equal(get_full_hdr(e)->basic[0], e->prev_txn_hdr)) &&
+			CHECK(wal_index_basic_hdr_advanced(e->pending_txn_hdr, e->prev_txn_hdr)) &&
+			CHECK(!write_lock_held(e)) &&
+			CHECK(e->pending_txn_frames != NULL);
+		if (!res) {
+			return false;
+		}
+		for (uint32_t i = 0; i < e->pending_txn_len; i++) {
+			res &= CHECK(e->pending_txn_frames[i].page != NULL);
+		}
+		return res;
+	}
+
+	if (sm_state(sm) == WTX_POLLED) {
+		return CHECK(wal_index_basic_hdr_equal(get_full_hdr(e)->basic[0], e->prev_txn_hdr)) &&
+			CHECK(wal_index_basic_hdr_advanced(e->pending_txn_hdr, e->prev_txn_hdr)) &&
+			CHECK(write_lock_held(e)) &&
+			CHECK(e->pending_txn_frames == NULL);
+	}
+
+	assert(0);
+}
+
+static int check_wal_integrity(sqlite3_file *f)
+{
+	/* TODO */
+	(void)f;
+	return SQLITE_OK;
+}
+
+/* sqlite3_io_methods implementations begin here */
+
+static sqlite3_file *get_orig(struct file *f)
+{
+	return (f->flags & SQLITE_OPEN_WAL) ? f->entry->wal_cur : f->orig;
+}
+
+static void maybe_close_entry(struct entry *e)
+{
+	if (e->refcount_main_db > 0 || e->refcount_wal > 0) {
+		return;
+	}
+
+	sqlite3_free(e->main_db_name);
+	sqlite3_free(e->wal_moving_name);
+	sqlite3_free(e->wal_cur_fixed_name);
+	if (e->wal_cur->pMethods != NULL) {
+		e->wal_cur->pMethods->xClose(e->wal_cur);
+	}
+	sqlite3_free(e->wal_cur);
+	sqlite3_free(e->wal_prev_fixed_name);
+	if (e->wal_prev->pMethods != NULL) {
+		e->wal_prev->pMethods->xClose(e->wal_prev);
+	}
+	sqlite3_free(e->wal_prev);
+
+	free_pending_txn(e);
+
+	pthread_rwlock_wrlock(&e->common->rwlock);
+	queue_remove(&e->link);
+	pthread_rwlock_unlock(&e->common->rwlock);
+	sqlite3_free(e);
+
+}
+
+static int vfs2_close(sqlite3_file *file)
+{
+	struct file *xfile = (struct file *)file;
+	int rv;
+
+	rv = SQLITE_OK;
+	if (xfile->flags & SQLITE_OPEN_MAIN_DB) {
+		if (xfile->orig->pMethods != NULL) {
+			rv = xfile->orig->pMethods->xClose(xfile->orig);
+		}
+		sqlite3_free(xfile->orig);
+		xfile->entry->refcount_main_db -= 1;
+		maybe_close_entry(xfile->entry);
+	} else if (xfile->flags & SQLITE_OPEN_WAL) {
+		xfile->entry->refcount_wal -= 1;
+		maybe_close_entry(xfile->entry);
+	} else if (xfile->orig->pMethods != NULL) {
+		rv = xfile->orig->pMethods->xClose(xfile->orig);
+		sqlite3_free(xfile->orig);
+	}
+	return rv;
+}
+
+static int vfs2_read(sqlite3_file *file, void *buf, int amt, sqlite3_int64 ofst)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xRead(orig, buf, amt, ofst);
+}
+
+static int wal_swap(struct entry *e,
+		    const struct wal_hdr *wal_hdr)
+{
+	PRE(e->pending_txn_len == 0);
+	PRE(e->pending_txn_frames == NULL);
+	int rv;
+
+	e->page_size = ByteGetBe32(wal_hdr->page_size);
+
+	/* Terminology: the outgoing WAL is the one that's moving
+	 * from cur to prev. The incoming WAL is the one that's moving
+	 * from prev to cur. */
+	sqlite3_file *phys_outgoing = e->wal_cur;
+	char *name_outgoing = e->wal_cur_fixed_name;
+	sqlite3_file *phys_incoming = e->wal_prev;
+	char *name_incoming = e->wal_prev_fixed_name;
+
+	tracef("wal swap outgoing=%s incoming=%s", name_outgoing, name_incoming);
+
+	/* Write the new header of the incoming WAL. */
+	rv = phys_incoming->pMethods->xWrite(phys_incoming, wal_hdr,
+					     sizeof(struct wal_hdr), 0);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+
+	/* In-memory WAL swap. */
+	e->wal_cur = phys_incoming;
+	e->wal_cur_fixed_name = name_incoming;
+	e->wal_prev = phys_outgoing;
+	e->wal_prev_fixed_name = name_outgoing;
+	e->wal_cursor = 0;
+	e->wal_prev_hdr = e->wal_cur_hdr;
+	e->wal_cur_hdr = *wal_hdr;
+
+	/* Move the moving name. */
+	rv = unlink(e->wal_moving_name);
+	if (rv != 0 && errno != ENOENT) {
+		tracef("unlink = IOERR");
+		return SQLITE_IOERR;
+	}
+	rv = link(name_incoming, e->wal_moving_name);
+	if (rv != 0) {
+		tracef("link = IOERR");
+		return SQLITE_IOERR;
+	}
+
+	/* TODO do we need an fsync here? */
+
+	/* Best-effort: invalidate the outgoing physical WAL so that nobody gets
+	 * confused. */
+	(void)phys_outgoing->pMethods->xWrite(phys_outgoing, &invalid_magic,
+					      sizeof(invalid_magic), 0);
+	return SQLITE_OK;
+}
+
+static int vfs2_wal_write_frame_hdr(struct entry *e,
+				    const struct wal_frame_hdr *fhdr,
+				    uint32_t x)
+{
+	struct vfs2_wal_frame *frames = e->pending_txn_frames;
+	if (no_pending_txn(e)) {
+		assert(x == e->wal_cursor);
+		e->pending_txn_start = x;
+	}
+	uint32_t n = e->pending_txn_len;
+	tracef("orig=%u start=%u n=%u", x, e->pending_txn_start, n);
+	x -= e->pending_txn_start;
+	assert(x <= n);
+	if (e->pending_txn_len == 0 && x == 0) {
+		/* check that the WAL-index hdr makes sense and save it */
+		struct wal_index_basic_hdr hdr = get_full_hdr(e)->basic[0];
+		assert(hdr.isInit == 1);
+		assert(hdr.mxFrame == e->pending_txn_start);
+		e->prev_txn_hdr = hdr;
+	}
+	if (x == n) {
+		/* FIXME reallocating every time seems bad */
+		sqlite3_uint64 z = (sqlite3_uint64)sizeof(*frames) * (sqlite3_uint64)(n + 1);
+		e->pending_txn_frames = sqlite3_realloc64(frames, z);
+		if (e->pending_txn_frames == NULL) {
+			return SQLITE_NOMEM;
+		}
+		struct vfs2_wal_frame *frame = &e->pending_txn_frames[n];
+		uint32_t commit = ByteGetBe32(fhdr->commit);
+		frame->page_number = ByteGetBe32(fhdr->page_number);
+		frame->commit = commit;
+		frame->page = NULL;
+		e->pending_txn_last_frame_commit = commit;
+		e->pending_txn_len++;
+	} else {
+		/* Overwriting a previously-written frame in the current
+		 * transaction. */
+		struct vfs2_wal_frame *frame = &e->pending_txn_frames[x];
+		frame->page_number = ByteGetBe32(fhdr->page_number);
+		frame->commit = ByteGetBe32(fhdr->commit);
+		sqlite3_free(frame->page);
+		frame->page = NULL;
+	}
+	sm_move(&e->wtx_sm, WTX_ACTIVE);
+	return SQLITE_OK;
+}
+
+static int vfs2_wal_post_write(struct entry *e,
+			       const void *buf,
+			       int amt,
+			       sqlite3_int64 ofst)
+{
+	uint32_t frame_size = VFS2_WAL_FRAME_HDR_SIZE + e->page_size;
+	if (amt == VFS2_WAL_FRAME_HDR_SIZE) {
+		ofst -= (sqlite3_int64)sizeof(struct wal_hdr);
+		assert(ofst % frame_size == 0);
+		sqlite3_int64 frame_ofst = ofst / (sqlite3_int64)frame_size;
+		return vfs2_wal_write_frame_hdr(e, buf, (uint32_t)frame_ofst);
+	} else if (amt == (int)e->page_size) {
+		sqlite3_int64 x = ofst - VFS2_WAL_FRAME_HDR_SIZE -
+				  (sqlite3_int64)sizeof(struct wal_hdr);
+		assert(x % frame_size == 0);
+		x /= frame_size;
+		x -= e->pending_txn_start;
+		assert(0 <= x && x < e->pending_txn_len);
+		struct vfs2_wal_frame *frame = &e->pending_txn_frames[x];
+		assert(frame->page == NULL);
+		frame->page = sqlite3_malloc(amt);
+		if (frame->page == NULL) {
+			return SQLITE_NOMEM;
+		}
+		memcpy(frame->page, buf, (size_t)amt);
+		sm_move(&e->wtx_sm, WTX_ACTIVE);
+		return SQLITE_OK;
+	} else {
+		assert(0);
+	}
+}
+
+static int vfs2_write(sqlite3_file *file,
+		      const void *buf,
+		      int amt,
+		      sqlite3_int64 ofst)
+{
+	struct file *xfile = (struct file *)file;
+	int rv;
+
+	if ((xfile->flags & SQLITE_OPEN_WAL) && ofst == 0) {
+		assert(amt == sizeof(struct wal_hdr));
+		const struct wal_hdr *hdr = buf;
+		struct entry *e = xfile->entry;
+		tracef("about to wal swap");
+		rv = wal_swap(e, hdr);
+		if (rv != SQLITE_OK) {
+			return rv;
+		}
+		/* check that the WAL-index hdr makes sense and save it */
+		struct wal_index_basic_hdr ihdr = get_full_hdr(e)->basic[0];
+		assert(ihdr.isInit == 1);
+		assert(ihdr.mxFrame == 0);
+		e->prev_txn_hdr = ihdr;
+		sm_move(&e->wtx_sm, WTX_ACTIVE);
+		return SQLITE_OK;
+	}
+
+	sqlite3_file *orig = get_orig(xfile);
+	rv = orig->pMethods->xWrite(orig, buf, amt, ofst);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+
+	if (xfile->flags & SQLITE_OPEN_WAL) {
+		struct entry *e = xfile->entry;
+		tracef("wrote to WAL name=%s amt=%d ofst=%lld", e->wal_cur_fixed_name, amt, ofst);
+		return vfs2_wal_post_write(e, buf, amt, ofst);
+	}
+
+	return SQLITE_OK;
+}
+
+static int vfs2_truncate(sqlite3_file *file, sqlite3_int64 size)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xTruncate(orig, size);
+}
+
+static int vfs2_sync(sqlite3_file *file, int flags)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xSync(orig, flags);
+}
+
+static int vfs2_file_size(sqlite3_file *file, sqlite3_int64 *size)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xFileSize(orig, size);
+}
+
+static int vfs2_lock(sqlite3_file *file, int mode)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xLock(orig, mode);
+}
+
+static int vfs2_unlock(sqlite3_file *file, int mode)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xUnlock(orig, mode);
+}
+
+static int vfs2_check_reserved_lock(sqlite3_file *file, int *out)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xCheckReservedLock(orig, out);
+}
+
+static int interpret_pragma(char **args)
+{
+	char **e = &args[0];
+	char *left = args[1];
+	PRE(left != NULL);
+	char *right = args[2];
+
+	if (strcmp(left, "journal_mode") == 0 && right != NULL && strcasecmp(right, "wal") != 0) {
+		*e = sqlite3_mprintf("dqlite requires WAL mode");
+		return SQLITE_ERROR;
+	}
+
+	return SQLITE_NOTFOUND;
+}
+
+static int vfs2_file_control(sqlite3_file *file, int op, void *arg)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	int rv;
+
+	if (op == SQLITE_FCNTL_COMMIT_PHASETWO && e->pending_txn_len != 0) {
+		/* Hide the transaction that was just written by resetting
+		 * the WAL index header. */
+		struct wal_index_full_hdr *hdr = get_full_hdr(e);
+		e->pending_txn_hdr = hdr->basic[0];
+		hdr->basic[0] = e->prev_txn_hdr;
+		hdr->basic[1] = hdr->basic[0];
+		e->wal_cursor += e->pending_txn_len;
+		sm_move(&xfile->entry->wtx_sm, WTX_HIDDEN);
+	} else if (op == SQLITE_FCNTL_PRAGMA) {
+		rv = interpret_pragma(arg);
+		if (rv != SQLITE_NOTFOUND) {
+			return rv;
+		}
+	} else if (op == SQLITE_FCNTL_PERSIST_WAL) {
+		/* TODO handle setting as well as getting (?) */
+		int *out = arg;
+		*out = 1;
+		return SQLITE_OK;
+	}
+
+	sqlite3_file *orig = get_orig(xfile);
+	rv = orig->pMethods->xFileControl(orig, op, arg);
+	return rv;
+}
+
+static int vfs2_sector_size(sqlite3_file *file)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xSectorSize(orig);
+}
+
+static int vfs2_device_characteristics(sqlite3_file *file)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xDeviceCharacteristics(orig);
+}
+
+static int vfs2_fetch(sqlite3_file *file,
+		      sqlite3_int64 ofst,
+		      int amt,
+		      void **out)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xFetch(orig, ofst, amt, out);
+}
+
+static int vfs2_unfetch(sqlite3_file *file, sqlite3_int64 ofst, void *buf)
+{
+	struct file *xfile = (struct file *)file;
+	sqlite3_file *orig = get_orig(xfile);
+	return orig->pMethods->xUnfetch(orig, ofst, buf);
+}
+
+static int vfs2_shm_map(sqlite3_file *file,
+			int regno,
+			int regsz,
+			int extend,
+			void volatile **out)
+{
+	struct file *xfile = (struct file *)file;
+	struct entry *e = xfile->entry;
+	void *region;
+	int rv;
+
+	if (e->shm_regions != NULL && regno < e->shm_regions_len) {
+		region = e->shm_regions[regno];
+		assert(region != NULL);
+	} else if (extend != 0) {
+		assert(regno == e->shm_regions_len);
+		region = sqlite3_malloc(regsz);
+		if (region == NULL) {
+			rv = SQLITE_NOMEM;
+			goto err;
+		}
+		memset(region, 0, (size_t)regsz);
+		/* FIXME reallocating every time seems bad */
+		sqlite3_uint64 z = (sqlite3_uint64)sizeof(*e->shm_regions) * (sqlite3_uint64)(e->shm_regions_len + 1);
+		void **regions = sqlite3_realloc64(e->shm_regions, z);
+		if (regions == NULL) {
+			rv = SQLITE_NOMEM;
+			goto err_after_region_malloc;
+		}
+		e->shm_regions = regions;
+		e->shm_regions[regno] = region;
+		e->shm_regions_len++;
+	} else {
+		region = NULL;
+	}
+
+	*out = region;
+	if (regno == 0 && region != NULL) {
+		e->shm_refcount++;
+	}
+	return SQLITE_OK;
+
+err_after_region_malloc:
+	sqlite3_free(region);
+err:
+	assert(rv != SQLITE_OK);
+	*out = NULL;
+	return rv;
+}
+
+static __attribute__((noinline)) int busy(void)
+{
+	return SQLITE_BUSY;
+}
+
+static int vfs2_shm_lock(sqlite3_file *file, int ofst, int n, int flags)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile != NULL);
+	struct entry *e = xfile->entry;
+
+	assert(ofst >= 0 && ofst + n <= SQLITE_SHM_NLOCK);
+	assert(n >= 1);
+	assert(n == 1 || (flags & SQLITE_SHM_EXCLUSIVE) != 0);
+
+	assert(flags == (SQLITE_SHM_LOCK | SQLITE_SHM_SHARED) ||
+	       flags == (SQLITE_SHM_LOCK | SQLITE_SHM_EXCLUSIVE) ||
+	       flags == (SQLITE_SHM_UNLOCK | SQLITE_SHM_SHARED) ||
+	       flags == (SQLITE_SHM_UNLOCK | SQLITE_SHM_EXCLUSIVE));
+
+	assert(xfile->flags & SQLITE_OPEN_MAIN_DB);
+
+	if (flags == (SQLITE_SHM_LOCK | SQLITE_SHM_SHARED)) {
+		for (int i = ofst; i < ofst + n; i++) {
+			if (e->shm_locks[i] == VFS2_EXCLUSIVE) {
+				return busy();
+			}
+		}
+
+		for (int i = ofst; i < ofst + n; i++) {
+			e->shm_locks[i]++;
+		}
+	} else if (flags == (SQLITE_SHM_LOCK | SQLITE_SHM_EXCLUSIVE)) {
+		for (int i = ofst; i < ofst + n; i++) {
+			if (e->shm_locks[i] > 0) {
+				return busy();
+			}
+		}
+
+		for (int i = ofst; i < ofst + n; i++) {
+			e->shm_locks[i] = VFS2_EXCLUSIVE;
+		}
+
+		/* XXX maybe this shouldn't be an assertion */
+		if (ofst == WAL_WRITE_LOCK) {
+			assert(n == 1);
+			assert(e->pending_txn_len == 0);
+		}
+	} else if (flags == (SQLITE_SHM_UNLOCK | SQLITE_SHM_SHARED)) {
+		for (int i = ofst; i < ofst + n; i++) {
+			assert(e->shm_locks[i] > 0);
+			e->shm_locks[i]--;
+		}
+	} else if (flags == (SQLITE_SHM_UNLOCK | SQLITE_SHM_EXCLUSIVE)) {
+		for (int i = ofst; i < ofst + n; i++) {
+			assert(e->shm_locks[i] == VFS2_EXCLUSIVE);
+			e->shm_locks[i] = 0;
+		}
+
+		if (ofst <= WAL_RECOVER_LOCK && WAL_RECOVER_LOCK < ofst + n) {
+			tracef("unlocking the recovery lock!");
+		}
+
+		if (ofst == WAL_WRITE_LOCK) {
+			/* Unlocking the write lock: roll back any uncommitted
+			 * transaction. */
+			assert(n == 1);
+			tracef("unlocking write lock");
+			/* TODO make sure this is correct */
+			if (e->pending_txn_last_frame_commit == 0) {
+				free_pending_txn(e);
+				sm_move(&e->wtx_sm, WTX_BASE);
+			}
+		} else if (ofst == WAL_CKPT_LOCK && n == 1) {
+			/* End of a checkpoint: if all frames have been backfilled,
+			 * move to EMPTY. */
+			assert(n == 1);
+			struct wal_index_full_hdr *ihdr = get_full_hdr(e);
+			if (ihdr->nBackfill == ihdr->basic[0].mxFrame) {
+				sm_move(&e->wtx_sm, WTX_EMPTY);
+			}
+		} /* else if (ofst <= WAL_RECOVER_LOCK && WAL_RECOVER_LOCK < ofst + n) {
+			sm_move(&e->wtx_sm, WTX_BASE);
+		} */
+	} else {
+		assert(0);
+	}
+
+	return SQLITE_OK;
+}
+
+static void vfs2_shm_barrier(sqlite3_file *file)
+{
+	(void)file;
+}
+
+static int vfs2_shm_unmap(sqlite3_file *file, int delete)
+{
+	(void)delete;
+	struct file *xfile = (struct file *)file;
+	struct entry *e = xfile->entry;
+	e->shm_refcount--;
+	if (e->shm_refcount == 0) {
+		for (int i = 0; i < e->shm_regions_len; i++) {
+			void *region = e->shm_regions[i];
+			assert(region != NULL);
+			sqlite3_free(region);
+		}
+		sqlite3_free(e->shm_regions);
+
+		e->shm_regions = NULL;
+		e->shm_regions_len = 0;
+		memset(e->shm_locks, 0, sizeof(e->shm_locks));
+	}
+	return SQLITE_OK;
+}
+
+static struct sqlite3_io_methods vfs2_io_methods = {
+	.iVersion = 3,
+	.xClose = vfs2_close,
+	.xRead = vfs2_read,
+	.xWrite = vfs2_write,
+	.xTruncate = vfs2_truncate,
+	.xSync = vfs2_sync,
+	.xFileSize = vfs2_file_size,
+	.xLock = vfs2_lock,
+	.xUnlock = vfs2_unlock,
+	.xCheckReservedLock = vfs2_check_reserved_lock,
+	.xFileControl = vfs2_file_control,
+	.xSectorSize = vfs2_sector_size,
+	.xDeviceCharacteristics = vfs2_device_characteristics,
+	.xShmMap = vfs2_shm_map,
+	.xShmLock = vfs2_shm_lock,
+	.xShmBarrier = vfs2_shm_barrier,
+	.xShmUnmap = vfs2_shm_unmap,
+	.xFetch = vfs2_fetch,
+	.xUnfetch = vfs2_unfetch
+};
+
+static int compare_wal_headers(struct wal_hdr a,
+			       struct wal_hdr b,
+			       bool *ordered)
+{
+	if (get_salt1(a.salts) == get_salt1(b.salts) + 1) {
+		*ordered = true;
+	} else if (get_salt1(b.salts) == get_salt1(a.salts) + 1) {
+		*ordered = false;
+	} else {
+		return SQLITE_ERROR;
+	}
+	return SQLITE_OK;
+}
+
+static int read_wal_hdr(sqlite3_file *wal, sqlite3_int64 *size, struct wal_hdr *hdr)
+{
+	int rv;
+
+	rv = wal->pMethods->xFileSize(wal, size);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+	if (*size >= (sqlite3_int64)sizeof(struct wal_hdr)) {
+		rv = wal->pMethods->xRead(wal, hdr, sizeof(*hdr), 0);
+		if (rv != SQLITE_OK) {
+			return rv;
+		}
+	} else {
+		*hdr = (struct wal_hdr){};
+	}
+	return SQLITE_OK;
+}
+
+static struct wal_index_full_hdr initial_full_hdr(struct wal_hdr whdr)
+{
+	struct wal_index_full_hdr ihdr = {};
+	ihdr.basic[0].iVersion = 3007000;
+	ihdr.basic[0].isInit = 1;
+	ihdr.basic[0].bigEndCksum = is_bigendian();
+	ihdr.basic[0].szPage = (uint16_t)ByteGetBe32(whdr.page_size);
+	struct cksums sums = {};
+	update_cksums(native_magic(), (const void *)&ihdr.basic[0], offsetof(struct wal_index_basic_hdr, cksums), &sums);
+	ihdr.basic[0].cksums = sums;
+	ihdr.basic[1] = ihdr.basic[0];
+	ihdr.marks[0] = 0;
+	ihdr.marks[1] = 0;
+	ihdr.marks[2] = READ_MARK_UNUSED;
+	ihdr.marks[3] = READ_MARK_UNUSED;
+	ihdr.marks[4] = READ_MARK_UNUSED;
+	return ihdr;
+}
+
+static void set_mx_frame(struct wal_index_full_hdr *ihdr, uint32_t mx, struct wal_frame_hdr fhdr)
+{
+	uint32_t num_pages = ByteGetBe32(fhdr.commit);
+	PRE(num_pages > 0);
+	ihdr->basic[0].iChange += 1;
+	ihdr->basic[0].mxFrame = mx;
+	ihdr->basic[0].nPage = num_pages;
+	/* XXX byte order */
+	ihdr->basic[0].frame_cksums.cksum1 = ByteGetBe32(fhdr.cksum1);
+	ihdr->basic[0].frame_cksums.cksum2 = ByteGetBe32(fhdr.cksum2);
+	struct cksums sums = {};
+	update_cksums(native_magic(), (const void *)&ihdr->basic[0], 40, &sums);
+	ihdr->basic[0].cksums = sums;
+	ihdr->basic[1] = ihdr->basic[0];
+}
+
+static void restart_full_hdr(struct wal_index_full_hdr *ihdr, struct wal_hdr new_whdr)
+{
+	/* cf. walRestartHdr */
+	ihdr->basic[0].mxFrame = 0;
+	ihdr->basic[0].salts = new_whdr.salts;
+	struct cksums sums = {};
+	update_cksums(native_magic(), (const void *)&ihdr->basic[0], 40, &sums);
+	ihdr->basic[0].cksums = sums;
+	ihdr->basic[1] = ihdr->basic[0];
+	ihdr->nBackfill = 0;
+	ihdr->nBackfillAttempted = 0;
+}
+
+static uint32_t wal_cursor_from_size(uint32_t page_size, sqlite3_int64 size)
+{
+	sqlite3_int64 whdr_size = (sqlite3_int64)sizeof(struct wal_hdr);
+	if (size < whdr_size) {
+		return 0;
+	}
+	sqlite3_int64 x = (size - whdr_size) / ((sqlite3_int64)sizeof(struct wal_frame_hdr) + (sqlite3_int64)page_size);
+	return (uint32_t)x;
+}
+
+static sqlite3_int64 wal_offset_from_cursor(uint32_t page_size, uint32_t cursor)
+{
+	return (sqlite3_int64)sizeof(struct wal_hdr) + (sqlite3_int64)cursor * ((sqlite3_int64)sizeof(struct wal_frame_hdr) + (sqlite3_int64)page_size);
+}
+
+static int open_entry(struct common *common, const char *name, struct entry *e)
+{
+	sqlite3_vfs *v = common->orig;
+	int path_cap = v->mxPathname + 1;
+	int file_cap = v->szOsFile;
+	int rv;
+
+	*e = (struct entry){};
+	e->common = common;
+	sm_init(&e->wtx_sm, wtx_invariant, NULL, wtx_states, WTX_CLOSED);
+
+	e->refcount_main_db = 1;
+
+	e->main_db_name = sqlite3_malloc(path_cap);
+	e->wal_moving_name = sqlite3_malloc(path_cap);
+	e->wal_cur_fixed_name = sqlite3_malloc(path_cap);
+	e->wal_prev_fixed_name = sqlite3_malloc(path_cap);
+	if (e->main_db_name == NULL ||
+			e->wal_moving_name == NULL ||
+			e->wal_cur_fixed_name == NULL ||
+			e->wal_prev_fixed_name == NULL) {
+		return SQLITE_NOMEM;
+	}
+
+	strcpy(e->main_db_name, name);
+
+	strcpy(e->wal_moving_name, name);
+	strcat(e->wal_moving_name, "-wal");
+
+	strcpy(e->wal_cur_fixed_name, name);
+	strcat(e->wal_cur_fixed_name, "-xwal1");
+
+	strcpy(e->wal_prev_fixed_name, name);
+	strcat(e->wal_prev_fixed_name, "-xwal2");
+
+	/* TODO EXRESCODE? */
+	int phys_wal_flags = SQLITE_OPEN_READWRITE|SQLITE_OPEN_CREATE|SQLITE_OPEN_WAL|SQLITE_OPEN_NOFOLLOW;
+
+	e->wal_cur = sqlite3_malloc(file_cap);
+	if (e->wal_cur == NULL) {
+		return SQLITE_NOMEM;
+	}
+	rv = v->xOpen(v, e->wal_cur_fixed_name, e->wal_cur, phys_wal_flags, NULL);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+
+	e->wal_prev = sqlite3_malloc(file_cap);
+	if (e->wal_prev == NULL) {
+		return SQLITE_NOMEM;
+	}
+	rv = v->xOpen(v, e->wal_prev_fixed_name, e->wal_prev, phys_wal_flags, NULL);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+
+	sqlite3_int64 size1;
+	struct wal_hdr hdr1;
+	rv = read_wal_hdr(e->wal_cur, &size1, &hdr1);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+	sqlite3_int64 size2;
+	struct wal_hdr hdr2;
+	rv = read_wal_hdr(e->wal_prev, &size2, &hdr2);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+
+	struct wal_hdr hdr_cur = hdr1;
+	sqlite3_int64 size_cur = size1;
+	struct wal_hdr hdr_prev = hdr2;
+
+	bool wal1_is_fresh;
+	if (size2 < (sqlite3_int64)sizeof(struct wal_hdr)) {
+		wal1_is_fresh = true;
+	} else if (size1 < (sqlite3_int64)sizeof(struct wal_hdr)) {
+		wal1_is_fresh = false;
+	} else {
+		rv = compare_wal_headers(hdr1, hdr2, &wal1_is_fresh);
+		if (rv != SQLITE_OK) {
+			return rv;
+		}
+	}
+	if (!wal1_is_fresh) {
+		void *temp;
+
+		temp = e->wal_cur;
+		e->wal_cur = e->wal_prev;
+		e->wal_prev = temp;
+
+		temp = e->wal_cur_fixed_name;
+		e->wal_cur_fixed_name = e->wal_prev_fixed_name;
+		e->wal_prev_fixed_name = temp;
+
+		hdr_cur = hdr2;
+		size_cur = size2;
+		hdr_prev = hdr1;
+	}
+
+	e->wal_cur_hdr = hdr_cur;
+	e->wal_prev_hdr = hdr_prev;
+
+	rv = unlink(e->wal_moving_name);
+	(void)rv;
+	rv = link(e->wal_cur_fixed_name, e->wal_moving_name);
+	(void)rv;
+
+	e->shm_regions = sqlite3_malloc(sizeof(void *[1]));
+	if (e->shm_regions == NULL) {
+		return SQLITE_NOMEM;
+	}
+	e->shm_regions[0] = sqlite3_malloc(VFS2_WAL_INDEX_REGION_SIZE);
+	if (e->shm_regions[0] == NULL) {
+		return SQLITE_NOMEM;
+	}
+	memset(e->shm_regions[0], 0, VFS2_WAL_INDEX_REGION_SIZE);
+	e->shm_regions_len = 1;
+
+	*get_full_hdr(e) = initial_full_hdr(hdr_cur);
+
+	e->wal_cursor = wal_cursor_from_size(e->page_size, size_cur);
+
+	int next = WTX_EMPTY;
+	if (size_cur >= wal_offset_from_cursor(0 /* this doesn't matter */, 0)) {
+		/* TODO verify the header here */
+		e->page_size = ByteGetBe32(hdr_cur.page_size);
+		next = WTX_FLUSH;
+	}
+	if (size_cur >= wal_offset_from_cursor(e->page_size, 1)) {
+		e->shm_locks[WAL_WRITE_LOCK] = VFS2_EXCLUSIVE;
+		next = WTX_FOLLOWING;
+	}
+	sm_move(&e->wtx_sm, next);
+
+	return SQLITE_OK;
+}
+
+static int set_up_entry(struct common *common, const char *name, int flags, struct entry **e)
+{
+	bool name_is_db = (flags & SQLITE_OPEN_MAIN_DB) != 0;
+	bool name_is_wal = (flags & SQLITE_OPEN_WAL) != 0;
+	assert(name_is_db ^ name_is_wal);
+	int rv;
+
+	struct entry *res = NULL;
+	pthread_rwlock_rdlock(&common->rwlock);
+	queue *q;
+	QUEUE_FOREACH(q, &common->queue)
+	{
+		struct entry *cur = QUEUE_DATA(q, struct entry, link);
+		if ((name_is_db && strcmp(cur->main_db_name, name) == 0) ||
+		    (name_is_wal && strcmp(cur->wal_moving_name, name) == 0)) {
+			res = cur;
+			break;
+		}
+	}
+	pthread_rwlock_unlock(&common->rwlock);
+	if (res != NULL) {
+		sqlite3_free(*e);
+		*e = res;
+		unsigned *refcount = name_is_db ? &res->refcount_main_db : &res->refcount_wal;
+		*refcount += 1;
+		return SQLITE_OK;
+	}
+
+	assert(name_is_db);
+	res = *e;
+	/* If open_entry fails we still want to link in the entry. Since we unconditionally
+	 * set pMethods in our file vtable, SQLite will xClose the file and vfs2_close
+	 * will run to clean up the partial work of open_entry. */
+	rv = open_entry(common, name, res);
+	pthread_rwlock_wrlock(&common->rwlock);
+	queue_insert_tail(&common->queue, &res->link);
+	pthread_rwlock_unlock(&common->rwlock);
+	return rv;
+}
+
+static int vfs2_open(sqlite3_vfs *vfs,
+		     const char *name,
+		     sqlite3_file *out,
+		     int flags,
+		     int *out_flags)
+{
+	struct file *xout = (struct file *)out;
+	struct common *common = vfs->pAppData;
+	int rv;
+
+	*xout = (struct file){};
+	xout->base.pMethods = &vfs2_io_methods;
+	xout->flags = flags;
+
+	if ((flags & SQLITE_OPEN_WAL) == 0) {
+		sqlite3_vfs *v = common->orig;
+		xout->orig = sqlite3_malloc(v->szOsFile);
+		if (xout->orig == NULL) {
+			return SQLITE_NOMEM;
+		}
+		rv = v->xOpen(v, name, xout->orig, flags, out_flags);
+		if (rv != SQLITE_OK) {
+			return rv;
+		}
+	}
+
+	if (flags & (SQLITE_OPEN_MAIN_DB|SQLITE_OPEN_WAL)) {
+		xout->entry = sqlite3_malloc(sizeof(*xout->entry));
+		if (xout->entry == NULL) {
+			return SQLITE_NOMEM;
+		}
+		rv = set_up_entry(common, name, flags, &xout->entry);
+		if (rv != SQLITE_OK) {
+			return rv;
+		}
+	}
+
+	if ((flags & SQLITE_OPEN_WAL) && out_flags != NULL) {
+		*out_flags = flags;
+	}
+
+	return SQLITE_OK;
+}
+
+/* TODO does this need to be customized? should it ever be called on one of our files? */
+static int vfs2_delete(sqlite3_vfs *vfs, const char *name, int sync_dir)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xDelete(data->orig, name, sync_dir);
+}
+
+static int vfs2_access(sqlite3_vfs *vfs, const char *name, int flags, int *out)
+{
+	/* TODO always report that the WAL exists (?) */
+	/* TODO other customizations? */
+	struct common *data = vfs->pAppData;
+	return data->orig->xAccess(data->orig, name, flags, out);
+}
+
+static int vfs2_full_pathname(sqlite3_vfs *vfs,
+			      const char *name,
+			      int n,
+			      char *out)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xFullPathname(data->orig, name, n, out);
+}
+
+static void *vfs2_dl_open(sqlite3_vfs *vfs, const char *filename)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xDlOpen(data->orig, filename);
+}
+
+static void vfs2_dl_error(sqlite3_vfs *vfs, int n, char *msg)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xDlError(data->orig, n, msg);
+}
+
+typedef void (*vfs2_sym)(void);
+static vfs2_sym vfs2_dl_sym(sqlite3_vfs *vfs, void *dl, const char *symbol)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xDlSym(data->orig, dl, symbol);
+}
+
+static void vfs2_dl_close(sqlite3_vfs *vfs, void *dl)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xDlClose(data->orig, dl);
+}
+
+static int vfs2_randomness(sqlite3_vfs *vfs, int n, char *out)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xRandomness(data->orig, n, out);
+}
+
+static int vfs2_sleep(sqlite3_vfs *vfs, int microseconds)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xSleep(data->orig, microseconds);
+}
+
+static int vfs2_current_time(sqlite3_vfs *vfs, double *out)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xCurrentTime(data->orig, out);
+}
+
+/* TODO update this to reflect syscalls that we make ourselves (not through the
+ * base VFS -- store last error in vfs [thread-local?]) */
+static int vfs2_get_last_error(sqlite3_vfs *vfs, int n, char *out)
+{
+	struct common *data = vfs->pAppData;
+	return data->orig->xGetLastError(data->orig, n, out);
+}
+
+static int vfs2_current_time_int64(sqlite3_vfs *vfs, sqlite3_int64 *out)
+{
+	struct common *data = vfs->pAppData;
+	if (data->orig->iVersion < 2) {
+		return SQLITE_ERROR;
+	}
+	return data->orig->xCurrentTimeInt64(data->orig, out);
+}
+
+/* sqlite3_vfs implementations end here */
+
+sqlite3_vfs *vfs2_make(sqlite3_vfs *orig, const char *name)
+{
+	struct common *common = sqlite3_malloc(sizeof(*common));
+	struct sqlite3_vfs *vfs = sqlite3_malloc(sizeof(*vfs));
+	if (common == NULL || vfs == NULL) {
+		return NULL;
+	}
+	common->orig = orig;
+	pthread_rwlock_init(&common->rwlock, NULL);
+	queue_init(&common->queue);
+	vfs->iVersion = 2;
+	vfs->szOsFile = sizeof(struct file);
+	vfs->mxPathname = orig->mxPathname;
+	vfs->zName = name;
+	vfs->pAppData = common;
+	vfs->xOpen = vfs2_open;
+	vfs->xDelete = vfs2_delete;
+	vfs->xAccess = vfs2_access;
+	vfs->xFullPathname = vfs2_full_pathname;
+	vfs->xDlOpen = vfs2_dl_open;
+	vfs->xDlError = vfs2_dl_error;
+	vfs->xDlSym = vfs2_dl_sym;
+	vfs->xDlClose = vfs2_dl_close;
+	vfs->xRandomness = vfs2_randomness;
+	vfs->xSleep = vfs2_sleep;
+	vfs->xCurrentTime = vfs2_current_time;
+	vfs->xGetLastError = vfs2_get_last_error;
+	vfs->xCurrentTimeInt64 = vfs2_current_time_int64;
+	return vfs;
+}
+
+int vfs2_unapply(sqlite3_file *file, struct vfs2_wal_slice first_to_unapply)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	PRE(salts_equal(first_to_unapply.salts, e->wal_cur_hdr.salts));
+	PRE(first_to_unapply.start + first_to_unapply.len <= e->wal_cursor);
+	struct wal_index_full_hdr *ihdr = get_full_hdr(e);
+	PRE(first_to_unapply.start >= ihdr->basic[0].mxFrame);
+	PRE(e->shm_locks[WAL_WRITE_LOCK] == VFS2_EXCLUSIVE);
+
+	e->wal_cursor = first_to_unapply.start;
+
+	if (e->wal_cursor == ihdr->basic[0].mxFrame) {
+		e->shm_locks[WAL_WRITE_LOCK] = 0;
+		sm_move(&e->wtx_sm, WTX_FLUSH);
+	} else {
+		sm_move(&e->wtx_sm, WTX_FOLLOWING);
+	}
+	return SQLITE_OK;
+}
+
+int vfs2_unhide(sqlite3_file *file)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	PRE(e->shm_locks[WAL_WRITE_LOCK] == VFS2_EXCLUSIVE);
+	e->shm_locks[WAL_WRITE_LOCK] = 0;
+
+	struct wal_index_full_hdr *hdr = get_full_hdr(e);
+	hdr->basic[0] = e->pending_txn_hdr;
+	hdr->basic[1] = e->pending_txn_hdr;
+	e->prev_txn_hdr = e->pending_txn_hdr;
+	e->pending_txn_hdr = (struct wal_index_basic_hdr){};
+	e->pending_txn_len = 0;
+	e->pending_txn_last_frame_commit = 0;
+
+	sm_move(&xfile->entry->wtx_sm, WTX_BASE);
+
+	return 0;
+}
+
+int vfs2_commit(sqlite3_file *file, struct vfs2_wal_slice stop)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	uint32_t commit = stop.start + stop.len;
+	PRE(e->wal_cursor >= commit);
+	PRE(salts_equal(stop.salts, e->wal_cur_hdr.salts));
+	PRE(e->shm_locks[WAL_WRITE_LOCK] == VFS2_EXCLUSIVE);
+	sqlite3_file *wal_cur = e->wal_cur;
+	struct wal_frame_hdr fhdr;
+	int rv = wal_cur->pMethods->xRead(wal_cur, &fhdr, sizeof(fhdr), wal_offset_from_cursor(e->page_size, stop.start + stop.len - 1));
+	if (rv == SQLITE_OK) {
+		return rv;
+	}
+	set_mx_frame(get_full_hdr(e), commit, fhdr);
+	if (commit == e->wal_cursor) {
+		e->shm_locks[WAL_WRITE_LOCK] = 0;
+		sm_move(&e->wtx_sm, WTX_FLUSH);
+	} else {
+		sm_move(&e->wtx_sm, WTX_FOLLOWING);
+	}
+	return 0;
+}
+
+int vfs2_commit_barrier(sqlite3_file *file)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	if (e->wal_cursor > 0) {
+		sqlite3_file *wal_cur = e->wal_cur;
+		struct wal_frame_hdr fhdr;
+		int rv = wal_cur->pMethods->xRead(wal_cur, &fhdr, sizeof(fhdr), wal_offset_from_cursor(e->page_size, e->wal_cursor - 1));
+		if (rv == SQLITE_OK) {
+			return rv;
+		}
+		set_mx_frame(get_full_hdr(e), e->wal_cursor, fhdr);
+		/* It's okay if the write lock isn't held */
+		e->shm_locks[WAL_WRITE_LOCK] = 0;
+		get_full_hdr(e)->basic[0].isInit = 0;
+		/* The next transaction will cause SQLite to run recovery which will complete the transition to BASE */
+		sm_move(&e->wtx_sm, WTX_FLUSH);
+	}
+	return 0;
+}
+
+int vfs2_poll(sqlite3_file *file, struct vfs2_wal_frame **frames, unsigned *n, struct vfs2_wal_slice *sl)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+
+	uint32_t len = e->pending_txn_len;
+	if (len > 0) {
+		/* Don't go through vfs2_shm_lock here since that has additional
+		 * checks that assume the context of being called from inside
+		 * SQLite. */
+		if (e->shm_locks[WAL_WRITE_LOCK] > 0) {
+			return 1;
+		}
+		e->shm_locks[WAL_WRITE_LOCK] = VFS2_EXCLUSIVE;
+	}
+
+	/* Note, not resetting pending_txn_{start,len} because they are used by later states */
+	if (n != NULL && frames != NULL) {
+		*n = len;
+		*frames = e->pending_txn_frames;
+	} else {
+		for (uint32_t i = 0; i < e->pending_txn_len; i++) {
+			sqlite3_free(e->pending_txn_frames[i].page);
+		}
+		sqlite3_free(e->pending_txn_frames);
+	}
+	e->pending_txn_frames = NULL;
+
+	if (sl != NULL) {
+		sl->len = len;
+		sl->salts = e->pending_txn_hdr.salts;
+		sl->start = e->prev_txn_hdr.mxFrame;
+		sl->len = len;
+	}
+
+	sm_move(&xfile->entry->wtx_sm, WTX_POLLED);
+
+	return 0;
+}
+
+void vfs2_destroy(sqlite3_vfs *vfs)
+{
+	struct common *data = vfs->pAppData;
+	pthread_rwlock_destroy(&data->rwlock);
+	sqlite3_free(data);
+	sqlite3_free(vfs);
+}
+
+int vfs2_abort(sqlite3_file *file)
+{
+	/* TODO maybe can "followerize" this and get rid of vfs2_unapply_after? */
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+
+	e->shm_locks[WAL_WRITE_LOCK] = 0;
+
+	struct wal_index_full_hdr *hdr = get_full_hdr(e);
+	hdr->basic[0] = e->prev_txn_hdr;
+	hdr->basic[1] = e->prev_txn_hdr;
+	e->pending_txn_hdr = (struct wal_index_basic_hdr){};
+
+	e->wal_cursor = e->pending_txn_start;
+	free_pending_txn(e);
+
+	sm_move(&xfile->entry->wtx_sm, WTX_BASE);
+	return 0;
+}
+
+int vfs2_read_wal(sqlite3_file *file,
+		  struct vfs2_wal_txn *txns,
+		  size_t txns_len)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	int rv;
+
+	/* TODO check wal integrity before reading it */
+	(void)check_wal_integrity;
+
+	int page_size = (int)e->page_size;
+	for (size_t i = 0; i < txns_len; i++) {
+		struct vfs2_wal_frame *f = sqlite3_malloc64(txns[i].meta.len * sizeof(*f));
+		if (f == NULL) {
+			goto oom;
+		}
+		txns[i].frames = f;
+		for (size_t j = 0; j < txns[i].meta.len; j++) {
+			void *p = sqlite3_malloc(page_size);
+			if (p == NULL) {
+				goto oom;
+			}
+			txns[i].frames[j].page = p;
+		}
+	}
+
+	for (size_t i = 0; i < txns_len; i++) {
+		sqlite3_file *wal;
+		unsigned read_lock;
+		bool from_wal_cur = salts_equal(txns[i].meta.salts, e->wal_cur_hdr.salts);
+		bool from_wal_prev = salts_equal(txns[i].meta.salts, e->wal_prev_hdr.salts);
+		assert(from_wal_cur ^ from_wal_prev);
+		if (from_wal_cur) {
+			rv = vfs2_pseudo_read_begin(file, e->wal_cursor, &read_lock);
+			if (rv != SQLITE_OK) {
+				return 1;
+			}
+			wal = e->wal_cur;
+		} else {
+			wal = e->wal_prev;
+		}
+
+		uint32_t start = txns[i].meta.start;
+		uint32_t len = txns[i].meta.len;
+		for (uint32_t j = 0; j < len; j++) {
+			sqlite3_int64 off = wal_offset_from_cursor(e->page_size, start + j);
+			struct wal_frame_hdr fhdr;
+			rv = wal->pMethods->xRead(wal, &fhdr, sizeof(fhdr), off);
+			if (rv != SQLITE_OK) {
+				return 1;
+			}
+			off += (sqlite3_int64)sizeof(fhdr);
+			rv = wal->pMethods->xRead(wal, txns[i].frames[j].page, page_size, off);
+			if (rv != SQLITE_OK) {
+				return 1;
+			}
+			txns[i].frames[j].page_number = ByteGetBe32(fhdr.page_number);
+			txns[i].frames[j].commit = ByteGetBe32(fhdr.commit);
+		}
+		if (from_wal_cur) {
+			vfs2_pseudo_read_end(file, read_lock);
+		}
+	}
+
+	return 0;
+
+oom:
+	for (uint32_t i = 0; i < txns_len; i++) {
+		for (uint32_t j = 0; j < txns[i].meta.len; j++) {
+			sqlite3_free(txns[i].frames[j].page);
+		}
+		sqlite3_free(txns[i].frames);
+		txns[i].frames = NULL;
+	}
+	return 1;
+}
+
+static int write_one_frame(struct entry *e, struct wal_frame_hdr hdr, void *data)
+{
+	int rv;
+
+	sqlite3_int64 off = wal_offset_from_cursor(e->page_size, e->wal_cursor);
+	rv = e->wal_cur->pMethods->xWrite(e->wal_cur, &hdr, sizeof(hdr), off);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+	off += (sqlite3_int64)sizeof(hdr);
+	rv = e->wal_cur->pMethods->xWrite(e->wal_cur, data, (int)e->page_size, off);
+	if (rv != SQLITE_OK) {
+		return rv;
+	}
+	e->wal_cursor += 1;
+	return SQLITE_OK;
+}
+
+static struct wal_hdr next_wal_hdr(const struct entry *e)
+{
+	struct wal_hdr ret;
+	struct wal_hdr old = e->wal_cur_hdr;
+	BytePutBe32(native_magic(), ret.magic);
+	BytePutBe32(3007000, ret.version);
+	BytePutBe32(e->page_size, ret.page_size);
+	uint32_t ckpoint_seqno = ByteGetBe32(old.ckpoint_seqno);
+	BytePutBe32(ckpoint_seqno + 1, ret.ckpoint_seqno);
+	uint32_t salt1;
+       	if (ckpoint_seqno == 0) {
+		salt1 = get_salt1(old.salts) + 1;
+	} else {
+		e->common->orig->xRandomness(e->common->orig, sizeof(salt1), (void *)&salt1);
+	}
+	BytePutBe32(salt1, ret.salts.salt1);
+	e->common->orig->xRandomness(e->common->orig, sizeof(ret.salts.salt2), (void *)&ret.salts.salt2);
+	return ret;
+}
+
+static struct wal_frame_hdr txn_frame_hdr(struct entry *e, struct cksums sums, struct vfs2_wal_frame frame)
+{
+	struct wal_frame_hdr fhdr;
+
+	BytePutBe32(frame.page_number, fhdr.page_number);
+	BytePutBe32(frame.commit, fhdr.commit);
+	update_cksums(ByteGetBe32(e->wal_cur_hdr.magic), (const void *)(&fhdr), 8, &sums);
+	update_cksums(ByteGetBe32(e->wal_cur_hdr.magic), frame.page, e->page_size, &sums);
+	fhdr.salts = e->wal_cur_hdr.salts;
+	BytePutBe32(sums.cksum1, fhdr.cksum1);
+	BytePutBe32(sums.cksum2, fhdr.cksum2);
+	return fhdr;
+}
+
+int vfs2_apply_uncommitted(sqlite3_file *file, uint32_t page_size, const struct vfs2_wal_frame *frames, unsigned len, struct vfs2_wal_slice *out)
+{
+	PRE(len > 0);
+	PRE(is_valid_page_size(page_size));
+	for (unsigned i = 0; i < len - 1; i++) {
+		PRE(frames[i].commit == 0);
+	}
+	PRE(frames[len - 1].commit > 0);
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	PRE(page_size == e->page_size);
+	int rv;
+
+	/* The write lock is always held if there is at least one
+	 * uncommitted frame in WAL-cur. In FOLLOWING state, we allow
+	 * adding more frames to WAL-cur even if there are already
+	 * some uncommitted frames. Hence we don't check the write
+	 * lock here before "acquiring" it, we just make sure that
+	 * it's held before returning.
+	 *
+	 * The write lock will be released when a call to vfs2_commit
+	 * or vfs2_unapply causes the number of committed frames in
+	 * WAL-cur (mxFrame) to equal the number of applies frames
+	 * (wal_cursor). */
+	e->shm_locks[WAL_WRITE_LOCK] = VFS2_EXCLUSIVE;
+	
+	struct wal_index_full_hdr *ihdr = get_full_hdr(e);
+	uint32_t mx = ihdr->basic[0].mxFrame;
+	if (mx > 0 && ihdr->nBackfill == mx) {
+		struct wal_hdr new_whdr = next_wal_hdr(e);
+		restart_full_hdr(ihdr, new_whdr);
+		rv = wal_swap(e, &new_whdr);
+		if (rv != SQLITE_OK) {
+			return 1;
+		}
+		/* sm_move(&e->wtx_sm, WTX_FLUSH); */
+	}
+
+	uint32_t start = e->wal_cursor;
+
+	struct cksums sums;
+	if (start > 0) {
+		/* TODO cache this in the entry? */
+		struct wal_frame_hdr prev_fhdr;
+		sqlite3_int64 off = wal_offset_from_cursor(e->page_size, e->wal_cursor - 1);
+		rv = e->wal_cur->pMethods->xRead(e->wal_cur, &prev_fhdr, sizeof(prev_fhdr), off);
+		if (rv != SQLITE_OK) {
+			return 1;
+		}
+		sums.cksum1 = ByteGetBe32(prev_fhdr.cksum1);
+		sums.cksum2 = ByteGetBe32(prev_fhdr.cksum2);
+	} else {
+		sums.cksum1 = ByteGetBe32(e->wal_cur_hdr.cksum1);
+		sums.cksum2 = ByteGetBe32(e->wal_cur_hdr.cksum2);
+	}
+
+	struct wal_frame_hdr fhdr = txn_frame_hdr(e, sums, frames[0]);
+	rv = write_one_frame(e, fhdr, frames[0].page);
+	if (rv != SQLITE_OK) {
+		return 1;
+	}
+
+	for (unsigned i = 1; i < len; i++) {
+		sums.cksum1 = ByteGetBe32(fhdr.cksum1);
+		sums.cksum2 = ByteGetBe32(fhdr.cksum2);
+		fhdr = txn_frame_hdr(e, sums, frames[i]);
+		rv = write_one_frame(e, fhdr, frames[i].page);
+		if (rv != SQLITE_OK) {
+			return 1;
+		}
+	}	
+
+	sm_move(&e->wtx_sm, WTX_FOLLOWING);
+	out->salts = e->wal_cur_hdr.salts;
+	out->start = start;
+	out->len = len;
+	return 0;
+}
+
+/* Get the index of the `i`th read lock in the array of
+ * shm locks. There are five read locks, and three non-read
+ * locks that come before the read locks.
+ * See https://sqlite.org/walformat.html#wal_locks. */
+static unsigned read_lock(unsigned i)
+{
+	PRE(i < 5);
+	return 3 + i;
+}
+
+int vfs2_pseudo_read_begin(sqlite3_file *file, uint32_t target, unsigned *out)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	struct wal_index_full_hdr *ihdr = get_full_hdr(e);
+
+	/* adapted from walTryBeginRead */
+	uint32_t max_mark = 0;
+	unsigned max_index = 0;
+	for (unsigned i = 1; i < WAL_NREADER; i++){
+		uint32_t cur = ihdr->marks[i];
+		if (max_mark <= cur && cur <= target) {
+			assert(cur != READ_MARK_UNUSED);
+			max_mark = cur;
+			max_index = i;
+		}
+	}
+	if (max_mark < target || max_index == 0) {
+		for (unsigned i = 1; i < WAL_NREADER; i++) {
+			if (e->shm_locks[read_lock(i)] > 0) {
+				continue;
+			}
+			ihdr->marks[i] = target;
+			max_mark = target;
+			max_index = i;
+			break;
+		}
+	}
+	if (max_index == 0) {
+		return 1;
+	}
+	*out = max_index;
+	return 0;
+}
+
+int vfs2_pseudo_read_end(sqlite3_file *file, unsigned i)
+{
+	struct file *xfile = (struct file *)file;
+	PRE(xfile->flags & SQLITE_OPEN_MAIN_DB);
+	struct entry *e = xfile->entry;
+	PRE(e->shm_locks[i] > 0);
+	e->shm_locks[i] -= 1;
+	return 0;
+}
diff -pruN 1.16.4-4/src/vfs2.h 1.16.5-1/src/vfs2.h
--- 1.16.4-4/src/vfs2.h	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/src/vfs2.h	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,105 @@
+#ifndef DQLITE_VFS2_H
+#define DQLITE_VFS2_H
+
+#include <sqlite3.h>
+
+#include <stddef.h>
+#include <stdint.h>
+
+/**
+ * Create a new VFS object that wraps the given VFS object.
+ *
+ * The returned VFS is allocated on the heap and lives until vfs2_destroy is
+ * called.  Its methods are thread-safe if those of the wrapped VFS are, but
+ * the methods of the sqlite3_file objects it creates are not thread-safe.
+ * Therefore, a database connection that's created using this VFS should only
+ * be used on the thread that opened it. The functions below that operate on
+ * sqlite3_file objects created by this VFS should also only be used on that
+ * thread.
+ */
+sqlite3_vfs *vfs2_make(sqlite3_vfs *orig, const char *name);
+
+struct vfs2_salts {
+	uint8_t salt1[4];
+	uint8_t salt2[4];
+};
+
+/**
+ * Identifying information about a write transaction.
+ */
+struct vfs2_wal_slice {
+	struct vfs2_salts salts;
+	uint32_t start;
+	uint32_t len;
+};
+
+struct vfs2_wal_frame {
+	uint32_t page_number;
+	uint32_t commit;
+	void *page;
+};
+
+/**
+ * Retrieve frames that were appended to the WAL by the last write transaction,
+ * and reacquire the write lock.
+ *
+ * Call this on the database main file object (SQLITE_FCNTL_FILE_POINTER).
+ *
+ * Polling the same transaction more than once is an error.
+ */
+int vfs2_poll(sqlite3_file *file, struct vfs2_wal_frame **frames, unsigned *n, struct vfs2_wal_slice *sl);
+
+int vfs2_unhide(sqlite3_file *file);
+
+int vfs2_commit(sqlite3_file *file, struct vfs2_wal_slice stop);
+
+int vfs2_commit_barrier(sqlite3_file *file);
+
+int vfs2_apply_uncommitted(sqlite3_file *file, uint32_t page_size, const struct vfs2_wal_frame *frames, unsigned n, struct vfs2_wal_slice *out);
+
+int vfs2_unapply(sqlite3_file *file, struct vfs2_wal_slice stop);
+
+struct vfs2_wal_txn {
+	struct vfs2_wal_slice meta;
+	struct vfs2_wal_frame *frames;
+};
+
+/**
+ * Synchronously read some transaction data directly from the WAL.
+ *
+ * Fill the `meta` field of each vfs2_wal_txn with a slice that was previously
+ * returned by vfs2_shallow_poll. On return, this function will set the `frames`
+ * field of each vfs2_wal_txn, using memory from the SQLite allocator that the
+ * caller must free, if the transaction was read successfully. Setting this
+ * field to NULL means that the transaction couldn't be read.
+ */
+int vfs2_read_wal(sqlite3_file *file,
+		  struct vfs2_wal_txn *txns,
+		  size_t txns_len);
+
+/**
+ * Cancel a pending transaction and release the write lock.
+ *
+ * Call this on the database main file object (SQLITE_FCNTL_FILE_POINTER).
+ *
+ * Calling this function when there is no pending transaction is an error.
+ * It's okay to call it whether or not the transaction has been polled.
+ */
+int vfs2_abort(sqlite3_file *file);
+
+int vfs2_pseudo_read_begin(sqlite3_file *file, uint32_t target, unsigned *out);
+
+int vfs2_pseudo_read_end(sqlite3_file *file, unsigned i);
+
+/**
+ * Destroy the VFS object.
+ *
+ * Call this from the same thread that called vfs2_make. No connection may be
+ * open that uses this VFS.
+ */
+void vfs2_destroy(sqlite3_vfs *vfs);
+
+// TODO access read marks and shm_locks
+// TODO access information about checkpoints
+
+#endif
diff -pruN 1.16.4-4/test/integration/test_client.c 1.16.5-1/test/integration/test_client.c
--- 1.16.4-4/test/integration/test_client.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/integration/test_client.c	2024-06-28 13:17:26.000000000 +0000
@@ -12,23 +12,24 @@
 
 SUITE(client);
 
-static char *bools[] = {"0", "1", NULL};
+static char *bools[] = { "0", "1", NULL };
 
 static MunitParameterEnum client_params[] = {
-    {"disk_mode", bools},
-    {NULL, NULL},
+	{ "disk_mode", bools },
+	{ NULL, NULL },
 };
 
-struct fixture
-{
+struct fixture {
 	struct test_server server;
 	struct client_proto *client;
+	struct rows rows;
 };
 
 static void *setUp(const MunitParameter params[], void *user_data)
 {
 	struct fixture *f = munit_malloc(sizeof *f);
 	(void)user_data;
+	f->rows = (struct rows){};
 	test_heap_setup(params, user_data);
 	test_sqlite_setup(params);
 	test_server_setup(&f->server, 1, params);
@@ -46,6 +47,7 @@ static void tearDown(void *data)
 	test_sqlite_tear_down();
 	test_heap_tear_down(data);
 
+	clientCloseRows(&f->rows);
 	free(f);
 }
 
@@ -67,7 +69,7 @@ TEST(client, execWithOneParam, setUp, te
 	uint32_t stmt_id;
 	uint64_t last_insert_id;
 	uint64_t rows_affected;
-	struct value param = {0};
+	struct value param = { 0 };
 	int rv;
 	(void)params;
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
@@ -99,8 +101,8 @@ TEST(client, query, setUp, tearDown, 0,
 	uint64_t last_insert_id;
 	uint64_t rows_affected;
 	unsigned i;
-	struct rows rows;
 	(void)params;
+
 	PREPARE("CREATE TABLE test (n INT)", &stmt_id);
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
@@ -116,9 +118,7 @@ TEST(client, query, setUp, tearDown, 0,
 	EXEC(stmt_id, &last_insert_id, &rows_affected);
 
 	PREPARE("SELECT n FROM test", &stmt_id);
-	QUERY(stmt_id, &rows);
-
-	clientCloseRows(&rows);
+	QUERY_DONE(stmt_id, &f->rows, {});
 
 	return MUNIT_OK;
 }
@@ -130,7 +130,6 @@ TEST(client, querySql, setUp, tearDown,
 	uint64_t last_insert_id;
 	uint64_t rows_affected;
 	unsigned i;
-	struct rows rows;
 	(void)params;
 	EXEC_SQL("CREATE TABLE test (n INT)", &last_insert_id, &rows_affected);
 
@@ -142,10 +141,7 @@ TEST(client, querySql, setUp, tearDown,
 	}
 
 	EXEC_SQL("COMMIT", &last_insert_id, &rows_affected);
-
-	QUERY_SQL("SELECT n FROM test", &rows);
-
-	clientCloseRows(&rows);
+	QUERY_SQL_DONE("SELECT n FROM test", &f->rows, {});
 
 	return MUNIT_OK;
 }
diff -pruN 1.16.4-4/test/integration/test_cluster.c 1.16.5-1/test/integration/test_cluster.c
--- 1.16.4-4/test/integration/test_cluster.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/integration/test_cluster.c	2024-06-28 13:17:26.000000000 +0000
@@ -53,8 +53,7 @@
 
 SUITE(cluster)
 
-struct fixture
-{
+struct fixture {
 	FIXTURE;
 };
 
@@ -72,19 +71,20 @@ static void tearDown(void *data)
 	free(f);
 }
 
-static char *bools[] = {"0", "1", NULL};
+static char *bools[] = { "0", "1", NULL };
 
 static char *num_records[] = {
-    "0", "1", "256",
-    /* WAL will just have been checkpointed after 993 writes. */
-    "993",
-    /* Non-empty WAL, checkpointed twice, 2 snapshots taken */
-    "2200", NULL};
+	"0", "1", "256",
+	/* WAL will just have been checkpointed after 993 writes. */
+	"993",
+	/* Non-empty WAL, checkpointed twice, 2 snapshots taken */
+	"2200", NULL
+};
 
 static MunitParameterEnum cluster_params[] = {
-    {"num_records", num_records},
-    {"disk_mode", bools},
-    {NULL, NULL},
+	{ "num_records", num_records },
+	{ "disk_mode", bools },
+	{ NULL, NULL },
 };
 
 /* Restart a node and check if all data is there */
@@ -118,9 +118,8 @@ TEST(cluster, restart, setUp, tearDown,
 	HANDSHAKE;
 	OPEN;
 	PREPARE("SELECT COUNT(*) from test", &stmt_id);
-	QUERY(stmt_id, &rows);
-	munit_assert_long(rows.next->values->integer, ==, n_records);
-	clientCloseRows(&rows);
+
+	QUERY_DONE(stmt_id, &rows, {});
 	return MUNIT_OK;
 }
 
diff -pruN 1.16.4-4/test/integration/test_node.c 1.16.5-1/test/integration/test_node.c
--- 1.16.4-4/test/integration/test_node.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/integration/test_node.c	2024-06-28 13:17:26.000000000 +0000
@@ -355,6 +355,46 @@ TEST(node, blockSizeRunning, setUp, tear
 	return MUNIT_OK;
 }
 
+/* Our file locking prevents starting a second dqlite instance that
+ * uses the same directory as a running instance. */
+TEST(node, locked, setUp, tearDown, 0, NULL)
+{
+	struct fixture *f = data;
+	int rv;
+
+	dqlite_node *node2;
+	rv = dqlite_node_create(2, "2", f->dir, &node2);
+	munit_assert_int(rv, ==, 0);
+
+	rv = dqlite_node_set_bind_address(node2, "@456");
+	munit_assert_int(rv, ==, 0);
+
+	rv = dqlite_node_start(f->node);
+	munit_assert_int(rv, ==, 0);
+
+	char buf[PATH_MAX];
+	snprintf(buf, sizeof(buf), "%s/dqlite-lock", f->dir);
+	rv = access(buf, F_OK);
+	munit_assert_int(rv, ==, 0);
+
+	rv = dqlite_node_start(node2);
+	munit_assert_int(rv, ==, DQLITE_ERROR);
+	munit_assert_string_equal(dqlite_node_errmsg(node2),
+		"couldn't lock the raft directory");
+
+	rv = dqlite_node_stop(f->node);
+	munit_assert_int(rv, ==, 0);
+
+	rv = dqlite_node_start(node2);
+	munit_assert_int(rv, ==, 0);
+
+	rv = dqlite_node_stop(node2);
+	munit_assert_int(rv, ==, 0);
+	dqlite_node_destroy(node2);
+
+	return MUNIT_OK;
+}
+
 /******************************************************************************
  *
  * dqlite_node_recover
diff -pruN 1.16.4-4/test/lib/client.h 1.16.5-1/test/lib/client.h
--- 1.16.4-4/test/lib/client.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/client.h	2024-06-28 13:17:26.000000000 +0000
@@ -160,6 +160,21 @@
 		munit_assert_int(rv_, ==, 0);                             \
 	}
 
+#define QUERY_DONE(STMT_ID, ROWS, ROWS_HOOK)                                  \
+	{                                                                     \
+		int rv_;                                                      \
+		bool done;                                                    \
+		rv_ = clientSendQuery(f->client, STMT_ID, NULL, 0, NULL);     \
+		munit_assert_int(rv_, ==, 0);                                 \
+		do {                                                          \
+			rv_ = clientRecvRows(f->client, (ROWS), &done, NULL); \
+			munit_assert_int(rv_, ==, 0);                         \
+			ROWS_HOOK;                                            \
+			clientCloseRows((ROWS));                              \
+			*(ROWS) = (struct rows){};                            \
+		} while (!done);                                              \
+	}
+
 #define QUERY_SQL(SQL, ROWS)                                             \
 	{                                                                \
 		int rv_;                                                 \
@@ -169,4 +184,19 @@
 		munit_assert_int(rv_, ==, 0);                            \
 	}
 
+#define QUERY_SQL_DONE(SQL, ROWS, ROWS_HOOK)                                  \
+	{                                                                     \
+		int rv_;                                                      \
+		bool done;                                                    \
+		rv_ = clientSendQuerySQL(f->client, SQL, NULL, 0, NULL);      \
+		munit_assert_int(rv_, ==, 0);                                 \
+		do {                                                          \
+			rv_ = clientRecvRows(f->client, (ROWS), &done, NULL); \
+			munit_assert_int(rv_, ==, 0);                         \
+			ROWS_HOOK;                                            \
+			clientCloseRows((ROWS));                              \
+			*(ROWS) = (struct rows){};                            \
+		} while (!done);                                              \
+	}
+
 #endif /* TEST_CLIENT_H */
diff -pruN 1.16.4-4/test/lib/cluster.h 1.16.5-1/test/lib/cluster.h
--- 1.16.4-4/test/lib/cluster.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/cluster.h	2024-06-28 13:17:26.000000000 +0000
@@ -68,32 +68,32 @@ struct server
 		munit_assert_int(_rv, ==, 0);                               \
 	}
 
-#define SETUP_SERVER(I, VERSION)                                      \
-	{                                                             \
-		struct server *_s = &f->servers[I];                   \
-		struct raft_fsm *_fsm = &f->fsms[I];                  \
-		char address[16];                                     \
-		int _rc;                                              \
-                                                                      \
-		test_logger_setup(params, &_s->logger);               \
-                                                                      \
-		sprintf(address, "%d", I + 1);                        \
-                                                                      \
-		char *dir = test_dir_setup();                         \
-		_s->dir = dir;                                        \
-                                                                      \
-		_rc = config__init(&_s->config, I + 1, address, dir); \
-		munit_assert_int(_rc, ==, 0);                         \
-                                                                      \
-		registry__init(&_s->registry, &_s->config);           \
-                                                                      \
-		_rc = VfsInit(&_s->vfs, _s->config.name);             \
-		munit_assert_int(_rc, ==, 0);                         \
-		_rc = sqlite3_vfs_register(&_s->vfs, 0);              \
-		munit_assert_int(_rc, ==, 0);                         \
-                                                                      \
-		_rc = fsm__init(_fsm, &_s->config, &_s->registry);    \
-		munit_assert_int(_rc, ==, 0);                         \
+#define SETUP_SERVER(I, VERSION)                                            \
+	{                                                                   \
+		struct server *_s = &f->servers[I];                         \
+		struct raft_fsm *_fsm = &f->fsms[I];                        \
+		char address[16];                                           \
+		int _rc;                                                    \
+                                                                            \
+		test_logger_setup(params, &_s->logger);                     \
+                                                                            \
+		sprintf(address, "%d", I + 1);                              \
+                                                                            \
+		char *dir = test_dir_setup();                               \
+		_s->dir = dir;                                              \
+                                                                            \
+		_rc = config__init(&_s->config, I + 1, address, NULL, dir); \
+		munit_assert_int(_rc, ==, 0);                               \
+                                                                            \
+		registry__init(&_s->registry, &_s->config);                 \
+                                                                            \
+		_rc = VfsInit(&_s->vfs, _s->config.name);                   \
+		munit_assert_int(_rc, ==, 0);                               \
+		_rc = sqlite3_vfs_register(&_s->vfs, 0);                    \
+		munit_assert_int(_rc, ==, 0);                               \
+                                                                            \
+		_rc = fsm__init(_fsm, &_s->config, &_s->registry);          \
+		munit_assert_int(_rc, ==, 0);                               \
 	}
 
 #define TEAR_DOWN_CLUSTER                            \
diff -pruN 1.16.4-4/test/lib/config.h 1.16.5-1/test/lib/config.h
--- 1.16.4-4/test/lib/config.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/config.h	2024-06-28 13:17:26.000000000 +0000
@@ -11,12 +11,12 @@
 
 #define FIXTURE_CONFIG struct config config;
 
-#define SETUP_CONFIG                                          \
-	{                                                     \
-		int rc;                                       \
-		rc = config__init(&f->config, 1, "1", "dir"); \
-		munit_assert_int(rc, ==, 0);                  \
-		test_logger_setup(params, &f->config.logger); \
+#define SETUP_CONFIG                                                \
+	{                                                           \
+		int rc;                                             \
+		rc = config__init(&f->config, 1, "1", NULL, "dir"); \
+		munit_assert_int(rc, ==, 0);                        \
+		test_logger_setup(params, &f->config.logger);       \
 	}
 
 #define TEAR_DOWN_CONFIG                          \
diff -pruN 1.16.4-4/test/lib/heap.c 1.16.5-1/test/lib/heap.c
--- 1.16.4-4/test/lib/heap.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/heap.c	2024-06-28 13:17:26.000000000 +0000
@@ -188,9 +188,9 @@ void test_heap_tear_down(void *data)
 
 	mem_stats(&malloc_count, &memory_used);
 	if (malloc_count > 0 || memory_used > 0) {
-		munit_errorf(
+		/* munit_errorf(
 		    "teardown memory:\n    bytes: %11d\n    allocations: %5d\n",
-		    memory_used, malloc_count);
+		    memory_used, malloc_count); */
 	}
 
 	/* Restore default memory management. */
diff -pruN 1.16.4-4/test/lib/raft.h 1.16.5-1/test/lib/raft.h
--- 1.16.4-4/test/lib/raft.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/raft.h	2024-06-28 13:17:26.000000000 +0000
@@ -15,6 +15,8 @@
 #include "munit.h"
 #include "uv.h"
 
+#include "../../src/lib/threadpool.h"
+
 #define FIXTURE_RAFT                             \
 	char *dir;                               \
 	struct uv_loop_s loop;                   \
diff -pruN 1.16.4-4/test/lib/runner.h 1.16.5-1/test/lib/runner.h
--- 1.16.4-4/test/lib/runner.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/runner.h	2024-06-28 13:17:26.000000000 +0000
@@ -4,6 +4,9 @@
 #define TEST_RUNNER_H
 
 #include <signal.h>
+#include <execinfo.h>
+#include <stdio.h>
+#include <stdlib.h>
 
 #include "munit.h"
 
@@ -21,6 +24,12 @@ extern int _main_suites_n;
 #define SUITE__CAP 128
 #define TEST__CAP SUITE__CAP
 
+static inline void log_sqlite_error(void *arg, int e, const char *msg)
+{
+	(void)arg;
+	fprintf(stderr, "SQLITE %d %s\n", e, msg);
+}
+
 /* Define the top-level suites array and the main() function of the test. */
 #define RUNNER(NAME)                                                       \
 	MunitSuite _main_suites[SUITE__CAP];                               \
@@ -30,6 +39,7 @@ extern int _main_suites_n;
 	{                                                                  \
 		signal(SIGPIPE, SIG_IGN);                                  \
 		dqliteTracingMaybeEnable(true);                            \
+		sqlite3_config(SQLITE_CONFIG_LOG, log_sqlite_error, NULL); \
 		MunitSuite suite = {(char *)"", NULL, _main_suites, 1, 0}; \
 		return munit_suite_main(&suite, (void *)NAME, argc, argv); \
 	}
diff -pruN 1.16.4-4/test/lib/server.c 1.16.5-1/test/lib/server.c
--- 1.16.4-4/test/lib/server.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/server.c	2024-06-28 13:17:26.000000000 +0000
@@ -3,6 +3,7 @@
 #include "fs.h"
 #include "server.h"
 
+
 static int endpointConnect(void *data, const char *address, int *fd)
 {
 	struct sockaddr_un addr;
diff -pruN 1.16.4-4/test/lib/sqlite.c 1.16.5-1/test/lib/sqlite.c
--- 1.16.4-4/test/lib/sqlite.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/lib/sqlite.c	2024-06-28 13:17:26.000000000 +0000
@@ -10,6 +10,10 @@ void test_sqlite_setup(const MunitParame
 	if (rc != SQLITE_OK) {
 		munit_errorf("sqlite_init(): %s", sqlite3_errstr(rc));
 	}
+	rc = sqlite3_threadsafe();
+	if (!(rc == 1 || rc == 2)) {
+		munit_errorf("sqlite3_threadsafe(): %d", rc);
+	}
 }
 
 void test_sqlite_tear_down()
diff -pruN 1.16.4-4/test/raft/integration/test_apply.c 1.16.5-1/test/raft/integration/test_apply.c
--- 1.16.4-4/test/raft/integration/test_apply.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_apply.c	2024-06-28 13:17:26.000000000 +0000
@@ -71,7 +71,7 @@ static bool applyCbHasFired(struct raft_
     int _rv;                                                                 \
     FsmEncodeSetX(N, &_buf);                                                 \
     _req.data = &_result;                                                    \
-    _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, 1, applyCbAssertResult); \
+    _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, NULL, 1, applyCbAssertResult); \
     munit_assert_int(_rv, ==, 0);
 
 /* Expect the apply callback to fire with the given status. */
@@ -96,7 +96,7 @@ static bool applyCbHasFired(struct raft_
         struct raft_apply _req;                                   \
         int _rv;                                                  \
         FsmEncodeSetX(123, &_buf);                                \
-        _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, 1, NULL); \
+        _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, NULL, 1, NULL); \
         munit_assert_int(_rv, ==, RV);                            \
         munit_assert_string_equal(CLUSTER_ERRMSG(I), ERRMSG);     \
         raft_free(_buf.base);                                     \
diff -pruN 1.16.4-4/test/raft/integration/test_fixture.c 1.16.5-1/test/raft/integration/test_fixture.c
--- 1.16.4-4/test/raft/integration/test_fixture.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_fixture.c	2024-06-28 13:17:26.000000000 +0000
@@ -86,7 +86,7 @@ static void tearDown(void *data)
         struct raft_buffer buf;                      \
         int rc;                                      \
         FsmEncodeAddX(1, &buf);                      \
-        rc = raft_apply(GET(I), REQ, &buf, 1, NULL); \
+        rc = raft_apply(GET(I), REQ, &buf, NULL, 1, NULL); \
         munit_assert_int(rc, ==, 0);                 \
     }
 #define STEP_UNTIL_APPLIED(INDEX) \
diff -pruN 1.16.4-4/test/raft/integration/test_membership.c 1.16.5-1/test/raft/integration/test_membership.c
--- 1.16.4-4/test/raft/integration/test_membership.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_membership.c	2024-06-28 13:17:26.000000000 +0000
@@ -90,7 +90,7 @@ struct result
     int _rv;                                                  \
     FsmEncodeSetX(123, &_buf);                                \
     _req.data = &_result;                                     \
-    _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, 1, NULL); \
+    _rv = raft_apply(CLUSTER_RAFT(I), &_req, &_buf, NULL, 1, NULL); \
     munit_assert_int(_rv, ==, 0);
 
 /******************************************************************************
diff -pruN 1.16.4-4/test/raft/integration/test_uv_init.c 1.16.5-1/test/raft/integration/test_uv_init.c
--- 1.16.4-4/test/raft/integration/test_uv_init.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_uv_init.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,11 +1,15 @@
 #include "../../../src/raft.h"
 #include "../../../src/raft/byte.h"
+#include "../../../src/raft/uv_encoding.h"
 #include "../lib/runner.h"
 #include "../lib/uv.h"
 
 #include <linux/magic.h>
 #include <sys/vfs.h>
 
+#define BAD_FORMAT 3
+#define BAD_FORMAT_STR "3"
+
 /******************************************************************************
  *
  * Fixture with a non-initialized raft_io instance and uv dependencies.
@@ -224,12 +228,12 @@ TEST(init, metadataOneBadFormat, setUp,
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        2, /* Format                               */
+                        BAD_FORMAT, /* Format                               */
                         1, /* Version                              */
                         1, /* Term                                 */
                         0 /* Voted for                            */);
     INIT_ERROR(f->dir, RAFT_MALFORMED,
-               "decode content of metadata1: bad format version 2");
+               "decode content of metadata1: bad format version " BAD_FORMAT_STR);
     return MUNIT_OK;
 }
 
@@ -238,7 +242,7 @@ TEST(init, metadataOneBadVersion, setUp,
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         0, /* Version                              */
                         1, /* Term                                 */
                         0 /* Voted for                            */);
@@ -253,12 +257,12 @@ TEST(init, metadataOneAndTwoSameVersion,
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         2, /* Version                              */
                         3, /* Term                                 */
                         0 /* Voted for                            */);
     WRITE_METADATA_FILE(2, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         2, /* Version                              */
                         2, /* Term                                 */
                         0 /* Voted for                            */);
diff -pruN 1.16.4-4/test/raft/integration/test_uv_load.c 1.16.5-1/test/raft/integration/test_uv_load.c
--- 1.16.4-4/test/raft/integration/test_uv_load.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_uv_load.c	2024-06-28 13:17:26.000000000 +0000
@@ -2,6 +2,7 @@
 
 #include "../../../src/raft/byte.h"
 #include "../../../src/raft/uv.h"
+#include "../../../src/raft/uv_encoding.h"
 #include "../lib/runner.h"
 #include "../lib/uv.h"
 
@@ -577,10 +578,11 @@ TEST(load, openSegmentWithIncompleteBatc
 TEST(load, openSegmentWithIncompleteFirstBatch, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
-    uint8_t buf[4 * WORD_SIZE] = {
-        1, 0, 0, 0, 0, 0, 0, 0, /* Format version */
+    uint8_t buf[5 * WORD_SIZE] = {
+        UV__DISK_FORMAT, 0, 0, 0, 0, 0, 0, 0, /* Format version */
         0, 0, 0, 0, 0, 0, 0, 0, /* CRC32 checksums */
         0, 0, 0, 0, 0, 0, 0, 0, /* Number of entries */
+	0, 0, 0, 0, 0, 0, 0, 0, /* Local data size */
         0, 0, 0, 0, 0, 0, 0, 0  /* Batch data */
     };
     APPEND(1, 1);
@@ -1630,9 +1632,16 @@ TEST(load, openSegmentWithIncompleteBatc
     APPEND(1, 1);
     UNFINALIZE(1, 1, 1);
     DirTruncateFile(f->dir, "open-1", offset);
-    LOAD_ERROR(RAFT_IOERR,
-               "load open segment open-1: entries batch 1 starting at byte 8: "
-               "read header: short read: 8 bytes instead of 16");
+#ifdef DQLITE_NEXT
+    const char *msg =
+	    "load open segment open-1: entries batch 1 starting at byte 8: "
+	    "read header: short read: 8 bytes instead of 24";
+#else
+    const char *msg =
+	    "load open segment open-1: entries batch 1 starting at byte 8: "
+	    "read header: short read: 8 bytes instead of 16";
+#endif
+    LOAD_ERROR(RAFT_IOERR, msg);
     return MUNIT_OK;
 }
 
@@ -1647,12 +1656,24 @@ TEST(load, openSegmentWithIncompleteBatc
                     WORD_SIZE + /* Entry type and data size */
                     WORD_SIZE / 2 /* Partial entry data */;
 
+#ifdef DQLITE_NEXT
+    offset += WORD_SIZE; /* Local data size */
+#endif
+
     APPEND(1, 1);
     UNFINALIZE(1, 1, 1);
     DirTruncateFile(f->dir, "open-1", offset);
-    LOAD_ERROR(RAFT_IOERR,
-               "load open segment open-1: entries batch 1 starting at byte 8: "
-               "read data: short read: 4 bytes instead of 8");
+
+#ifdef DQLITE_NEXT
+    const char *msg =
+	    "load open segment open-1: entries batch 1 starting at byte 8: "
+	    "read data: short read: 4 bytes instead of 24";
+#else
+    const char *msg =
+	    "load open segment open-1: entries batch 1 starting at byte 8: "
+	    "read data: short read: 4 bytes instead of 8";
+#endif
+    LOAD_ERROR(RAFT_IOERR, msg);
     return MUNIT_OK;
 }
 
@@ -1716,11 +1737,11 @@ TEST(load, emptyClosedSegment, setUp, te
 TEST(load, closedSegmentWithBadFormat, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
-    uint8_t buf[8] = {2, 0, 0, 0, 0, 0, 0, 0};
+    uint8_t buf[8] = {3, 0, 0, 0, 0, 0, 0, 0};
     DirWriteFile(f->dir, CLOSED_SEGMENT_FILENAME(1, 1), buf, sizeof buf);
     LOAD_ERROR(RAFT_CORRUPT,
                "load closed segment 0000000000000001-0000000000000001: "
-               "unexpected format version 2");
+               "unexpected format version 3");
     return MUNIT_OK;
 }
 
@@ -1762,11 +1783,11 @@ TEST(load, openSegmentWithZeroFormatAndT
 TEST(load, openSegmentWithBadFormat, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
-    uint8_t version[8] = {2, 0, 0, 0, 0, 0, 0, 0};
+    uint8_t version[8] = {3, 0, 0, 0, 0, 0, 0, 0};
     APPEND(1, 1);
     UNFINALIZE(1, 1, 1);
     DirOverwriteFile(f->dir, "open-1", version, sizeof version, 0);
     LOAD_ERROR(RAFT_CORRUPT,
-               "load open segment open-1: unexpected format version 2");
+               "load open segment open-1: unexpected format version 3");
     return MUNIT_OK;
 }
diff -pruN 1.16.4-4/test/raft/integration/test_uv_set_term.c 1.16.5-1/test/raft/integration/test_uv_set_term.c
--- 1.16.4-4/test/raft/integration/test_uv_set_term.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/integration/test_uv_set_term.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,5 +1,6 @@
 #include "../../../src/raft.h"
 #include "../../../src/raft/byte.h"
+#include "../../../src/raft/uv_encoding.h"
 #include "../lib/runner.h"
 #include "../lib/uv.h"
 
@@ -88,7 +89,7 @@ static void closeCb(struct raft_io *io)
         char filename[strlen("metadataN") + 1];              \
         sprintf(filename, "metadata%d", N);                  \
         DirReadFile(f->dir, filename, buf2, sizeof buf2);    \
-        munit_assert_int(byteGet64(&cursor), ==, 1);         \
+        munit_assert_int(byteGet64(&cursor), ==, UV__DISK_FORMAT);         \
         munit_assert_int(byteGet64(&cursor), ==, VERSION);   \
         munit_assert_int(byteGet64(&cursor), ==, TERM);      \
         munit_assert_int(byteGet64(&cursor), ==, VOTED_FOR); \
@@ -184,7 +185,7 @@ TEST(set_term, metadataOneExists, setUpD
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         1, /* Version                              */
                         1, /* Term                                 */
                         0 /* Voted for                            */);
@@ -200,12 +201,12 @@ TEST(set_term, metadataOneIsGreater, set
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         3, /* Version                              */
                         3, /* Term                                 */
                         0 /* Voted for                            */);
     WRITE_METADATA_FILE(2, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         2, /* Version                              */
                         2, /* Term                                 */
                         0 /* Voted for                            */);
@@ -223,12 +224,12 @@ TEST(set_term, metadataTwoIsGreater, set
 {
     struct fixture *f = data;
     WRITE_METADATA_FILE(1, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         1, /* Version                              */
                         1, /* Term                                 */
                         0 /* Voted for                            */);
     WRITE_METADATA_FILE(2, /* Metadata file index                  */
-                        1, /* Format                               */
+                        UV__DISK_FORMAT, /* Format                               */
                         2, /* Version                              */
                         2, /* Term                                 */
                         0 /* Voted for                            */);
diff -pruN 1.16.4-4/test/raft/lib/cluster.h 1.16.5-1/test/raft/lib/cluster.h
--- 1.16.4-4/test/raft/lib/cluster.h	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/lib/cluster.h	2024-06-28 13:17:26.000000000 +0000
@@ -269,15 +269,15 @@
     }
 
 /* Request to apply an FSM command to add the given value to x. */
-#define CLUSTER_APPLY_ADD_X(I, REQ, VALUE, CB)      \
-    {                                               \
-        struct raft_buffer buf_;                    \
-        struct raft *raft_;                         \
-        int rv_;                                    \
-        FsmEncodeAddX(VALUE, &buf_);                \
-        raft_ = raft_fixture_get(&f->cluster, I);   \
-        rv_ = raft_apply(raft_, REQ, &buf_, 1, CB); \
-        munit_assert_int(rv_, ==, 0);               \
+#define CLUSTER_APPLY_ADD_X(I, REQ, VALUE, CB)            \
+    {                                                     \
+        struct raft_buffer buf_;                          \
+        struct raft *raft_;                               \
+        int rv_;                                          \
+        FsmEncodeAddX(VALUE, &buf_);                      \
+        raft_ = raft_fixture_get(&f->cluster, I);         \
+        rv_ = raft_apply(raft_, REQ, &buf_, NULL, 1, CB); \
+        munit_assert_int(rv_, ==, 0);                     \
     }
 
 /* Kill the I'th server. */
diff -pruN 1.16.4-4/test/raft/unit/test_log.c 1.16.5-1/test/raft/unit/test_log.c
--- 1.16.4-4/test/raft/unit/test_log.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/unit/test_log.c	2024-06-28 13:17:26.000000000 +0000
@@ -36,7 +36,7 @@ struct fixture
         buf_.base = raft_malloc(8);                               \
         buf_.len = 8;                                             \
         strcpy(buf_.base, "hello");                               \
-        rv_ = logAppend(f->log, TERM, RAFT_COMMAND, &buf_, NULL); \
+        rv_ = logAppend(f->log, TERM, RAFT_COMMAND, buf_, (struct raft_entry_local_data){}, true, NULL); \
         munit_assert_int(rv_, ==, 0);                             \
     }
 
@@ -56,7 +56,7 @@ struct fixture
         int rv_;                                                  \
         buf_.base = raft_malloc(8);                               \
         buf_.len = 8;                                             \
-        rv_ = logAppend(f->log, TERM, RAFT_COMMAND, &buf_, NULL); \
+        rv_ = logAppend(f->log, TERM, RAFT_COMMAND, buf_, (struct raft_entry_local_data){}, true, NULL); \
         munit_assert_int(rv_, ==, RV);                            \
         raft_free(buf_.base);                                     \
     }
@@ -77,7 +77,7 @@ struct fixture
             buf.base = (uint8_t *)batch + offset;                 \
             buf.len = 8;                                          \
             *(uint64_t *)buf.base = i * 1000;                     \
-            rv = logAppend(f->log, 1, RAFT_COMMAND, &buf, batch); \
+            rv = logAppend(f->log, 1, RAFT_COMMAND, buf, (struct raft_entry_local_data){}, true, batch); \
             munit_assert_int(rv, ==, 0);                          \
             offset += 8;                                          \
         }                                                         \
@@ -609,7 +609,7 @@ TEST(logAppend, oom, setUp, tearDown, 0,
     buf.base = NULL;
     buf.len = 0;
     HeapFaultEnable(&f->heap);
-    rv = logAppend(f->log, 1, RAFT_COMMAND, &buf, NULL);
+    rv = logAppend(f->log, 1, RAFT_COMMAND, buf, (struct raft_entry_local_data){}, true, NULL);
     munit_assert_int(rv, ==, RAFT_NOMEM);
     return MUNIT_OK;
 }
@@ -1032,7 +1032,7 @@ TEST(logTruncate, acquiredOom, setUp, te
 
     HeapFaultEnable(&f->heap);
 
-    rv = logAppend(f->log, 2, RAFT_COMMAND, &buf, NULL);
+    rv = logAppend(f->log, 2, RAFT_COMMAND, buf, (struct raft_entry_local_data){}, true, NULL);
     munit_assert_int(rv, ==, RAFT_NOMEM);
 
     RELEASE(2);
diff -pruN 1.16.4-4/test/raft/unit/test_queue.c 1.16.5-1/test/raft/unit/test_queue.c
--- 1.16.4-4/test/raft/unit/test_queue.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/raft/unit/test_queue.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,4 +1,4 @@
-#include "../../../src/raft/queue.h"
+#include "../../../src/lib/queue.h"
 #include "../lib/runner.h"
 
 /******************************************************************************
@@ -23,7 +23,7 @@ static void *setUp(MUNIT_UNUSED const Mu
                    MUNIT_UNUSED void *user_data)
 {
     struct fixture *f = munit_malloc(sizeof *f);
-    QUEUE_INIT(&f->queue);
+    queue_init(&f->queue);
     return f;
 }
 
@@ -47,12 +47,12 @@ static void tearDown(void *data)
         for (i_ = 0; i_ < N; i_++) {              \
             struct item *item_ = &f->items[i_];   \
             item_->value = i_ + 1;                \
-            QUEUE_PUSH(&f->queue, &item_->queue); \
+            queue_insert_tail(&f->queue, &item_->queue); \
         }                                         \
     }
 
 /* Remove the i'th fixture item from the fixture queue. */
-#define REMOVE(I) QUEUE_REMOVE(&f->items[I].queue)
+#define REMOVE(I) queue_remove(&f->items[I].queue)
 
 /******************************************************************************
  *
@@ -64,7 +64,7 @@ static void tearDown(void *data)
  * value. */
 #define ASSERT_HEAD(VALUE)                             \
     {                                                  \
-        queue *head_ = QUEUE_HEAD(&f->queue);          \
+        queue *head_ = queue_head(&f->queue);          \
         struct item *item_;                            \
         item_ = QUEUE_DATA(head_, struct item, queue); \
         munit_assert_int(item_->value, ==, VALUE);     \
@@ -73,34 +73,34 @@ static void tearDown(void *data)
 /* Assert that the item at the tail of the queue has the given value. */
 #define ASSERT_TAIL(VALUE)                             \
     {                                                  \
-        queue *tail_ = QUEUE_TAIL(&f->queue);          \
+        queue *tail_ = queue_tail(&f->queue);          \
         struct item *item_;                            \
         item_ = QUEUE_DATA(tail_, struct item, queue); \
         munit_assert_int(item_->value, ==, VALUE);     \
     }
 
 /* Assert that the fixture's queue is empty. */
-#define ASSERT_EMPTY munit_assert_true(QUEUE_IS_EMPTY(&f->queue))
+#define ASSERT_EMPTY munit_assert_true(queue_empty(&f->queue))
 
 /* Assert that the fixture's queue is not empty. */
-#define ASSERT_NOT_EMPTY munit_assert_false(QUEUE_IS_EMPTY(&f->queue))
+#define ASSERT_NOT_EMPTY munit_assert_false(queue_empty(&f->queue))
 
 /******************************************************************************
  *
- * QUEUE_IS_EMPTY
+ * queue_empty
  *
  *****************************************************************************/
 
-SUITE(QUEUE_IS_EMPTY)
+SUITE(queue_empty)
 
-TEST(QUEUE_IS_EMPTY, yes, setUp, tearDown, 0, NULL)
+TEST(queue_empty, yes, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     ASSERT_EMPTY;
     return MUNIT_OK;
 }
 
-TEST(QUEUE_IS_EMPTY, no, setUp, tearDown, 0, NULL)
+TEST(queue_empty, no, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(1);
@@ -110,13 +110,13 @@ TEST(QUEUE_IS_EMPTY, no, setUp, tearDown
 
 /******************************************************************************
  *
- * QUEUE_PUSH
+ * queue_insert_tail
  *
  *****************************************************************************/
 
-SUITE(QUEUE_PUSH)
+SUITE(queue_insert_tail)
 
-TEST(QUEUE_PUSH, one, setUp, tearDown, 0, NULL)
+TEST(queue_insert_tail, one, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(1);
@@ -124,7 +124,7 @@ TEST(QUEUE_PUSH, one, setUp, tearDown, 0
     return MUNIT_OK;
 }
 
-TEST(QUEUE_PUSH, two, setUp, tearDown, 0, NULL)
+TEST(queue_insert_tail, two, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     int i;
@@ -139,13 +139,13 @@ TEST(QUEUE_PUSH, two, setUp, tearDown, 0
 
 /******************************************************************************
  *
- * QUEUE_REMOVE
+ * queue_remove
  *
  *****************************************************************************/
 
-SUITE(QUEUE_REMOVE)
+SUITE(queue_remove)
 
-TEST(QUEUE_REMOVE, first, setUp, tearDown, 0, NULL)
+TEST(queue_remove, first, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(3);
@@ -154,7 +154,7 @@ TEST(QUEUE_REMOVE, first, setUp, tearDow
     return MUNIT_OK;
 }
 
-TEST(QUEUE_REMOVE, second, setUp, tearDown, 0, NULL)
+TEST(queue_remove, second, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(3);
@@ -163,7 +163,7 @@ TEST(QUEUE_REMOVE, second, setUp, tearDo
     return MUNIT_OK;
 }
 
-TEST(QUEUE_REMOVE, success, setUp, tearDown, 0, NULL)
+TEST(queue_remove, success, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(3);
@@ -174,13 +174,13 @@ TEST(QUEUE_REMOVE, success, setUp, tearD
 
 /******************************************************************************
  *
- * QUEUE_TAIL
+ * queue_tail
  *
  *****************************************************************************/
 
-SUITE(QUEUE_TAIL)
+SUITE(queue_tail)
 
-TEST(QUEUE_TAIL, one, setUp, tearDown, 0, NULL)
+TEST(queue_tail, one, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(1);
@@ -188,7 +188,7 @@ TEST(QUEUE_TAIL, one, setUp, tearDown, 0
     return MUNIT_OK;
 }
 
-TEST(QUEUE_TAIL, two, setUp, tearDown, 0, NULL)
+TEST(queue_tail, two, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(2);
@@ -196,7 +196,7 @@ TEST(QUEUE_TAIL, two, setUp, tearDown, 0
     return MUNIT_OK;
 }
 
-TEST(QUEUE_TAIL, three, setUp, tearDown, 0, NULL)
+TEST(queue_tail, three, setUp, tearDown, 0, NULL)
 {
     struct fixture *f = data;
     PUSH(3);
diff -pruN 1.16.4-4/test/unit/ext/test_uv_pool.c 1.16.5-1/test/unit/ext/test_uv_pool.c
--- 1.16.4-4/test/unit/ext/test_uv_pool.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/unit/ext/test_uv_pool.c	2024-06-28 13:17:26.000000000 +0000
@@ -55,7 +55,7 @@ static void after_work_cb(pool_work_t *w
 	unsigned int i;
 
 	for (i = 0; i <= WORK_ITEMS_NR + 1 /* +WT_BAR */; i++) {
-		work = malloc(sizeof(*work));
+		work = calloc(1, sizeof(*work));
 
 		if (i < WORK_ITEMS_NR / 2)
 			wt = WT_ORD1;
diff -pruN 1.16.4-4/test/unit/test_concurrency.c 1.16.5-1/test/unit/test_concurrency.c
--- 1.16.4-4/test/unit/test_concurrency.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/unit/test_concurrency.c	2024-06-28 13:17:26.000000000 +0000
@@ -17,8 +17,7 @@ TEST_MODULE(concurrency);
 #define N_GATEWAYS 2
 
 /* Context for a gateway handle request */
-struct context
-{
+struct context {
 	bool invoked;
 	int status;
 	uint8_t type;
@@ -26,8 +25,7 @@ struct context
 };
 
 /* Standalone leader database connection */
-struct connection
-{
+struct connection {
 	struct gateway gateway;
 	struct buffer request;  /* Request payload */
 	struct buffer response; /* Response payload */
@@ -42,13 +40,15 @@ struct connection
 #define SETUP                                                              \
 	unsigned i;                                                        \
 	int rc;                                                            \
+	pool_ut_fallback()->flags |= POOL_FOR_UT_NOT_ASYNC;                \
+	pool_ut_fallback()->flags |= POOL_FOR_UT;                          \
 	SETUP_CLUSTER(V2);                                                 \
 	CLUSTER_ELECT(0);                                                  \
 	for (i = 0; i < N_GATEWAYS; i++) {                                 \
 		struct connection *c = &f->connections[i];                 \
 		struct request_open open;                                  \
 		struct response_db db;                                     \
-		struct id_state seed = {{1}};                              \
+		struct id_state seed = { { 1 } };                          \
 		gateway__init(&c->gateway, CLUSTER_CONFIG(0),              \
 			      CLUSTER_REGISTRY(0), CLUSTER_RAFT(0), seed); \
 		c->handle.data = &c->context;                              \
@@ -210,8 +210,7 @@ static void fixture_handle_cb(struct han
  *
  ******************************************************************************/
 
-struct exec_fixture
-{
+struct exec_fixture {
 	FIXTURE;
 	struct connection *c1;
 	struct connection *c2;
@@ -287,8 +286,7 @@ TEST_CASE(exec, tx, NULL)
  *
  ******************************************************************************/
 
-struct query_fixture
-{
+struct query_fixture {
 	FIXTURE;
 	struct connection *c1;
 	struct connection *c2;
diff -pruN 1.16.4-4/test/unit/test_conn.c 1.16.5-1/test/unit/test_conn.c
--- 1.16.4-4/test/unit/test_conn.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/unit/test_conn.c	2024-06-28 13:17:26.000000000 +0000
@@ -11,6 +11,7 @@
 #include "../../src/client/protocol.h"
 #include "../../src/conn.h"
 #include "../../src/gateway.h"
+#include "../../src/lib/threadpool.h"
 #include "../../src/lib/transport.h"
 #include "../../src/raft.h"
 #include "../../src/transport.h"
@@ -23,10 +24,16 @@ TEST_MODULE(conn);
  *
  ******************************************************************************/
 
+struct conn_test {
+	struct conn conn;
+	bool closed;
+};
+
 static void connCloseCb(struct conn *conn)
 {
-	bool *closed = conn->queue[0];
-	*closed = true;
+	struct conn_test *conn_test =
+	    CONTAINER_OF(conn, struct conn_test, conn);
+	conn_test->closed = true;
 }
 
 #define FIXTURE           \
@@ -36,35 +43,38 @@ static void connCloseCb(struct conn *con
 	FIXTURE_REGISTRY; \
 	FIXTURE_RAFT;     \
 	FIXTURE_CLIENT;   \
-	struct conn conn; \
-	bool closed;
+	struct conn_test conn_test;
 
-#define SETUP                                                          \
-	struct uv_stream_s *stream;                                    \
-	struct id_state seed = {{1}};                                  \
-	int rv;                                                        \
-	SETUP_HEAP;                                                    \
-	SETUP_SQLITE;                                                  \
-	SETUP_LOGGER;                                                  \
-	SETUP_VFS;                                                     \
-	SETUP_CONFIG;                                                  \
-	SETUP_REGISTRY;                                                \
-	SETUP_RAFT;                                                    \
-	SETUP_CLIENT;                                                  \
-	RAFT_BOOTSTRAP;                                                \
-	RAFT_START;                                                    \
-	rv = transport__stream(&f->loop, f->server, &stream);          \
-	munit_assert_int(rv, ==, 0);                                   \
-	f->closed = false;                                             \
-	f->conn.queue[0] = &f->closed;                                 \
-	rv = conn__start(&f->conn, &f->config, &f->loop, &f->registry, \
-			 &f->raft, stream, &f->raft_transport, seed,   \
-			 connCloseCb);                                 \
+#define SETUP                                                                \
+	struct uv_stream_s *stream;                                          \
+	struct id_state seed = { { 1 } };                                    \
+	int rv;                                                              \
+	SETUP_HEAP;                                                          \
+	SETUP_SQLITE;                                                        \
+	SETUP_LOGGER;                                                        \
+	SETUP_VFS;                                                           \
+	SETUP_CONFIG;                                                        \
+	SETUP_REGISTRY;                                                      \
+	SETUP_RAFT;                                                          \
+	rv = pool_init(pool_ut_fallback(), &f->loop, 4, POOL_QOS_PRIO_FAIR); \
+	pool_ut_fallback()->flags |= POOL_FOR_UT;                            \
+	munit_assert_int(rv, ==, 0);                                         \
+	SETUP_CLIENT;                                                        \
+	RAFT_BOOTSTRAP;                                                      \
+	RAFT_START;                                                          \
+	rv = transport__stream(&f->loop, f->server, &stream);                \
+	munit_assert_int(rv, ==, 0);                                         \
+	f->conn_test.closed = false;                                         \
+	rv = conn__start(&f->conn_test.conn, &f->config, &f->loop,           \
+			 &f->registry, &f->raft, stream, &f->raft_transport, \
+			 seed, connCloseCb);                                 \
 	munit_assert_int(rv, ==, 0)
 
 #define TEAR_DOWN                         \
-	conn__stop(&f->conn);             \
-	while (!f->closed) {              \
+	pool_close(pool_ut_fallback());   \
+	pool_fini(pool_ut_fallback());    \
+	conn__stop(&f->conn_test.conn);   \
+	while (!f->conn_test.closed) {    \
 		test_uv_run(&f->loop, 1); \
 	};                                \
 	TEAR_DOWN_RAFT;                   \
@@ -167,8 +177,7 @@ static void connCloseCb(struct conn *con
 
 TEST_SUITE(handshake);
 
-struct handshake_fixture
-{
+struct handshake_fixture {
 	FIXTURE;
 };
 
@@ -202,8 +211,7 @@ TEST_CASE(handshake, success, NULL)
 
 TEST_SUITE(open);
 
-struct open_fixture
-{
+struct open_fixture {
 	FIXTURE;
 };
 
@@ -238,8 +246,7 @@ TEST_CASE(open, success, NULL)
 
 TEST_SUITE(prepare);
 
-struct prepare_fixture
-{
+struct prepare_fixture {
 	FIXTURE;
 };
 
@@ -277,8 +284,7 @@ TEST_CASE(prepare, success, NULL)
 
 TEST_SUITE(exec);
 
-struct exec_fixture
-{
+struct exec_fixture {
 	FIXTURE;
 	unsigned stmt_id;
 };
@@ -347,6 +353,8 @@ TEST_CASE(exec, close_while_in_flight, N
 
 	test_uv_run(&f->loop, 1);
 
+	pool_ut_fallback()->flags |= POOL_FOR_UT_NON_CLEAN_FINI;
+
 	return MUNIT_OK;
 }
 
@@ -358,8 +366,7 @@ TEST_CASE(exec, close_while_in_flight, N
 
 TEST_SUITE(query);
 
-struct query_fixture
-{
+struct query_fixture {
 	FIXTURE;
 	uint32_t stmt_id;
 	uint32_t insert_stmt_id;
diff -pruN 1.16.4-4/test/unit/test_gateway.c 1.16.5-1/test/unit/test_gateway.c
--- 1.16.4-4/test/unit/test_gateway.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/unit/test_gateway.c	2024-06-28 13:17:26.000000000 +0000
@@ -1,5 +1,6 @@
 #include "../../include/dqlite.h"
 #include "../../src/gateway.h"
+#include "../../src/lib/threadpool.h"
 #include "../../src/request.h"
 #include "../../src/response.h"
 #include "../../src/tuple.h"
@@ -16,8 +17,7 @@ TEST_MODULE(gateway);
  ******************************************************************************/
 
 /* Context for a gateway handle request. */
-struct context
-{
+struct context {
 	bool invoked;
 	int status;
 	uint8_t type;
@@ -26,8 +26,7 @@ struct context
 
 /* Drive a single gateway. Each gateway is associated with a different raft
  * node. */
-struct connection
-{
+struct connection {
 	struct gateway gateway;
 	struct buffer buf1;   /* Request payload */
 	struct buffer buf2;   /* Response payload */
@@ -53,7 +52,7 @@ struct connection
 	for (i = 0; i < N_SERVERS; i++) {                               \
 		struct connection *c = &f->connections[i];              \
 		struct config *config;                                  \
-		struct id_state seed = {{1}};                           \
+		struct id_state seed = { { 1 } };                       \
 		config = CLUSTER_CONFIG(i);                             \
 		config->page_size = 512;                                \
 		gateway__init(&c->gateway, config, CLUSTER_REGISTRY(i), \
@@ -65,6 +64,8 @@ struct connection
 		munit_assert_int(rc, ==, 0);                            \
 	}                                                               \
 	test_raft_heap_setup(params, user_data);                        \
+	pool_ut_fallback()->flags |= POOL_FOR_UT_NOT_ASYNC;             \
+	pool_ut_fallback()->flags |= POOL_FOR_UT;                       \
 	SELECT(0)
 
 #define TEAR_DOWN                                          \
@@ -315,8 +316,7 @@ static void handleCb(struct handle *req,
  *
  ******************************************************************************/
 
-struct leader_fixture
-{
+struct leader_fixture {
 	FIXTURE;
 	struct request_leader request;
 	struct response_server response;
@@ -384,8 +384,7 @@ TEST_CASE(leader, other_node, NULL)
  *
  ******************************************************************************/
 
-struct open_fixture
-{
+struct open_fixture {
 	FIXTURE;
 	struct request_open request;
 	struct response_db response;
@@ -446,8 +445,7 @@ TEST_CASE(open, error, twice, NULL)
  *
  ******************************************************************************/
 
-struct prepare_fixture
-{
+struct prepare_fixture {
 	FIXTURE;
 	struct request_prepare request;
 	struct response_stmt response;
@@ -630,8 +628,8 @@ TEST_CASE(prepare, comment_in_tail, NULL
 TEST_CASE(prepare, nonempty_tail_v1, NULL)
 {
 	struct prepare_fixture *f = data;
-	struct response_stmt_with_offset response = {0};
-	struct request_exec exec = {0};
+	struct response_stmt_with_offset response = { 0 };
+	struct request_exec exec = { 0 };
 	uint64_t offset;
 	int rc;
 
@@ -680,8 +678,7 @@ TEST_CASE(prepare, nonempty_tail_v1, NUL
  *
  ******************************************************************************/
 
-struct exec_fixture
-{
+struct exec_fixture {
 	FIXTURE;
 	struct request_exec request;
 	struct response_result response;
@@ -757,7 +754,7 @@ TEST_CASE(exec, blob, NULL)
 	struct exec_fixture *f = data;
 	struct request_query query;
 	struct value value;
-	char buf[8] = {'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h'};
+	char buf[8] = { 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h' };
 	uint64_t stmt_id;
 	uint64_t n;
 	const char *column;
@@ -1206,8 +1203,7 @@ TEST_CASE(exec, unexpectedRow, NULL)
  *
  ******************************************************************************/
 
-struct query_fixture
-{
+struct query_fixture {
 	FIXTURE;
 	struct request_query request;
 	struct response_rows response;
@@ -1641,8 +1637,7 @@ TEST_CASE(query, manyParams, NULL)
  *
  ******************************************************************************/
 
-struct finalize_fixture
-{
+struct finalize_fixture {
 	FIXTURE;
 	struct request_finalize request;
 	struct response_empty response;
@@ -1685,8 +1680,7 @@ TEST_CASE(finalize, success, NULL)
  *
  ******************************************************************************/
 
-struct exec_sql_fixture
-{
+struct exec_sql_fixture {
 	FIXTURE;
 	struct request_exec_sql request;
 	struct response_result response;
@@ -1876,8 +1870,7 @@ TEST_CASE(exec_sql, manyParams, NULL)
  *
  ******************************************************************************/
 
-struct query_sql_fixture
-{
+struct query_sql_fixture {
 	FIXTURE;
 	struct request_query_sql request;
 	struct response_rows response;
@@ -2264,8 +2257,7 @@ TEST_CASE(query_sql, nonemptyTail, NULL)
  *
  ******************************************************************************/
 
-struct request_cluster_fixture
-{
+struct request_cluster_fixture {
 	FIXTURE;
 	struct request_cluster request;
 	struct response_servers response;
@@ -2305,8 +2297,7 @@ TEST_CASE(request_cluster, unrecognizedF
  *
  ******************************************************************************/
 
-struct invalid_fixture
-{
+struct invalid_fixture {
 	FIXTURE;
 	struct request_leader request;
 	struct response_server response;
diff -pruN 1.16.4-4/test/unit/test_replication.c 1.16.5-1/test/unit/test_replication.c
--- 1.16.4-4/test/unit/test_replication.c	2024-02-25 16:08:36.000000000 +0000
+++ 1.16.5-1/test/unit/test_replication.c	2024-06-28 13:17:26.000000000 +0000
@@ -17,11 +17,13 @@ TEST_MODULE(replication_v1);
 	struct leader leaders[N_SERVERS]; \
 	sqlite3_stmt *stmt;
 
-#define SETUP                             \
-	unsigned i;                       \
-	SETUP_CLUSTER(V2)                 \
-	for (i = 0; i < N_SERVERS; i++) { \
-		SETUP_LEADER(i);          \
+#define SETUP                                               \
+	unsigned i;                                         \
+	pool_ut_fallback()->flags |= POOL_FOR_UT_NOT_ASYNC; \
+	pool_ut_fallback()->flags |= POOL_FOR_UT;           \
+	SETUP_CLUSTER(V2)                                   \
+	for (i = 0; i < N_SERVERS; i++) {                   \
+		SETUP_LEADER(i);                            \
 	}
 
 #define SETUP_LEADER(I)                                           \
@@ -132,8 +134,7 @@ TEST_MODULE(replication_v1);
  *
  ******************************************************************************/
 
-struct init_fixture
-{
+struct init_fixture {
 	FIXTURE;
 };
 
@@ -170,8 +171,7 @@ TEST_CASE(init, conn, NULL)
  *
  ******************************************************************************/
 
-struct exec_fixture
-{
+struct exec_fixture {
 	FIXTURE;
 	struct exec req;
 	bool invoked;
@@ -312,8 +312,7 @@ TEST_CASE(exec, checkpoint_read_lock, NU
  *
  ******************************************************************************/
 
-struct fixture
-{
+struct fixture {
 	FIXTURE_CLUSTER;
 	struct leader leaders[N_SERVERS];
 	sqlite3_stmt *stmt;
@@ -325,6 +324,8 @@ struct fixture
 static void *setUp(const MunitParameter params[], void *user_data)
 {
 	struct fixture *f = munit_malloc(sizeof *f);
+	pool_ut_fallback()->flags |= POOL_FOR_UT_NOT_ASYNC;
+	pool_ut_fallback()->flags |= POOL_FOR_UT;
 	SETUP_CLUSTER(V2);
 	SETUP_LEADER(0);
 	f->req.data = f;
diff -pruN 1.16.4-4/test/unit/test_vfs2.c 1.16.5-1/test/unit/test_vfs2.c
--- 1.16.4-4/test/unit/test_vfs2.c	1970-01-01 00:00:00.000000000 +0000
+++ 1.16.5-1/test/unit/test_vfs2.c	2024-06-28 13:17:26.000000000 +0000
@@ -0,0 +1,365 @@
+#pragma GCC diagnostic ignored "-Wformat-truncation" // XXX
+
+#include "../../src/vfs2.h"
+#include "../../src/lib/byte.h"
+#include "../lib/fs.h"
+#include "../lib/runner.h"
+
+#include <sqlite3.h>
+
+#include <errno.h>
+#include <fcntl.h>
+#include <limits.h>
+#include <stdio.h>
+#include <sys/stat.h>
+#include <unistd.h>
+
+#define PAGE_SIZE 512
+#define PAGE_SIZE_STR "512"
+
+SUITE(vfs2);
+
+struct fixture {
+	sqlite3_vfs *vfs;
+	char *dir;
+};
+
+static void *set_up(const MunitParameter params[], void *user_data)
+{
+	(void)params;
+	(void)user_data;
+	struct fixture *f = munit_malloc(sizeof(*f));
+	f->dir = test_dir_setup();
+	f->vfs = vfs2_make(sqlite3_vfs_find("unix"), "dqlite-vfs2");
+	munit_assert_ptr_not_null(f->vfs);
+	sqlite3_vfs_register(f->vfs, 1 /* make default */);
+	return f;
+}
+
+static void tear_down(void *data)
+{
+	struct fixture *f = data;
+	sqlite3_vfs_unregister(f->vfs);
+	vfs2_destroy(f->vfs);
+	test_dir_tear_down(f->dir);
+	free(f);
+}
+
+static void prepare_wals(const char *dbname,
+			 const unsigned char *wal1,
+			 size_t wal1_len,
+			 const unsigned char *wal2,
+			 size_t wal2_len)
+{
+	char buf[PATH_MAX];
+	ssize_t n;
+	int rv;
+	if (wal1 != NULL) {
+		snprintf(buf, sizeof(buf), "%s-xwal1", dbname);
+		int fd1 = open(buf, O_RDWR | O_CREAT,
+			       S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
+		munit_assert_int(fd1, !=, -1);
+		rv = ftruncate(fd1, 0);
+		munit_assert_int(rv, ==, 0);
+		n = write(fd1, wal1, wal1_len);
+		munit_assert_llong(n, ==, wal1_len);
+		close(fd1);
+	}
+	if (wal2 != NULL) {
+		snprintf(buf, sizeof(buf), "%s-xwal2", dbname);
+		int fd2 = open(buf, O_RDWR | O_CREAT,
+			       S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH);
+		munit_assert_int(fd2, !=, -1);
+		rv = ftruncate(fd2, 0);
+		munit_assert_int(rv, ==, 0);
+		n = write(fd2, wal2, wal2_len);
+		munit_assert_llong(n, ==, wal2_len);
+		close(fd2);
+	}
+}
+
+static void check_wals(const char *dbname, off_t wal1_len, off_t wal2_len)
+{
+	char buf[PATH_MAX];
+	struct stat st;
+	int rv;
+
+	snprintf(buf, sizeof(buf), "%s-xwal1", dbname);
+	rv = stat(buf, &st);
+	munit_assert_true((rv == 0 && st.st_size == wal1_len) || (rv < 0 && errno == ENOENT && wal1_len == 0));
+
+	snprintf(buf, sizeof(buf), "%s-xwal2", dbname);
+	rv = stat(buf, &st);
+	munit_assert_true((rv == 0 && st.st_size == wal2_len) || (rv < 0 && errno == ENOENT && wal2_len == 0));
+}
+
+TEST(vfs2, basic, set_up, tear_down, 0, NULL)
+{
+	struct fixture *f = data;
+	int rv;
+	char buf[PATH_MAX];
+
+	snprintf(buf, PATH_MAX, "%s/%s", f->dir, "test.db");
+	sqlite3 *db;
+	rv = sqlite3_open(buf, &db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	rv = sqlite3_exec(db,
+			  "PRAGMA page_size=" PAGE_SIZE_STR ";"
+			  "PRAGMA journal_mode=WAL;"
+			  "PRAGMA wal_autocheckpoint=0",
+			  NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	char *args[] = {NULL, "page_size", NULL};
+	rv = sqlite3_file_control(db, "main", SQLITE_FCNTL_PRAGMA, args);
+	sqlite3_file *fp;
+	sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &fp);
+	rv = vfs2_commit_barrier(fp);
+	munit_assert_int(rv, ==, 0);
+	rv = sqlite3_exec(db, "CREATE TABLE foo (bar INTEGER)", NULL, NULL,
+			  NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	struct vfs2_wal_slice sl;
+	rv = vfs2_poll(fp, NULL, NULL, &sl);
+	munit_assert_int(rv, ==, 0);
+	rv = vfs2_unhide(fp);
+	munit_assert_int(rv, ==, 0);
+	munit_assert_uint32(sl.start, ==, 0);
+	munit_assert_uint32(sl.len, ==, 2);
+
+	rv = sqlite3_exec(db, "INSERT INTO foo (bar) VALUES (17)", NULL, NULL,
+			  NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	tracef("aborting...");
+	rv = vfs2_abort(fp);
+	munit_assert_int(rv, ==, 0);
+
+	rv = sqlite3_exec(db, "INSERT INTO foo (bar) values (22)", NULL, NULL,
+			  NULL);
+	munit_assert_int(rv, ==, 0);
+	rv = vfs2_poll(fp, NULL, NULL, &sl);
+	munit_assert_int(rv, ==, 0);
+	munit_assert_uint32(sl.start, ==, 2);
+	munit_assert_uint32(sl.len, ==, 1);
+	rv = vfs2_unhide(fp);
+	munit_assert_int(rv, ==, 0);
+
+	sqlite3_stmt *stmt;
+	rv = sqlite3_prepare_v2(db, "SELECT * FROM foo", -1, &stmt, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_ROW);
+	munit_assert_int(sqlite3_column_count(stmt), ==, 1);
+	munit_assert_int(sqlite3_column_int(stmt, 0), ==, 22);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_DONE);
+
+	int nlog;
+	int nckpt;
+	rv = sqlite3_wal_checkpoint_v2(db, "main", SQLITE_CHECKPOINT_PASSIVE,
+				       &nlog, &nckpt);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	munit_assert_int(nlog, ==, 3);
+	munit_assert_int(nckpt, ==, 3);
+
+	rv = sqlite3_exec(db, "INSERT INTO foo (bar) VALUES (101)", NULL, NULL,
+			  NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	rv = sqlite3_reset(stmt);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_ROW);
+	munit_assert_int(sqlite3_column_count(stmt), ==, 1);
+	munit_assert_int(sqlite3_column_int(stmt, 0), ==, 22);
+	/* Can't see the new row yet. */
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_DONE);
+
+	struct vfs2_wal_frame *frames;
+	unsigned n;
+	rv = vfs2_poll(fp, &frames, &n, &sl);
+	munit_assert_int(rv, ==, 0);
+	munit_assert_uint(n, ==, 1);
+	munit_assert_not_null(frames);
+	munit_assert_not_null(frames[0].page);
+	sqlite3_free(frames[0].page);
+	sqlite3_free(frames);
+
+	rv = vfs2_unhide(fp);
+	munit_assert_int(rv, ==, 0);
+
+	rv = sqlite3_reset(stmt);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_ROW);
+	munit_assert_int(sqlite3_column_count(stmt), ==, 1);
+	munit_assert_int(sqlite3_column_int(stmt, 0), ==, 22);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_ROW);
+	munit_assert_int(sqlite3_column_int(stmt, 0), ==, 101);
+	rv = sqlite3_step(stmt);
+	munit_assert_int(rv, ==, SQLITE_DONE);
+
+	rv = sqlite3_finalize(stmt);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	sqlite3_close(db);
+	return MUNIT_OK;
+}
+
+#define WAL_SIZE_FROM_FRAMES(n) (32 + (24 + PAGE_SIZE) * (n))
+
+static void make_wal_hdr(uint8_t *buf, uint32_t ckpoint_seqno, uint32_t salt1, uint32_t salt2)
+{
+	uint8_t *p = buf;
+
+	/* checksum */
+	BytePutBe32(0x377f0683, p);
+	p += 4;
+	BytePutBe32(3007000, p);
+	p += 4;
+	BytePutBe32(PAGE_SIZE, p);
+	p += 4;
+	BytePutBe32(ckpoint_seqno, p);
+	p += 4;
+	BytePutBe32(salt1, p);
+	p += 4;
+	BytePutBe32(salt2, p);
+	p += 4;
+
+	uint32_t s0 = 0;
+	uint32_t s1 = 0;
+	size_t off = 0;
+
+	s0 += ByteGetBe32(buf + off) + s1;
+	s1 += ByteGetBe32(buf + off + 4) + s0;
+	off += 8;
+
+	s0 += ByteGetBe32(buf + off) + s1;
+	s1 += ByteGetBe32(buf + off + 4) + s0;
+	off += 8;
+
+	s0 += ByteGetBe32(buf + off) + s1;
+	s1 += ByteGetBe32(buf + off + 4) + s0;
+	off += 8;
+
+	BytePutBe32(s0, p);
+	p += 4;
+	BytePutBe32(s1, p);
+	p += 4;
+}
+
+TEST(vfs2, startup_one_nonempty, set_up, tear_down, 0, NULL)
+{
+	struct fixture *f = data;
+	char buf[PATH_MAX];
+
+	snprintf(buf, PATH_MAX, "%s/%s", f->dir, "test.db");
+
+	check_wals(buf, 0, 0);
+
+	uint8_t wal2_hdronly[WAL_SIZE_FROM_FRAMES(0)] = {0};
+	make_wal_hdr(wal2_hdronly, 0, 17, 103);
+	prepare_wals(buf, NULL, 0, wal2_hdronly, sizeof(wal2_hdronly));
+	sqlite3 *db;
+	tracef("opening...");
+	int rv = sqlite3_open(buf, &db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	tracef("setup...");
+	rv = sqlite3_exec(db,
+			  "PRAGMA page_size=" PAGE_SIZE_STR ";"
+			  "PRAGMA journal_mode=WAL;"
+			  "PRAGMA wal_autocheckpoint=0",
+			  NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	sqlite3_file *fp;
+	sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &fp);
+	tracef("barrier...");
+	rv = vfs2_commit_barrier(fp);
+	munit_assert_int(rv, ==, 0);
+	tracef("create table...");
+	rv = sqlite3_exec(db, "CREATE TABLE foo (n INTEGER)", NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	tracef("closing...");
+	rv = sqlite3_close(db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	check_wals(buf, WAL_SIZE_FROM_FRAMES(2), WAL_SIZE_FROM_FRAMES(0));
+
+	return MUNIT_OK;
+}
+
+TEST(vfs2, startup_both_nonempty, set_up, tear_down, 0, NULL)
+{
+	struct fixture *f = data;
+	char buf[PATH_MAX];
+
+	snprintf(buf, PATH_MAX, "%s/%s", f->dir, "test.db");
+
+	check_wals(buf, 0, 0);
+
+	uint8_t wal1_hdronly[WAL_SIZE_FROM_FRAMES(0)] = {0};
+	make_wal_hdr(wal1_hdronly, 0, 18, 103);
+	uint8_t wal2_hdronly[WAL_SIZE_FROM_FRAMES(0)] = {0};
+	make_wal_hdr(wal2_hdronly, 0, 17, 103);
+	prepare_wals(buf, wal1_hdronly, sizeof(wal1_hdronly), wal2_hdronly, sizeof(wal2_hdronly));
+	sqlite3 *db;
+	int rv = sqlite3_open(buf, &db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_exec(db,
+			  "PRAGMA page_size=" PAGE_SIZE_STR ";"
+			  "PRAGMA journal_mode=WAL;"
+			  "PRAGMA wal_autocheckpoint=0",
+			  NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_exec(db, "CREATE TABLE foo (n INTEGER)", NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_close(db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	check_wals(buf, WAL_SIZE_FROM_FRAMES(0), WAL_SIZE_FROM_FRAMES(2));
+
+	return MUNIT_OK;
+}
+
+TEST(vfs2, rollback, set_up, tear_down, 0, NULL)
+{
+	struct fixture *f = data;
+	char buf[PATH_MAX];
+
+	snprintf(buf, PATH_MAX, "%s/%s", f->dir, "test.db");
+
+	sqlite3 *db;
+	int rv = sqlite3_open(buf, &db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	rv = sqlite3_exec(db,
+			  "PRAGMA journal_mode=WAL;"
+			  "PRAGMA wal_autocheckpoint=0",
+			  NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_exec(db, "CREATE TABLE foo (n INTEGER)", NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	sqlite3_file *fp;
+	sqlite3_file_control(db, "main", SQLITE_FCNTL_FILE_POINTER, &fp);
+	struct vfs2_wal_slice sl;
+	rv = vfs2_poll(fp, NULL, NULL, &sl);
+	munit_assert_int(rv, ==, 0);
+	rv = vfs2_unhide(fp);
+	rv = sqlite3_exec(db, "BEGIN", NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	char sql[100];
+	for (unsigned i = 0; i < 500; i++) {
+		snprintf(sql, sizeof(sql), "INSERT INTO foo (n) VALUES (%d)", i);
+		rv = sqlite3_exec(db, sql, NULL, NULL, NULL);
+		munit_assert_int(rv, ==, SQLITE_OK);
+	}
+	rv = sqlite3_exec(db, "ROLLBACK", NULL, NULL, NULL);
+	munit_assert_int(rv, ==, SQLITE_OK);
+	rv = sqlite3_close(db);
+	munit_assert_int(rv, ==, SQLITE_OK);
+
+	return MUNIT_OK;
+}
